#  -*-perl-*-
# runtest.t
# test-dnssec.sh

use strict;                 
use warnings;

use Getopt::Long;

use Test::More;

# OPTIONS:
#
# -o print checker output at end of script
# -p print whole problem text for problems which
#    are incorrect for some reason
# -v print the Checkers debug messages at level 5

my %OPT;

&GetOptions( 
	'o!' => \$OPT{'print_checker_output'},
	'p!' => \$OPT{'verbose_bad_problems'},
	't!' => \$OPT{'verbose_testns'},
	'v!' => \$OPT{'debug_messages'}
);

# Make sure we don't miss warnings

$SIG{__WARN__} = \&perl_warning_handler;

BEGIN {
    # Net::DNS::TestNS import

    # We can't do any problem code testing without TestNS,
    # check it's there first. We could do this with use_ok,
    # but then a whole load of path garbage gets printed

    # We don't do this for the other modules, because they're required
    # for operation of DelCheck so we want people to worry :-)

    if ( eval {  require Net::DNS::TestNS; } ) {
	Net::DNS::TestNS->import();
    }
    else {
	plan skip_all =>  "Net::DNS::TestNS not found";
    }

    # Other Modules

    plan tests => 23;

    # 1: use DelCheck module

    die if ( ! use_ok( 'Net::DelCheck', qw( &errors ) ) );

    # 2: use XML::DOM

    die if ( ! use_ok( 'XML::DOM' ) );

}

SKIP: {

    my $curdir = $0;
    $curdir =~ s/\/[^\/]+$//;

    # If this test requires IPv6, and it's not available
    # according to the module, skip it
    # This is indicated by the presence of a file 'ipv6.needed'
    # in the test directory

    skip "IPv6 not available", 21 if ( 
             ( ! &Net::DelCheck::ipv6_available ) &&
	     ( -e join '/', $curdir, 'ipv6.needed' )
    );

    # Some tests are disabled
    # This is indicated by the presence of a file 'ipv6.needed'
    # The file contains the reason it's disabled.

    my $disabled_file_path = join '/', $curdir, 'test.disabled';

    if ( -e join '/', $curdir, 'test.disabled' ) {
	skip "Test Disabled, see test.disabled", 21;
    }

    my $server_prefix = 'server';
    my $server_affix = 'xml';

    my $delcheck_prefix = 'delcheck';
    my $delcheck_affix = 'conf';

    my $resolver_conf = join '/', $curdir, 'resolver.xml';

    my $output_conf = join '/', $curdir, 'output.conf';
    
    # 3: there's a server conf
    
    my $bad_conf_flag;
    
    my ( $server_names, $server_confs ) = &get_server_glue_details( $curdir, $server_prefix, $server_affix );

    ok( ( @$server_names > 0 ), "found > 0 server configs" ) || ( $bad_conf_flag = 1 );

    # 4: there's a delcheck conf
    
    my @delcheck_confs = &get_delcheck_confs( $curdir, $delcheck_prefix, $delcheck_affix );

    ok( ( @delcheck_confs > 0 ), "found > 0 DelCheck configs" ) || ( $bad_conf_flag = 1 );

    # 5: there's a resolver conf

    ok( -f $resolver_conf, "$resolver_conf exists" ) || ( $bad_conf_flag = 1 );
    
    # 6: there's an output conf and we can read it
    
    my ( $conf_err, @expected_codes ) = &read_output_conf( $output_conf );
    
    if ( ! ok( ! defined $conf_err, 'read output config file' ) ) {
	$bad_conf_flag = 1;
	print STDERR "ERROR: $conf_err\n";
    }
    
  SKIP: {
      if ( $bad_conf_flag ) {
	  skip "bad test configuration", 18;
      }
      else {
	  &do_test( \@delcheck_confs, $server_names, $server_confs, $resolver_conf, @expected_codes );
      }
  }
    
    exit 0;
}

sub do_test
{
    my ( $test_confs, $server_names, $server_confs, $resolver_conf, @expected_codes ) = @_;
    
    # Make checker
    # Note: making a new checker each time doesn't test the
    # persistent checker functionallity
    
    my $checker;
    if ($OPT{'debug_messages'}){
	    $checker= new Net::DelCheck( @$test_confs, { 'ModuleDiagnosticLevel' => '5' });
	}else{
	    $checker= new Net::DelCheck( @$test_confs );
	}

    # 7: made checker

  SKIP: {
      if ( ! isa_ok( $checker, 'Net::DelCheck', 'checker' ) ) {
	  print STDERR "Module error: ". join "\n", &Net::DelCheck::errors, '';
	  skip "DelCheck creation error", 16;
      }

      # Make test servers
      
      my $servers_ok_flag = 1;
      
      my @server_refs;
      foreach my $server_name ( @$server_names ) {

    next if (($server_name =~ /^[0-9\.]+$/) || ($server_name =~ /:{2,8}/)); # glue picked up automatically by TestNS
	  
		my $server_conf = shift @$server_confs;

		my $server_ref = new Net::DNS::TestNS( $server_conf,
 						       {
 							   Verbose => $OPT{'verbose_testns'} ,
 						       }
						       );
		
       		if ( ! $server_ref ){		
			die $Net::DNS::TestNS::errorcondition;
			}



	  if ( defined $server_ref ) {
	      push @server_refs, $server_ref;
	  }
	  else {
	      $servers_ok_flag = 0;
	  }
      }

      # 8: made servers
      
      ok( $servers_ok_flag, 'made test servers' );

      # Make resolving server
      
      my $resolver = new Net::DNS::TestNS( $resolver_conf );
	$resolver->verbose(1) if ( $OPT{'verbose_testns'} );
      
      # 9: made a resolver
      
      isa_ok( $resolver, 'Net::DNS::TestNS', 'resolving server' );
      
      # Run check if we have a checker and some servers
      
    SKIP: {
	skip "no checker to use", 14 unless defined $checker;
	skip "no servers to use", 14 unless $servers_ok_flag;
	skip "no resolver to use", 14 unless defined $resolver;
	
	foreach my $server ( @server_refs ) { $server->run; }
	
	$resolver->run;

	my @errors = $checker->get_data( 'testzone.local', @$server_names );
	
	# 10: no errors from &get_data
	
	if ( ! is( @errors, 0, 'no errors returned by &get_data' ) ) {
	    print STDERR "&get_data errors: ", join "\n", @errors, "\n";
	}
	
	my $xml_schema = $checker->schema;
	
	# Schema checks
	
      SKIP: {
	  # 11: schema defined
	  
	  skip "schema not defined", 3 unless ok( defined $xml_schema, 'schema defined' );
	  
	  # 12: schema longer than 100
	  
	  ok( length $xml_schema > 100, 'schema reasonable length' );
	  
	  # 13: schema starts with <xsd:schema
	  
	  my $xml_schema_start = substr( $xml_schema, 0, 100 );

	  like( $xml_schema_start, qr/<xsd:schema/, 'schema starts with xsd:schema' );
	  
	  # 14: schema ends with /xsd:schema>
	  
	  my $xml_schema_end = substr( $xml_schema, ( ( length $xml_schema ) - 30 ), 30 );
	  like( $xml_schema_end, qr/<\/xsd:schema>$/, 'schema ends with /xsd:schema' );
	  }
	
	# Output checks
      SKIP: {
	  my $xml_out = $checker->check();
    #print STDERR $xml_out;
	  # 15: report output defined

	  skip 'output not defined', 6 unless ok( defined $xml_out, 'output defined' );
	  
	  my $report_tag_name = $checker->xml_tag_name( 'report' );
	  my $entity_tag_name = $checker->xml_tag_name( 'problem_entity' );
	  my $problem_tag_name = $checker->xml_tag_name( 'problem' );
	  my $no_problems_tag_name = $checker->xml_tag_name( 'no_problem' );
	  
	  my $entity_attribute_name =  $checker->xml_attribute_name( 'problem_entity_name' );
	  my $problem_attribute_name = $checker->xml_attribute_name( 'problem_code' );
	  
	  # 16: report root tag defined
	  
	  skip 'report tag not defined', 7 unless ok( defined $report_tag_name, 'report root tag name defined' );
	  
	  # 17: report entity tag defined
	  
	  skip 'entity tag not defined', 6 unless ok( defined $entity_tag_name, 'entity tag name defined'  );
	  
	  # 18: problem tag defined
	  
	  skip 'problem tag not defined', 5 unless ok( defined $problem_tag_name, 'problem tag name defined' );
	  
	  # 19: no_problem tag defined
	  
	  skip 'no_problem tag not defined', 4 unless ok( defined $no_problems_tag_name, 'no_problem tag name defined' );
	  
	  # 20: entity attribute defined
	  
	  skip 'entity attribute not defined', 3 unless ok( defined $entity_attribute_name, 'entity attribute name defined' );
	  
	  # 21: problem attribute defined
	  
	  skip 'problem attribute not defined', 2 unless ok( defined $problem_attribute_name, 'problem attribute name defined' );

	SKIP: {
	    
	    my ( $doc, $err ) = &create_dom( $xml_out );
	    
	    if ( defined $err ) {
		print STDERR $xml_out if ( $OPT{'print_checker_output'} );
		print STDERR $err;
		skip( "couldn't create DOM document", 2 );
	    }
	    
	    # 22: $doc is a reference to the right type
	    
	    skip "bad DOM document ref", 1  unless isa_ok( $doc, 'XML::DOM::Document', 'DOM document' );
	    
	    # 23: output matches the stuff defined in $output_conf

	    my @errors = &check_document_contents( 
		$doc, 
		$report_tag_name,
		$entity_tag_name, $entity_attribute_name,
		$problem_tag_name, $problem_attribute_name,
 	        $no_problems_tag_name,
                @expected_codes
            );

	    unless ( ok( ( @errors == 0 ), 'problem codes in output of &check' ) ) {
		print STDERR 'ERROR: ';
		print STDERR join "\nERROR: ", @errors;
		print STDERR "\n";
	    }
	    
	    print STDERR $xml_out if ( $OPT{'print_checker_output'} );
	}
      }
	
	# Cleanup after checker
	
	$checker->cleanup;
    }
      
      # Stop all nameservers
      
      foreach my $server ( @server_refs ) { $server->stop;  }
      
      $resolver->stop if ( defined $resolver );
  }
}

sub create_dom
{
    my $output = shift;

    my ( $doc, $err );

    if ( $output ) {
	my $parser = new XML::DOM::Parser;
	
	# Parse into DOM doc
	    
	eval {
	    $doc = $parser->parse( $output );
	};
	
	# If we managed to parse it, test the problem codes/arguments it contains
	
	if ( $@ ) {
	    $err = $@;	
	}
    }
    else {
	$err = 'no output from module';
    }
    
    return ( $doc, $err );
}

sub check_document_contents
{
    my ( 
	 $doc, 
	 $report_tag_name,
	 $entity_tag_name, $entity_attribute_name, 
	 $problem_tag_name, $problem_attribute_name,
	 $no_problems_tag_name,
	 @expected_codes 
    ) = @_;

    my @errors;

    # First check all expected problem codes occur with correct
    # entity and match expected regex.

    # @expected_codes is an array of PROBLEM_CODE, ENTITY, REGEX
    # REGEX is a pattern the problem description for PROBLEM_CODE/ENTITY
    # should match

    # second parameter to getElementsByTagName is 'no recurse'

    my %expected_spec;

    my $root_elem = $doc->getDocumentElement();

    # Check we got a root element and it's name is correct

    if ( defined $root_elem ) {
	my $root_name = $root_elem->getTagName;

	if ( $root_name ne $report_tag_name ) {
	    push  @errors, "root element has bad name '$root_name', should be: $report_tag_name)";
	}
    }
    else {
	push @errors, 'document root element undefined';
	return @errors;
    }

    my @entity_nodes = $root_elem->getElementsByTagName( $entity_tag_name, 0 );

    # Check we got all codes that we expected to get for each entity

    foreach my $code_line ( @expected_codes ) {
    
	my ( $exp_entity, $exp_pcode, $exp_times, @exp_regex ) = split /\s*,\s*/, $code_line;

	my $first_found_node;

	my $found_times = 0;

	if ( defined $exp_entity && defined $exp_pcode ) {

	    if ( defined $exp_times ) {
		if ( $exp_times !~ /^\d+$/ ) {
		    push @errors, "bad output specification in output conf file: $code_line ('times' not numeric)";
		    $exp_times = 1;
		}
	    }
	    else {
		$exp_times = 1;
	    }

	    $exp_pcode = uc $exp_pcode;

	    $exp_pcode = 'PROBLEM_'. $exp_pcode if ( $exp_pcode !~ /^PROBLEM_/ );

	    $expected_spec{$exp_entity}{$exp_pcode} = \@exp_regex;
	    
	    # Check we got the right number of the needed code, and that
            # the first one matches all of @reg_exp
	    
	    foreach my $entity_node ( @entity_nodes ) {

		my $attr_node = $entity_node->getAttributeNode( $entity_attribute_name );
		if ( defined $attr_node ) {
		    
		    if ( 
			 ( $attr_node->getValue =~ /$exp_entity/ ) 
		     ) {
			foreach my $pcode_node ( $entity_node->getElementsByTagName( $problem_tag_name, 0 ) ) {

			    my $node_string = $pcode_node->toString;

			    if ( $node_string =~ /UNDEFINED/ ) {
				push @errors, "found an UNDEFINED value in a $problem_tag_name TAG";

			    }
			    
			    my $pcode_attr_node =  $pcode_node->getAttributeNode( $problem_attribute_name );

			    if ( defined $pcode_attr_node ) {
				
				if ( $pcode_attr_node->getValue eq $exp_pcode ) {
				    $first_found_node = $pcode_node;

				    $found_times++;
				}
			    }
			}
		    }
		}
		last if defined $first_found_node;
	    }

	    # See if we got it and it matched

	    if ( defined $first_found_node ) {

		my $found_node_string = $first_found_node->toString;

		my $err_base = "problem code '$exp_pcode' for entity '$exp_entity' ";

		if ( $found_times != $exp_times ) {
		    push @errors, $err_base. "occured $found_times time(s), expected $exp_times occurence(s)";
		}

		foreach my $exp_regex ( @exp_regex ) {

		    if ( $found_node_string !~ /$exp_regex/s ) {
			my $err_msg = $err_base. "didn't match /$exp_regex/s";

			if ( $OPT{'verbose_bad_problems'} ) {
			    $err_msg .= ":\n$found_node_string";
			}

			push @errors, $err_msg;
		    }
		}
	    }
	    else {
		push @errors, "didn't find problem code '$exp_pcode' for entity '$exp_entity'";
	    }
	}
	else {
	    push @errors, "bad output specification in output conf file: $code_line";
	}
    }

    # Then check that no entity/problemcode combinations occur which were not specified

    foreach my $entity_node ( @entity_nodes ) {

	my $entity_attr_node =  $entity_node->getAttributeNode( $entity_attribute_name );

	if ( defined $entity_attr_node ) {
	    my $entity_name =  $entity_node->getAttributeNode( $entity_attribute_name )->getValue;
	    
	    foreach my $problem_node ( $entity_node->getElementsByTagName( $problem_tag_name, 0  ) ) {

		my $problem_attr_node = $problem_node->getAttributeNode( $problem_attribute_name );

		if ( defined $problem_attr_node ) {

		    my $problem_name =  $problem_attr_node->getValue;
		
		    # Check there an expected entity name/problem
		    # matching this occuring entityname/problem

		    my $found_spec_flag = 0;

		    # Program doesn't work if I don't do this, don't
		    # ask me why...

		    my @bob = %expected_spec;

		    while ( my ( $expentname, $ref ) = each %expected_spec ) {
			if ( 
			     ( $entity_name =~ /$expentname/ ) &&
			     ( exists $expected_spec{$expentname}{$problem_name} ) 
 		        ) {
			    $found_spec_flag = 1;
			    last;
			}
		    }

		    if ( ! $found_spec_flag ) {
			my $err_msg = "unexpected problem code '$problem_name' found for entity '$entity_name'";
		    
			if ( $OPT{'verbose_bad_problems'} ) {
			    my $problem_string = $problem_node->toString;
			    $err_msg .= ":\n$problem_string";
			}		    
		    
			push @errors, $err_msg;
		    }
		}
		else {
		    push @errors, "found a $problem_tag_name tag without a $problem_attribute_name attribute";
		}
	    }
	}
	else {
	    push @errors, "found a $entity_tag_name tag without a $entity_attribute_name attribute";
	}
    }

    # Check existence of no_problem tag

    my @no_problem_nodes = $root_elem->getElementsByTagName( $no_problems_tag_name, 0 );

    # If codes we're expected, check that didn't get any
    # 'no problem' children of the root

    if ( @expected_codes > 0 ) {
	if ( @no_problem_nodes > 0 ) {
	    push @errors, "found one or more $no_problems_tag_name element even though problems were expected";
	}
    }

    # If no problem codes were expected, check that we got a (single)
    # no problems node

    else {
   	if ( @no_problem_nodes == 0 ) {
	    push @errors, "no problems expected, but missing $no_problems_tag_name element"; 
	}
	elsif ( @no_problem_nodes > 1 ) {
	    push @errors, "multiple $no_problems_tag_name elements found!"; 
	}
    }

    return @errors;
}

sub read_output_conf
{
    my $conf_path = shift;

    my ( $err, @expected );

    if ( open( OUTPUTCONF, $conf_path ) ) {
	while ( <OUTPUTCONF> ) {
	    next if /^\s*#/;
	    next if /^\s*$/;
	    chop;
	    push @expected, $_;
	}
	close OUTPUTCONF;
    }
    else {
	$err = "couldn't open $conf_path: $!";
    }
    
    return $err, @expected;
}

sub get_delcheck_confs
{
    my ( $curdir, $delcheck_prefix, $delcheck_affix ) = @_;

    my @delcheck_confs;

    if ( opendir( THISDIR, $curdir ) ) {
	foreach my $file ( readdir THISDIR ) {
	    if ( $file =~ /^$delcheck_prefix.*$delcheck_affix$/ ) {
		push @delcheck_confs, ( join '/', $curdir, $file );
	    }
	}
	closedir THISDIR;
    }

    return sort @delcheck_confs;
}

sub get_server_details
{
    my ( $curdir, $server_prefix, $server_affix ) = @_;

    my ( @server_names, @server_confs );

    if ( opendir( THISDIR, $curdir ) ) {
	foreach my $file ( readdir THISDIR ) {
	    if ( $file =~ /^$server_prefix\.(.*)\.$server_affix$/ ) {
		push @server_names, $1;
		push @server_confs, ( join '/', $curdir, $file );
	    }
	}
	closedir THISDIR;
    }

    return ( \@server_names, \@server_confs );
}


sub get_server_glue_details
{
    my ( $curdir, $server_prefix, $server_affix ) = @_;

    my ( @server_names, @server_confs );

    if ( opendir( THISDIR, $curdir ) ) {
	foreach my $file ( readdir THISDIR ) {
	    if ( $file =~ /^$server_prefix\.(.*)\.$server_affix$/ ) {
		push @server_names, $1;
    my @ips = get_glue_ips ($curdir, $1);
    push @server_names, @ips;
		push @server_confs, ( join '/', $curdir, $file );
	    }
	}
	closedir THISDIR;
    }

    return ( \@server_names, \@server_confs );
}

sub get_glue_ips 
{
  my $curdir = shift;
  my $name = shift;

  # cut out the port number
  $name =~ s/:.+$//i;

  my @ips;

  my $resolver_conf = join ('/', ($curdir, 'resolver.xml'));

  #print STDERR "starting get_glue_ips with $curdir, $name\n";

  open (FILE, "< $resolver_conf") || die "Cant open file $resolver_conf:$!";

  while (<FILE>) {
    # format: ns3.testzone.local. 3600 IN A 3.4.5.6
    if (/^\s*$name\.\s+\d+?\s+IN\s+A\s+(.+)\s*$/i) {
      push @ips, $1;
      next;
    }
    if (/^\s*$name\.\s+\d+?\s+IN\s+AAAA\s+(.+)\s*$/i) {
      push @ips, $1;
      next;
    }
  }
  close (FILE);
  return @ips;
}

sub perl_warning_handler
{
    my $warning = shift;

    print STDERR "$0: caught perl warning:\n";
    print STDERR $warning;
    print STDERR "Death! <gurgle>\n";
    exit -1;
}
