package Net::DelCheck;

# $Id: DelCheck.pm,v 2.11 2006/08/16 12:49:25 katie Exp $

# If you want to know how to use this module, better type
# 'perldoc Net::DelCheck' now. Textual comments are all
# program code help.

# See end for copyright.

#use 5.008;  # NOT CLEAR WHY...

use strict;
use warnings;

use Net::DNS::Resolver;
use AppConfig qw( :argcount );
use XML::Writer;
use Net::IP qw( &ip_reverse &ip_is_ipv6 &ip_is_ipv4 &ip_expand_address );
use IO::Scalar;
use Time::Local;
use XML::DOM;
use Socket;
use Digest::MD5 qw(md5_hex);

require Exporter;

our @ISA = qw( Exporter );

use vars qw( 
	     $REVISION
	     $VERSION 
	     );

our @EXPORT_OK = qw( 
		     &errors
		     
		     &test_format &normalise_domain_name 
		     &get_ip_version &is_reverse_zone
		     &seconds_to_descr_string
		     &list_problem_codes &problem_code_no_args
		     &ipv6_available
);

# Version of the library

$VERSION = '2.40';

$REVISION=(qw$Revision: 2.11 $)[1]; # 



# Version of the template instantiation code
# This must chang whenever the number of problem codes changes,
# or the number or order of arguments changes.
# See the perldoc for the template file for more info.

our $TEMPLATE_VERSION = '20060620';

# Test if we can use IPv6
# Waiting for Net::DNS to make a flag available for this, since
# we can't do anything with IPv6 unless the installed Net::DNS
# thinks it can as well.

our $IPV6_AVAILABLE;

BEGIN {


    my $NetDNSVersion = $Net::DNS::VERSION;
    $NetDNSVersion =~ s/_//;  


    
    $IPV6_AVAILABLE = 
	( 
	  $NetDNSVersion > 0.4900 &&       
          # eventually we may have IPv6
	  # actually its planned to be in
	  # version 0.50 (Olaf)
	  # Note that we also try to bind to ::1
	  eval { require Socket6; } &&
	  eval { require IO::Socket::INET6; } &&
	  $IO::Socket::INET6::VERSION >= 2.01 &&
	  IO::Socket::INET6->new(
				 Proto => 'tcp',
				 LocalAddr => 'localhost'
				 ) 
	  
	  
	  );
    
}
	       
##################### INTERNAL DATA STRUCTURE NOTES ###########################

# The module stores data under four key values of $self:

# 'config'
#
# reference to an AppConfig object, created by &init_config

# 'template'
#
# Reference to an XML::DOM::Document created by parsing the template XML file
# (configuration variable ModuleTemplatePath).

# 'report'
#
# reference to a XML::DOM::Document in which the report
# is built programmatically

# 'data'
#
# This is where collected data is stored, in the following structure:
#
# delegation
#    zone                       (normalized) name of the zone [&get_data]
#    bare_ns_list               bare_ns_list stripped from ModuleIgnorNameservers
#
# ns
#    byname                     data sorted by nameserver name
#        <name>
#            port               [&preprocess_ns_list]
#            rrs                [&get_nameserver_data]
#                a_list         Net::DNS::RR::A list
#                aaaa_list      Net::DNS::RR::AAAA list
#                cname_list     Net::DNS::RR::CNAME list
#                ptr_list       Net::DNS::RR::PTR list
#                glue_a_list    Net::DNS::RR::A list for own glue names @ own_ip
#                glue_aa_list   Net::DNS::RR::AAAA list own glue names @ own_ip
#            glue               list of glue ips given by the caller
#                
#    byip                       [get_nameserver_data]               
#        <ip>
#            name_list         list of all names for this IP
#            glue_name_list    list of names that are glues
# zone
#    <ip>                      [&get_zone_data_at_ip]
#      aa_flag                 Whether this IP returned AA for this zone SOA
#      rrs                     
#         ns_list              Net::DNS::RR::NS list
#         soa_list             Net::DNS::RR::SOA list
#         soa_rrsig_list       Net::DNS::RR::RRSIG list from ans section of SOA query
#         dnskey_list          Net::DNS::RR::DNSKEY list
#         dnskey_rrsig_list    Net::DNS::RR::RRSIG list from ans section of DNSKEY query
#         any_rrsig_list       Net::DNS::RR::RRSIG list from auth section of SOA query
#         other_list           Other Net::DNS::RR types   (not used
#                                          since ANY queries where obsolted)
#      mname_rrs               [&get_resolved_mnames]
#         a_list
#         aaaa_list
#
# Function names in [] indicate the function which creates the item

# ---------------------- CLASS VARIABLES ---------------------------- #

# Possibly occuring problem codes and their number of arguments.
# These are used to verify the template file.

my %PROBLEM_ARG_COUNT = (
		 'PROBLEM_LOOKUP' => 4, 
		 'PROBLEM_BAD_IP' => 2,
		 'PROBLEM_UNUSABLE_IP' => 3,
		 'PROBLEM_UNUSABLE_HOSTNAME' => 2,
		 'PROBLEM_BAD_PORT' => 2,
		 'PROBLEM_NO_REVERSE_MAPPING' => 2,
		 'PROBLEM_WRONG_REVERSE_MAPPING' => 2,
		 'PROBLEM_TOO_FEW_DELEGATED_NAMESERVERS' => 3,
		 'PROBLEM_DUPLICATE_DELEGATED_NS_IP' => 2,
		 'PROBLEM_DELEGATED_NS_ON_SAME_SUBNET' => 4,
		 'PROBLEM_SERIAL_NUMBERS_DIFFER' => 6,
		 'PROBLEM_SOA_VALUES_DIFFER' => 5,
		 'PROBLEM_INCONSISTENT_NS_LISTING' => 5,
		 'PROBLEM_REQUIRED_NAMESERVER_MISSING' => 4,
		 'PROBLEM_PROHIBITED_NAMESERVER_PRESENT' => 2,
		 'PROBLEM_DUPLICATE_DELEGATED_NS_NAME' => 2,
		 'PROBLEM_DUPLICATE_NS_NAME_LISTED_AT_ZONE' => 4,
		 'PROBLEM_DELEGATED_NS_NOT_LISTED_AT_ZONE' => 3,
		 'PROBLEM_LISTED_NAMESERVER_NOT_IN_DELEGATED_LIST' => 4,
		 'PROBLEM_SOA_RECORD_FIELD_EMPTY' => 1,
		 'PROBLEM_SERIAL_NOT_RECOMMENDED_FORMAT' => 1,
		 'PROBLEM_SERIAL_IN_FUTURE' => 1,
		 'PROBLEM_NO_SOA_AT_SERVER' => 2,
		 'PROBLEM_MULTIPLE_SOA_AT_SERVER' => 2,
		 'PROBLEM_NON_AA_NAMESERVER' => 2,
		 'PROBLEM_MNAME_NOT_LISTED_AS_NS' => 2,
		 'PROBLEM_MNAME_DOESNT_RESOLVE' => 1,
		 'PROBLEM_RNAME_DUPLICATE_ZONE' => 1,
		 'PROBLEM_REVERSE_DOMAIN_IN_RNAME' => 1,
		 'PROBLEM_RNAME_HAS_AT_SIGN' => 1,
		 'PROBLEM_RNAME_BAD_SYNTAX' => 1,
		 'PROBLEM_REFRESH_OUT_OF_RANGE' => 4,
		 'PROBLEM_RETRY_OUT_OF_RANGE' => 6,
		 'PROBLEM_EXPIRE_OUT_OF_RANGE' => 4,
		 'PROBLEM_EXPIRE_LESS_THAN_REFRESH' => 2,
		 'PROBLEM_EXPIRE_LESS_THAN_REFRESH_RETRY_SUM' => 3,
		 'PROBLEM_MINIMUM_OUT_OF_RANGE' => 4,
		 'PROBLEM_NS_NAME_NOT_CANONICAL' => 1,
		 'PROBLEM_DOMAIN_NAME_TOO_LONG' => 4,
		 'PROBLEM_DOMAIN_NAME_LABEL_EMPTY' => 2,
		 'PROBLEM_DOMAIN_NAME_LABEL_TOO_LONG' => 5,
		 'PROBLEM_HOSTNAME_IN_NUMERIC_TLD' => 2,
		 'PROBLEM_HOSTNAME_LABEL_ILLEGAL_CHARACTERS' => 3,
		 'PROBLEM_HOSTNAME_LABEL_ILLEGAL_START_OR_END' => 3,
		 'PROBLEM_HOSTNAME_ONE_LABEL' => 2,
		 'PROBLEM_COULDNT_LOOKUP_NS_ADDRESS' => 1,
		 'PROBLEM_SKIPPED_SERIAL_NUMBER' => 1,
		 'PROBLEM_MAILCHECK_CONNECTION_FAILED' => 3,
		 'PROBLEM_MAILCHECK_MX_LOOKUP' => 2,
		 'PROBLEM_MAILCHECK_A_LOOKUP' => 2,
		 'PROBLEM_MAILCHECK_SMTP_FAILURE' => 3,
		 'PROBLEM_RRSET_TTLS_NOT_EQUAL' => 4,
     'PROBLEM_NS_TTL_TOO_SHORT'  => 2,
     'PROBLEM_NOT_ALL_SERVERS_RETURN_DNSSECDATA' => 1,
     'PROBLEM_INCONSISTENT_DNSKEY_LISTING' => 5,
     'PROBLEM_NO_MATCHING_DNSKEY_FOR_DS' => 1,
     'PROBLEM_DS_POINTING_TO_NON_SEP_DNSKEY' => 3,
     'PROBLEM_SEP_FLAG_NOT_USED' =>0,
     'PROBLEM_SEP_KEY_NOT_SELF_SIGNED' =>3, 
     'PROBLEM_DS_DIGEST_ALGORITHM_NOT_KNOWN' => 1,
     'PROBLEM_DS_AND_DNSKEYRR_DO_NOT_MATCH' => 2,
     'PROBLEM_DNSSEC_ALGORITHM_NOT_IMPLEMENTED' => 3,
     'PROBLEM_CHAIN_OF_TRUST_DOES_NOT_VALIDATE' => 3,
     'PROBLEM_NO_CHAINS_OF_TRUST_VALIDATE' => 0,
     'PROBLEM_SIGNATURE_VALIDITY_NEARS_EXPIRATION' => 3,
     'PROBLEM_TTL_LARGE_FRACTION_OF_SIGNATURE_VALIDITY' => 2,
     'PROBLEM_GLUE_OUTSIDE_ZONE' => 1,
     'PROBLEM_GLUE_IP_MISSING' => 1,
     'PROBLEM_GLUE_IP_RR_MISSING' => 1,
);

############################# PUBLIC OBJECT METHODS ######################################

# Block for package error functions. This is just to avoid a global variable.
# Only &errors is exported.

{
    my @package_errors;

    # Function header: see perldoc

    sub errors { 
	if ( wantarray ) {
	    return @package_errors;
	    @package_errors = (); 
	}
	else {
	    return ( scalar ( @package_errors ) );
	}
    }

    sub add_error { push @package_errors, shift; }
    
    sub clear_errors {@package_errors = (); }
}

# Function header: see perldoc

sub new {
    my ( $that, @config_specs ) = @_;


    my $class = ref( $that ) || $that;

    my $self = {};

    bless ( $self, $class );

    # This is to make sure the error array is emptied between
    # two calls to the new method.
    $self->clear_errors();


    $self->init_config( @config_specs );
 
    if ( &errors ) {
	return undef;
    }
    else {
	return $self;
    }
}

# Function header: see perldoc

sub init_config
{
    my ( $self, @config_specs ) = @_;

    # For explanation of what each directive does, see perldoc


    $self->{'config'} = AppConfig->new( 
	{
	    'ERROR' => \&appconfig_error_handler,
	    'GLOBAL' => {
		'DEFAULT' => undef
	    }
	},

	'ModuleDiagnosticLevel'      =>    { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 0 },
	'ModuleDiagnosticFormat'     =>    { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 'txt' },
	'ModuleDiagnosticPacketHeaders' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'ModuleLocalResolvers'       =>    { ARGCOUNT => ARGCOUNT_LIST },
	'ModuleNameserverIgnoreList' =>    { ARGCOUNT => ARGCOUNT_LIST },
        'ModuleQueryRemoteDefaultPortNumber' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 53 },
	'ModuleQueryRemoteRetry'      =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 3 },
	'ModuleQueryRemoteRetrans'    =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 5 },
	'ModuleQueryRemoteUDPTimeout' =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 30 },
	'ModuleQueryRemoteNameserverV4' => { ARGCOUNT => ARGCOUNT_ONE },
	'ModuleQueryRemoteNameserverV6' => { ARGCOUNT => ARGCOUNT_ONE },
	'ModuleQueryLocalPortNumber'  =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 53 },
	'ModuleQueryLocalRetry'       =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 3 },
	'ModuleQueryLocalRetrans'     =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 5 },
	'ModuleQueryLocalUDPTimeout'  =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 20 },
	'ModuleReportXMLDeclaration' =>    { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => '<?xml version="1.0"?>' },
	'ModuleReportXMLTags'        =>    { ARGCOUNT => ARGCOUNT_HASH },
	'ModuleReportXMLAttributes'  =>    { ARGCOUNT => ARGCOUNT_HASH },	
	'ModuleSchemaTypes'          =>    { ARGCOUNT => ARGCOUNT_HASH },
	'ModuleSchemaNamespaceURI'   =>    { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 'http://www.w3.org/2001/XMLSchema' },
	'ModuleSchemaNamespaceName'  =>    { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 'xsd' },
	'ModuleSOAKeys'               =>   { ARGCOUNT => ARGCOUNT_LIST },
	'ModuleTemplatePath'          =>   { ARGCOUNT => ARGCOUNT_ONE },
	'ModuleTemplateXMLArgTag'     =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 'ARG' },
	'ModuleTemplateXMLArgNoAttr'  =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 'no' },
	'ModuleUseIPv6'                =>  { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 0 },
	'ModuleMailTimeoutSocket'     =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 5 },
	'ModuleMailTimeoutRead'     =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 10 },
	'ModuleMailEnvelopeFrom'       =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => undef },
	'ModuleMailCache'       =>   { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => undef },

	'CheckHostnamesAgainstDomainNameSyntax'   => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckHostnamesForIllegalCharacters' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckHostnamesForIllegalStartEndCharacters' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckHostnamesHaveMultipleLabels'            => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckNameserverNameCanonical'      => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckNameserverHasReverse'         => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckNameserverHasMatchingReverse' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckNameserverMinimumTTL'         => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckNumberOfNameserverNames'      => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckForDuplicateNameserverIPs'    => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckIPProximity'                  => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckForDifferingSerialNumbers'    => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckForDifferingSOAValues'        => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckForInconsistentNSListings'    => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckForDuplicateDelegatedNameserverNames' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckForDuplicateNameserverNamesAtZone'    => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckAllDelegatedNameserversListedAtZone'  => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckListedNameserversInDelegatedList'     => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckSerialAllNumeric'             => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckSerialRecommendedFormat'      => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckMultipleSOARecordsAtServer'   => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckNoSOARecordAtServer'          => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckSOARecordAtServerAuthoritatative' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckMnameInNSSetAtServer'         => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckMnameResolvable'              => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },	
        'CheckRnameNoDuplicateZone'       => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckRnameNotReverseZone'          => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckRnameHasNoAtSign'             => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckRnameAgainstDomainNameSyntax' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckRnameMiscSyntax'              => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckRnameMail'                  => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 0 },	
        'CheckRefreshRange'                 => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckRetryRange'                   => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckExpireRange'                  => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckExpireAgainstRefresh'         => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckExpireAgainstRefreshRetrySum' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
        'CheckMinimumRange'                 => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckForRequiredNameservers'       => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckForProhibitedNameservers'     => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'CheckSerialNotInFuture'            => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },	
	'CheckSkippedSerial'                => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },	
	'CheckRRsetTTL'                  => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },	
  'CheckNotAllServersReturnDNSSECData' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckInconsistentDNSKEYListing' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckNoMatchingDNSKEYForDS'     => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckDSPointingToNonSEPDNSKEY'  => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckSepFlagNotUsed'            => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckSepKeyNotSelfSigned'       => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckDSDigestAlgorithmNotKnown' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckDSAndDNSKEYRRDoNotMatch'   => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckDNSSECAlgorithmNotImplemented' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckChainOfTrustDoesNotValidate' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckNoChainsOfTrustValidate'    => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckSignatureValidityNearsExpiration' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckTtlLargeFractionOfSignatureValidity' => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckSOAMnameDiffer'                      => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 0 },
  'CheckGlueOutsideZone'                     => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckGlueIPMissing'                       => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
  'CheckGlueIPRRMissing'                     => { ARGCOUNT => ARGCOUNT_NONE, DEFAULT => 1 },
	'ParamMaxDomainNameLength'        => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 255 },
	'ParamMaxDomainNameLabelLength'   => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 63 },
	'ParamMinimumNumberNameservers'   => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 2 },
	'ParamFirstReasonableSerialYear'  => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 1980 },
	'ParamSerialFutureAllowanceHours' => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 24 },
	'ParamRefreshMin'                 => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 1200 },
	'ParamRefreshMax'                 => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 43200 },
	'ParamExpireMin'                  => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 1209600 },
	'ParamExpireMax'                  => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 2419200 },
	'ParamMinimumMin'                 => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 3600 },
	'ParamMinimumMax'                 => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 86400 },
	'ParamMinimumNameserverTTL'       => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 60 },
	'ParamRetryMinFractionRefresh'    => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 0.05 },
	'ParamRetryMaxFractionRefresh'    => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 1.0 },
	'ParamRequiredNameserverList'     => { ARGCOUNT => ARGCOUNT_LIST },
	'ParamRequiredNameserverSpec'     => { ARGCOUNT => ARGCOUNT_ONE, DEFAULT => 'exactly-one' },
	'ParamProhibitedNameserverList'   => { ARGCOUNT => ARGCOUNT_LIST }
   );

    # AppConfig doesn't seem to provide a nice way to default-value hashes and lists,
    # set them manually.

    if ( defined $self->{'config'} ) {
	
	$self->{'config'}->ModuleReportXMLTags( 'problem=PROBLEM' );
	$self->{'config'}->ModuleReportXMLTags( 'problem_entity=PROBLEM_ENTITY' );
	$self->{'config'}->ModuleReportXMLTags( 'report=DELCHECK_REPORT' );
	$self->{'config'}->ModuleReportXMLTags( 'no_problem=NO_PROBLEM' );

	$self->{'config'}->ModuleReportXMLAttributes( 'template_version=version' );
	$self->{'config'}->ModuleReportXMLAttributes( 'problem_entity_name=name' );
	$self->{'config'}->ModuleReportXMLAttributes( 'problem_code=code' );


	$self->{'config'}->ModuleSchemaTypes( 'problem_code=ProblemCodeDefinition' );
	$self->{'config'}->ModuleSchemaTypes( 'problem=ProblemDefinition' );
	$self->{'config'}->ModuleSchemaTypes( 'problem_entity=ProblemEntityDefinition' );
	$self->{'config'}->ModuleSchemaTypes( 'report=DelcheckReportDefinition' );

	$self->{'config'}->ModuleSOAKeys( 'mname' );
	$self->{'config'}->ModuleSOAKeys( 'serial' );
	$self->{'config'}->ModuleSOAKeys( 'rname' );
	$self->{'config'}->ModuleSOAKeys( 'refresh' );
	$self->{'config'}->ModuleSOAKeys( 'retry' );
	$self->{'config'}->ModuleSOAKeys( 'expire' );
	$self->{'config'}->ModuleSOAKeys( 'minimum' );

	if ( @config_specs ) {

	    &read_configs( $self->{'config'}, @config_specs );

	    $self->initialise_template
		if ( &errors == 0 );
	}
	else {
	    &test_config( $self->{'config'} );
	}




    }
    else  {
	&add_error( 'failed to create AppConfig object for config' );
    }
}

# Function header: see perldoc

sub get_data
{
    my $self = shift;

    # Check if zone and/or nameservers were specified

    my ( $zone, @nss ) = @_;

    # If we're starting a new check, delete all the old data and
    # problems

    delete $self->{'data'}      if ( exists $self->{'data'} );

    $self->initialise_report;

    $self->dprint( "Get data", 0 );
    $self->dprint();
    $self->dprint( "Zone: $zone", 0 );
    $self->dprint( "Nameservers: ". ( join ', ', @nss ). "\n", 0 );


    my @errors;

    # Check we were given a zone

    if ( $zone ) {
	$self->{'data'}{'delegation'}{'zone'} = &normalise_domain_name( $zone );
    }
    else {
	push @errors, 'no zone specified';
    }

    # If we were given some nameservers take a look at them

    my $arg_error;

    if ( @nss ) {
      my $arg_error = $self->preprocess_ns_list( @nss );
      push @errors, $arg_error if ($arg_error);
    }
    else {
	push @errors, 'no nameservers specified';
    }

    # reporting

    # If no errors....

    if ( @errors ) {
	$self->dprint( "argument errors, aborting", 0 );
    }
    else {
	# Diagnostics

	$self->dprint( "Delegated NS: ". &list_text( @{ $self->{'data'}{'delegation'}{'bare_ns_list'} } ), 0 );

	while ( my ( $ns_name, $ns_data ) =  each %{ $self->{'data'}{'ns'}{'byname'} } ) {
	    $self->dprint( "$ns_name: ". $ns_data->{'port'},  1 );
	    $self->dprint( "glue IP : ". join (', ', @{$ns_data->{'glue'}}) ,  1 ) if exists ($ns_data->{'glue'});
	}

  # Check possible glue records

  $self->check_glue_nameservers;

	# Get data about the nameservers

	$self->get_nameserver_data;

	# Get data from the nameservers about the zone

	$self->get_zone_data_from_nameservers;

	# Resolve mnames returned
	# No point if we're not going to check it later

	$self->get_resolved_mnames
	    if ( $self->check_enabled( 'CheckMnameResolvable' ) );
  }

    return @errors;
}

# Function header: see perldoc

sub check
{
    my $self = shift;
    my @ds = @_;

    $self->dprint();
    $self->dprint( 'Run all checks', 0 );

    # Do checks

    $self->check_ns_details();

    $self->check_zone_at_ips();

    $self->check_ns_set();

    $self->check_glue_rrs();

    $self->check_dnssec_data(@ds) if (scalar @ds);

    # If no problems occured, (as shown by there being no problem entities)
    # we need to insert a possible NO_PROBLEM element from the template file
    # into the report.

    my $problem_entity_nodes = $self->{'report'}->getDocumentElement->getElementsByTagName( 
            $self->conf( 'ModuleReportXMLTags', 'problem_entity' ), 0 
    );

    if ( $problem_entity_nodes->getLength == 0 ) {
	$self->insert_no_problem_element_into_report;
    }

    # Form the output, from the XML declaration plus the 'report' document

    my $output_scalar = 	
	join '', 
	$self->conf( 'ModuleReportXMLDeclaration' ), 
	$self->{'report'}->getDocumentElement->toString;

    return $output_scalar;
}

# Function header: see perldoc

sub xml_tag_name
{
    my $self = shift;

    my $tag_type = shift;

    return $self->{'config'}->ModuleReportXMLTags->{$tag_type};
}

# Function header: see perldoc

sub xml_attribute_name
{
    my $self = shift;

    my $attribute_type = shift;

    return $self->{'config'}->ModuleReportXMLAttributes->{$attribute_type};
}

# Function header: see perldoc

sub cleanup
{
    my $self = shift;

    if ( defined $self->{'template'} ) {
	$self->{'template'}->dispose;
    }

    if ( defined $self->{'report'} ) {
	$self->{'report'}->dispose;
    }
}

# Function header: see perldoc

sub unanimous_primary
{
    my $self = shift;

    $self->dprint( "&unanimous_primary", 0 );

    my %ns_names_by_mname;

    while (  my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {
	
	my $mname = $self->extract_usable_mname_from_soas( $ip );

	push @{ $ns_names_by_mname{$mname} }, @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} }
  	   if ( defined $mname );
    }


    my $mname_to_return;

    # Get the number of unique mnames listed

    my @all_mnames = keys %ns_names_by_mname;

    $self->dprint( "mnames found: ". &list_text( @all_mnames ), 1 );

    # There should only be one mname listed at all nameservers

    if ( @all_mnames == 1 ) {

	my @ns_with_mname = @{ $ns_names_by_mname{$all_mnames[0]}};

	$self->dprint( ( "listed at: ". &list_text( @ns_with_mname ) ), 1 );

	# It should be listed at all the delegated servers

	my $ok_flag = 1;

	foreach my $delegated_ns ( @{ $self->{'data'}{'delegation'}{'bare_ns_list'} } ) {
	    if ( ! grep { $delegated_ns eq $_ } @ns_with_mname ) {
		$self->dprint( "$delegated_ns does not list mname $all_mnames[0], can't return anything", 1 );
		$ok_flag = 0;
		last;
	    }
	}

	if ( $ok_flag ) {
	    $mname_to_return = $all_mnames[0];
	    $self->dprint( "found a unanimous mname: $mname_to_return", 1 );
	}
    }
    else {
	$self->dprint( "Multiple mnames found, return undef", 1 );
    }

    return $mname_to_return;
}

# Function header: see perldoc

sub schema
{
    my $self = shift;

    my $xml_schema;

    # Shorter names for programming convenience

    my $ns = $self->conf( 'ModuleSchemaNamespaceURI' );
    my $nsp = $self->conf( 'ModuleSchemaNamespaceName' );

    # XML::Writer only writes to file handles, need to fake
    # one to our output scalar

    # Some negative aspects of XML::Writer:
    # - for attribute values, can't specify a namespace
    # - setting a default prefix doesn't seem to work

    my $output_handle = new IO::Scalar \$xml_schema;

    my $writer = new XML::Writer(
			     DATA_MODE => 1,
			     DATA_INDENT => 3,
			     OUTPUT => $output_handle,
			     NAMESPACES => 1,
			     PREFIX_MAP => { $ns => $nsp }

    );

    $writer->addPrefix( $nsp );

    # Print XML declaration

    $writer->xmlDecl();

    $writer->startTag( [ $ns, 'schema' ] );
    
    # First generate definitions for all the data types

    $self->schema_write_problem_code_definition( $writer );
    
    $self->schema_write_problem_definition( $writer );

    $self->schema_write_problem_entity_definition( $writer );

    $self->schema_write_report_definition( $writer );

    # Then write a single element, which is of the report data type

    $writer->emptyTag( 
		       [ $ns, 'element' ], 
		       'name' => $self->conf( 'ModuleReportXMLTags', 'report' ),
		       'type' => $self->conf( 'ModuleSchemaTypes', 'report' ),
    );

    $writer->endTag( [ $ns, 'schema' ] );
    $writer->end();

    $output_handle->close();

    return $xml_schema;
}

############################## PUBLIC CLASS METHODS ######################################

# Function header: see perldoc

sub get_ip_version
{
    my $ip = shift;

    if ( &ip_is_ipv6( $ip ) ) {
	return 6;
    }
    elsif ( &ip_is_ipv4( $ip ) ) {
	return 4;
    }
    else {
	return 0;
    }
}

# Function header: see perldoc

sub list_problem_codes
{
    return sort keys %PROBLEM_ARG_COUNT;
}

sub ipv6_available
{
    return $IPV6_AVAILABLE;
}

# Function header: see perldoc

sub problem_code_no_args
{
    my $code = shift;

    return $PROBLEM_ARG_COUNT{$code} || undef;
}

# Function header: see perldoc

sub seconds_to_descr_string
{
    my $seconds_left = shift;

    return '0 seconds' if ( $seconds_left == 0 );

    # Things we will divide the seconds up into

    my @second_intervals = (
		'week', 604800,
		'day', 86400,
		'hour', 3600,
		'minute', 60,
                'second', 1
    );

    my @descr_strings;

    # We take a maximum of 3 descriptions because it looks silly saying e.g.
    # 5 years, 10 months, 6 days, 4 hours, 20 seconds

    while ( @descr_strings < 3 && ( my $interval_name = shift @second_intervals ) ) {

	my $interval_seconds = shift @second_intervals;

        # If there are more than $interval_seconds left to convert,
        # then there's at least one of unit $interval_name to be included
        # in the final string

	if ( $seconds_left >= $interval_seconds ) {

            # How many times does $interval_name fit into the seconds
            # remaining ?

	    my $no_intervals = int ( $seconds_left / $interval_seconds );

	    $seconds_left -= $no_intervals * $interval_seconds;

            # Make a string up for this $interval name

            my $descr_string = join ' ', $no_intervals, $interval_name;

            # Add an 's' if there are more than one of this $interval_name

            $descr_string .= 's' if ( $no_intervals > 1 );

            # Add it to the ones we have so far

            push @descr_strings, $descr_string;
	}
    }

    # If there were more than 0 seconds left, then we stopped before we'd
    # converted the whole lot, because being too accurate looks silly
    # sometimes. We note this by putting an 'approximately' sign in
    # front of the result

    my $approx_string = $seconds_left != 0 ? '~ ' : '';

    return join '', $approx_string, ( &list_text( @descr_strings ) );

}

# Function header: see perldoc

sub is_reverse_zone
{
    my $domainname = shift;

    # IPv4 reverse zone
    
    if ( $domainname =~ /(^|\.)in-addr\.arpa\.?/i ) {
	return 4;
    }	

    # IPv6 reverse zone
    
    elsif ( $domainname =~ /(^|\.)ip6\.(int|arpa)\.?/i ) {
	return 6;
    }

    # What the hell is it then ?

    else {
	return 0;
    }
}

# Function header: see perldoc

sub test_format
{
    my ( $to_check, $expected_format ) = @_;

    if ( $expected_format eq 'ip' ) {
	return 0
	    if ( ! &ip_is_ipv6( $to_check ) && ! &ip_is_ipv4( $to_check ) );
    }

    elsif ( $expected_format eq 'port' ) {
	return 0
	    if ( 
		 ( $to_check !~ /^([0-9]+)$/ ) || 
		 ( $to_check < 1 ) || 
		 ( $to_check >= 2**16 )
	    );

    }

    elsif ( $expected_format eq 'domain' || $expected_format eq 'hostname' ) {
	
	# Maximum length restriction
	
	return 0
	    if ( length $to_check > 255 );

        # No zero length labels please

	return 0
            if ( $to_check =~ /\.\./ );

	# Label length restriction

	my @labels = split /\./, $to_check;

	return 0
	    if ( grep { length $_ < 1 || length $_ > 63 } @labels );
    }

    # Lee 20020410

    # Removed the no-pure-numeric-labels restriction
    # See justification in &check_hostname

    if ( $expected_format eq 'hostname' ) {

	my @labels = split /\./, $to_check;

	return 0
	    if ( grep 
		{ 
		    $_ !~ /^[a-zA-Z0-9-]+$/  ||     # bad characters
 		    $_ =~ /^-/ ||                   # starts with '-'
		    $_ =~ /-$/                      # ends with '-'
		} @labels 
            ); 
    }

    return 1;
}

# Function header: see perldoc

sub normalise_domain_name
{
    my $dn = shift;

    return if ! $dn;

    $dn =~ s/^\s*//;      # Remove preceeding whitespace
    $dn =~ s/\s*$//;      # Remove trailing whitespace

    return '.' if ( $dn eq '.' );

    $dn = lc $dn;	  # Domain names are case insensitive
    $dn =~ s/\.$//;	  # Remove a terminating dot if present
    $dn =~ s/^\.//;	  # Remove a preceeding dot if present

    return $dn;
}

############################# PRIVATE OBJECT METHODS #####################################

# Purpose:		access configuration variables via function interface
# In:	$key:		configuration variable name
# In:	$subkey:	when the configuration variable is a hash, the $subkey
#                       element is returned
# Out:	$value:		value of the specified configuration variable
#
sub conf
{
    my ( $self, $key, $subkey ) = @_;
    # Check that it's a conf directive that we know about

    $self->program_error( "&conf: unknown configuration variable: $key", 1 )
	if ( ! $self->{'config'}->_exists( $key ) );

    # OK, let's get the value

    my $return_value = $self->{'config'}->$key;

    # If a 'subkey' is defined, we assume the config variable is a hash
    # In that case, $return_value is a hashref and we need to return 
    # the corresponding value from the hash

    if ( $subkey ) {
	if ( ref( $return_value ) ne 'HASH' ) {
	    $self->program_error( "&conf: configuration variable $key not a hash", 1 );	
	}

	$return_value = $self->{'config'}->$key->{$subkey};

	if ( ! defined $return_value ) {
	    &program_error( "&conf: requested HASH configuration variable ($key/$subkey) not defined" );
	}	
    }
    

    return $return_value;
}

# Purpose:		get data about the zone from each nameserver
#
sub get_zone_data_from_nameservers
{
    my $self = shift;

    my $zone = $self->{'data'}{'delegation'}{'zone'};

    $self->dprint();
    $self->dprint( "Get data about $zone from nameserver IP's", 0 );

    # Get a resolver configured for remote queries

    my $remote_res = $self->get_resolver( 'remote',1 );

    # Go through each IP we found for the nameservers, querying
    # about $zone at that IP

    while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'ns'}{'byip'} } ) {

	my $names_for_ip_text = &list_text( @{ $ip_ref->{'name_list'} } );

	$self->dprint();
	$self->dprint( "Get data about $zone from $ip ($names_for_ip_text)", 2 );
	
	# To be improved: uses the port no of the first one defined
	# in the case where the same NS is specified with same
	# IP but different ports

	my $port = $self->{'data'}{'ns'}{'byname'}{$ip_ref->{'name_list'}[0]}{'port'};

	$self->get_zone_data_at_ip( $remote_res, $ip, $names_for_ip_text, $port, $zone );
    }	
}

# Purpose:		for each SOA record we got back for the zone, try
#                       and resolve the mname to an IP
#

sub get_resolved_mnames

{
    my $self = shift;

    $self->dprint();

    $self->dprint( "Resolve returned mnames", 0 );


    my $res = $self->get_resolver( 'local',0 );	

    while (  my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

	my $mname = $self->extract_usable_mname_from_soas( $ip );

	my $names_for_ip_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );

	$self->dprint( "Resolve mname returned from $ip ($names_for_ip_text)", 1 );	

	if ( defined $mname ) {

	    # Check it's not a special hostname, stop if it is

	    next if ( ! $self->check_hostname_usable( $names_for_ip_text, $mname, 'the mname field of the SOA record' ) );
 
	    $self->dprint( "Resolve $mname", 2 );	
	    
	    # Get A records about name $mname
	    
	    my ( $answers, $err, undef, undef ) = $self->lookup_records( $res, $mname, 'A', 0 );
	    
	    if ( $err ) {
		$self->p( 'PROBLEM_LOOKUP', $names_for_ip_text, 
			      'A', $mname, &list_text( $res->nameservers ), $err
		);

		$self->dprint( "Error: $err", 3 );
	    }

	    # Record any A we find

	    foreach my $rr ( @$answers ) {
		if ( $rr->type eq 'A' ) {
		    push @{ $ip_ref->{'mname_rrs'}{'a_list'} }, $rr;
		}
	    }
	    
	    @$answers=(); $err=0;
	    ( $answers, $err, undef, undef ) = $self->lookup_records( $res, $mname, 'AAAA', 0 );
	    
	    if ( $err ) {
		$self->p( 'PROBLEM_LOOKUP', $names_for_ip_text, 
			  'A', $mname, &list_text( $res->nameservers ), $err
			);
		
		$self->dprint( "Error: $err", 3 );
	    }
	    
	    # Record  AAAA we find
	    
	    foreach my $rr ( @$answers ) {
		if ( $rr->type eq 'AAAA' ) {
		    push @{ $ip_ref->{'mname_rrs'}{'aaaa_list'} }, $rr;
		}
	    }

      # if this is a glue, add glue A/AAAA resolved earlier to the mname_rrs list.

      if ( $self->is_glue($mname)) {
         if ($self->{'data'}{'ns'}{'byname'}{$mname}{'rrs'}{'glue_a_list'}) {
           push @{ $ip_ref->{'mname_rrs'}{'a_list'} },
                @ {$self->{'data'}{'ns'}{'byname'}{$mname}{'rrs'}{'glue_a_list'} };
         }

         if ($self->{'data'}{'ns'}{'byname'}{$mname}{'rrs'}{'glue_aaaa_list'}) {
           push @{ $ip_ref->{'mname_rrs'}{'aaaa_list'} },
                @ {$self->{'data'}{'ns'}{'byname'}{$mname}{'rrs'}{'glue_aaaa_list'} };
         }

      } 

	}
	else {
	    $self->dprint( "no usable mname there", 2 );	
	}
    }
}

# Purpose: check whether a hostname is a subdomain of a zone

sub is_name_inside_zone
{
    my $self = shift;
    my $name = shift;

    my $zone = $self->{'data'}{'delegation'}{'zone'};
    if ($name =~ /^.+\.$zone$/i) {
      return 1;
    }
    return 0;
}

# Purpose: to check - is this legitimate glue record or not?

sub is_glue
{
   my $self = shift;
   my $name = shift;

   if ($self->is_name_inside_zone($name) &&
       (exists $self->{'data'}{'ns'}{'byname'}{$name}{'glue'} ) ) {
     return 1;
   }
   return 0;
}

# Purpose: get a list of glue names attached to an IP
#

sub get_glue_name_list
{
    my $self = shift;
    my $ip = shift;

    my @list;

    my $ip_ref = $self->{'data'}{'ns'}{'byip'}->{$ip};
    if ($ip_ref && exists $ip_ref->{'glue_name_list'}) {
      push @list, @{ $ip_ref->{'glue_name_list'} };
    }
    return @list;
}

# Purpose: split ipv4 and ipv6 glue records

sub get_glue_ip_list
{
    my $self = shift;
    my $name = shift;

    my (@ipv4_list, @ipv6_list);

    if ($self->is_name_inside_zone($name)) { # only take into account IP for acceptable glue
      if (exists $self->{'data'}{'ns'}{'byname'}{$name}{'glue'}) {
        foreach my $ip (@{$self->{'data'}{'ns'}{'byname'}{$name}{'glue'}}) {
          if (&ip_is_ipv4($ip)) {
            push @ipv4_list, $ip;
          }
          elsif (&ip_is_ipv6($ip)) {
            push @ipv6_list, $ip;
          }
          # safely ignore all the rest of the garbage which shouldn't be there anyway.
        }
      }
    }
    return (\@ipv4_list, \@ipv6_list);
}

# Purpose: do the glue-related checks
# we have to do them at this stage, because it may help to determine IP address for the server

sub check_glue_nameservers
{
    my $self = shift;
    
    $self->dprint();
    $self->dprint ("Check the possible glue records");

    my $zone = $self->{'data'}{'delegation'}{'zone'};

    foreach my $ns_name ( keys %{$self->{'data'}{'ns'}{'byname'} } ) {
      if ( $self->is_name_inside_zone($ns_name)) {
        if (exists $self->{'data'}{'ns'}{'byname'}{$ns_name}{'glue'}) {
          # legitimate glue, do the IP checks
          foreach my $ip (@{$self->{'data'}{'ns'}{'byname'}{$ns_name}{'glue'}}) {
            # try creating RR object
            if (&ip_is_ipv4($ip) ) {
              my $rr = new Net::DNS::RR("$zone. IN A $ip");
              my @ip_problems = $self->check_for_ip_problems($rr, 'Net::DNS::RR::A');
              foreach my $problem ( @ip_problems ) { $self->p( @$problem ) };
            }
            if (&ip_is_ipv6($ip) ) {
              my $rr = new Net::DNS::RR("$zone. IN AAAA $ip");
              my @ip_problems = $self->check_for_ip_problems($rr, 'Net::DNS::RR::AAAA');
              foreach my $problem ( @ip_problems ) { $self->p( @$problem ) };
            }
          }
        }
        else {
          # problem: glue is inside the zone but no IP is given
          $self->p('PROBLEM_GLUE_IP_MISSING', 'set', $ns_name) if ($self->check_enabled( 'CheckGlueIPMissing' ) );
        }
      }
      else { # outside the zone
        if (exists $self->{'data'}{'ns'}{'byname'}{$ns_name}{'glue'}) {
          # problem: glue is outside the zone
          $self->p('PROBLEM_GLUE_OUTSIDE_ZONE', 'set', $ns_name) if ($self->check_enabled( 'CheckGlueOutsideZone' ) );
        }  
      }
    }

}

# Purpose:		get data about a nameserver name
# Comments:             It's an A and AAAA query on the name, plus reverse
#                       lookup on a possible returned IP
#

sub get_nameserver_data
{
    my $self = shift;

    $self->dprint();
    $self->dprint( "Get data locally about nameservers", 0 );

    # Make a Resolver object to do the querying

    my $local_res = $self->get_resolver( 'local',0 );

    # We search only through *unique* names, to avoid querying
    # duplicate names twice

    foreach my $ns_name ( keys %{$self->{'data'}{'ns'}{'byname'} } ) {

	# Maybe it's localhost or something like that, then don't
	# do anything with it

	next if ( ! $self->check_hostname_usable( 'set', $ns_name, 'the delegated nameserver list' ) );

	$self->get_nameserver_name_data( $local_res, $ns_name );

	# We could get several IP's back for a particular name,
	# of different IP protocal versions. We need to decide
	# which one to use...

	my @problems = $self->pick_an_ip_for_ns( $ns_name );

	foreach my $problem ( @problems ) { $self->p( @$problem ) };

	my $specified_port = $self->{'data'}{'ns'}{'byname'}{$ns_name}{'port'};

	if ( my $ip = $self->{'data'}{'ns'}{'byname'}{$ns_name}{'ip'} ) {
	    
	    $self->dprint ( "Using IP: $ip", 3 );

	    $self->dprint( "Attempting to get PTR record", 2 );

	    $self->get_ptr_rr_data( $local_res, $ip, $ns_name );
	}
	else {
	    $self->dprint( "No IP found for $ns_name", 3 );
	}

	# Check port, we can't query later without it

	$self->dprint( "Check supplied port for $ns_name: $specified_port", 1 );
	
	if ( ! &test_format( $specified_port, 'port' ) ) {

	    $specified_port = '*undefined*' if ( ! defined  $specified_port );

	    $self->p( 'PROBLEM_BAD_PORT', $ns_name, 
		      $specified_port, $ns_name
	    );

	    $self->dprint( "Aborting: $ns_name: bad port specified: $specified_port", 2 );
	}
    }

    # Get a list of nameservers keyed by IP
    # Note an IP key may have several data values if some
    # of the delegated names mapped to the same IP

    while ( my ( $ns_name, $ns_ref ) = each %{ $self->{'data'}{'ns'}{'byname'} } ) {
	if ( defined $ns_ref && defined $ns_ref->{'ip'} ) {
	    push @{ $self->{'data'}{'ns'}{'byip'}{$ns_ref->{'ip'}}{'name_list'} }, $ns_name;
      if ($self->is_glue($ns_name)) {
        push @{ $self->{'data'}{'ns'}{'byip'}{$ns_ref->{'ip'}}{'glue_name_list'} }, $ns_name;
      }
	}
    }
}

# Purpose:		check a hostname is one we can use
# In:	$entity:	the associated entity
# In:   $hostname:      hostname to check
# In:   $location:      where the hostname was found
# Out:	$ok_falg:	boolean, true = it's usable
#
# Comments: 
#
# This is a check that it's not a special host of some kind,
# not that it's syntactically valid
#
sub check_hostname_usable
{
    my $self = shift;

    my ( $entity, $hostname, $location ) = @_;

    my $ok_flag = 1;

    if ( $hostname =~ /^localhost(\.|$)/ ) {
	$self->p( 
		  'PROBLEM_UNUSABLE_HOSTNAME', $entity, 
		  $hostname, $location
	);
	$ok_flag = 0;
    }elsif ( $hostname =~ /\.\d+$/ ) {
	$self->p( 
		  'PROBLEM_HOSTNAME_IN_NUMERIC_TLD', $entity, 
		  $hostname, $location
		  );
	$ok_flag = 0;
    }
	 

	 
    return $ok_flag;
}

# Purpose:		do an A and AAAA query on a nameserver name
# In:	$res:		reference to a Net::DNS::Resolver object
# Out:	$ns_name:	name to query for
# Comments: 
#
sub get_nameserver_name_data
{
    my ( $self, $res, $ns_name ) = @_;

    $self->dprint( "Get data about $ns_name", 1 );
    
    # Lee 20020410
    # Made error description clear, previously was prone to
    # interpretation as a DNS lookup error
    # Added an extra diagnostic too

    if ( ! &test_format( $ns_name, 'hostname' ) ) {
        $self->dprint( "not making query for bad hostname: $ns_name", 3 );
	return;
    }

    # Get A records about name $ns_name

    my ( $answers, $err, undef, undef ) = $self->lookup_records( $res, $ns_name, 'A', 0 );
    my $received_cname=0;

    if ( $err ) {
	$self->p( 'PROBLEM_LOOKUP', $ns_name, 
		  'A', $ns_name, &list_text( $res->nameservers ), $err
        );

	$self->dprint( "Error: $err", 3 );
    }

    foreach my $rr ( @$answers ) {
	if ( $rr->type eq 'A' ) {
	    push @{ $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'a_list'} }, $rr;
	}
	elsif ( $rr->type eq 'CNAME' ) {
	    push @{ $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'cname_list'} }, $rr;
	    $received_cname=1;
	}
    }


    return if $received_cname; # No need to query for AAAA if a CNAME was found.

    # Get AAAA records about name $ns_name
    @$answers=(); $err=0;

    ( $answers, $err, undef, undef ) = $self->lookup_records( $res, $ns_name, 'AAAA', 0 );
    
    if ( $err ) {
	$self->p( 'PROBLEM_LOOKUP', $ns_name, 
		  'AAAA', $ns_name, &list_text( $res->nameservers ), $err
		);
	
	$self->dprint( "Error: $err", 3 );
    }
    
    foreach my $rr ( @$answers ) {
	if ( $rr->type eq 'AAAA' ) {
	    push @{ $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'aaaa_list'} }, $rr;
	}
    }
    





}

# Purpose:		check that the contents of the template file meet our
#                       minimum requirements
#
sub test_template
{
    my $self = shift;

    my $root_element = $self->{'template'}->getDocumentElement;

    if ( ! defined $root_element ) {
	&add_template_error( $self->conf( 'ModuleTemplatePath' ), 'no root element' );
	return;
    }

    my $template_version_attribute = $self->conf( 'ModuleReportXMLAttributes', 'template_version' );

    # Test template version is the same as the current one

    my $version_node = $root_element->getAttributeNode( $template_version_attribute );

    if ( ! defined $version_node ) {
	&add_template_error( $self->conf( 'ModuleTemplatePath' ), 'no "version" attribute in root node' );
	return;
    }

    my $template_version = $version_node->getValue;

    if ( $template_version ne $TEMPLATE_VERSION ) {
	&add_template_error( $self->conf( 'ModuleTemplatePath' ), "template version mismatch. The template version is ". $template_version. "while the libarary is configured to use ".$TEMPLATE_VERSION .".Please consult the Changes file, adjust your template, then set the version number to $TEMPLATE_VERSION." );
	return;
    }

    # Check problem codes occur exactly one time, at top level

    my $problem_tag = $self->conf( 'ModuleReportXMLTags', 'problem' );    

    # Record which problem codes we've seen already and how many times

    my %seen_codes;

    # Problem nodes are assumed to be direct children of the root
    # (0=no recurse)

    my $problem_nodes = $root_element->getElementsByTagName( $problem_tag, 0 );

    for ( my $index = 0; $index < $problem_nodes->getLength; $index++ ) {

	my $problem_node = $problem_nodes->item( $index );

	my $problem_code_attribute = $self->conf( 'ModuleReportXMLAttributes', 'problem_code' );

	# Get node representing the problem code attribute

	my $code_node = $problem_node->getAttributeNode( $problem_code_attribute );

	# Examine the code if we found one, otherwise moan about it :-)

	if ( defined $code_node ) {
	    my $template_code = $code_node->getValue;

	    # If this code is defined, record it's occurence and test it's
	    # possible child ARG tags, otherwise moan about it

	    if ( exists $PROBLEM_ARG_COUNT{$template_code} ) {

		$seen_codes{$code_node->getValue}++;

		$self->test_argument_tags( $template_code, $problem_node );

	    }
	    else {
		&add_template_error( $self->conf( 'ModuleTemplatePath' ), "unknown problem code $template_code" );	
	    }
	}
	else {
	    &add_template_error( $self->conf( 'ModuleTemplatePath' ), "found a <$problem_tag> tag without '$problem_code_attribute' attribute" );		
	    next;
	}
    }

    # Check that each defined problem code occured at least once, and
    # that no problem code occured more than once

    foreach my $defined_problem_code ( keys %PROBLEM_ARG_COUNT ) {
	my $occurences = $seen_codes{$defined_problem_code};

	if ( ! defined $occurences ) {
	    &add_template_error( $self->conf( 'ModuleTemplatePath' ), "no <$problem_tag> tag for $defined_problem_code" );
	}
	elsif ( $occurences > 1 ) {
	    &add_template_error( $self->conf( 'ModuleTemplatePath' ), "problem code $defined_problem_code occurs $occurences times" );
	}
    }

    # Finally, test there aren't too many 'NO_PROBLEM' tags
    # Having one is optional.

    my $no_problem_tag = $self->conf( 'ModuleReportXMLTags', 'no_problem' );

    my $no_problem_nodes = $root_element->getElementsByTagName( $no_problem_tag, 0 );

    if ( $no_problem_nodes->getLength > 1 ) {
	&add_template_error( $self->conf( 'ModuleTemplatePath' ), "more than one $no_problem_tag element children of root" );
    }

}

# Purpose:		test that all the ARG nodes in a problem node meet our
#                       requirements 
# In: $problem_code:    code that the ARG node is contained by
# In: $problem_node:    node to check
#
sub test_argument_tags
{
    my ( $self, $problem_code, $problem_node ) = @_;

    # Check correct number of arguments

    my $argument_tag = $self->conf( 'ModuleTemplateXMLArgTag' );

    my $expected_num_arguments = $PROBLEM_ARG_COUNT{$problem_code};
		
    # Recursively get argument tags

    my $argument_nodes = $problem_node->getElementsByTagName( $argument_tag, 1 );

    for ( my $index = 0; $index < $argument_nodes->getLength; $index++ ) {

	my $argument_node = $argument_nodes->item( $index );

	my $argument_num_attribute = $self->conf( 'ModuleTemplateXMLArgNoAttr' );

	my $num_node = $argument_node->getAttributeNode( $argument_num_attribute );

	# If there's a number attribute, check it's argument

	if ( defined $num_node ) {
	    
	    my $argument_number = $num_node->getValue;
	    # Argument number must be numeric


	    my $oldwarn=$SIG{'__WARN__'};
	    $SIG{'__WARN__'} =sub{return;};

	    if ( $argument_number !~ /^\d+$/ ) {
		$SIG{'__WARN__'} = $oldwarn;
		&add_template_error( $self->conf( 'ModuleTemplatePath' ), "problem tag $problem_code contains a non-numeric argument number ($argument_number)" );	
	    }
	    
	    # Argument number starts at 0
	    
	    elsif ( $argument_number == 0 ) {
		$SIG{'__WARN__'} = $oldwarn;
		&add_template_error( $self->conf( 'ModuleTemplatePath' ), "problem tag $problem_code contains a reference to argument 0 (arguments start at 1)" );	
	    }
	    

	    # Argument number must be not more than the number of
	    # arguments that we know the problem code supports

	    elsif ( $argument_number > $expected_num_arguments ) {
		$SIG{'__WARN__'} = $oldwarn;
		&add_template_error( $self->conf( 'ModuleTemplatePath' ), "problem tag $problem_code contains a reference to argument $argument_number, this code only has $expected_num_arguments argument(s)" );	
	    }

	    # Extra... not really needed
	    $SIG{'__WARN__'} = $oldwarn;
	}
	else {
	    &add_template_error( $self->conf( 'ModuleTemplatePath' ), "no '$argument_num_attribute' in problem tag $problem_code" );	
	}	
    }
}

# Purpose:			record the details of a problem which has been spotted
#				by one of the checking methods
# In:	$problem_code:		the code (problem identifier)
# In:	$problem_entity:	the things which the problem relates to e.g. nameserver
#				name, set of nameserver names
# In:	@problems_arguments	the specific details involved in this occurence of
#				the problem. These are instantiated into a problem
#                               template from ModuleReportTemplate and added to
#                               $self->{'report'}.
#
sub p
{
    my ( $self, $problem_code, $problem_entity, @problem_arguments ) = @_;

    

    
    $self->program_error(  "Code called &p with undefined argument (\$problem_code)", 1 ) unless defined($problem_code);
    $self->program_error(  "Code called &p with undefined argument (\$problem_entity)", 1 ) unless defined($problem_entity);


    $self->dprint( "$problem_code for entity $problem_entity" );
    $self->dprint( "Args: ". ( join ', ', @problem_arguments ) );


    my $no_arguments = scalar( @problem_arguments );

    # Sanity check: problem code supplied must be defined

    if ( ! defined $PROBLEM_ARG_COUNT{$problem_code} ) {
	$self->program_error(  "Code called &p with problem code ($problem_code) which is undefined in %PROBLEM_ARG_COUNT", 1 );
    }
    
    # Sanity check: number of arguments supplied by caller must
    # equal defined number of arguments for the code.

    if ( $PROBLEM_ARG_COUNT{$problem_code} != @problem_arguments ) {
	$self->program_error(  "Code called &p with bad number of arguments ($no_arguments) for problem code $problem_code", 1 );
    }

    # Find the appropriate node in the template file for $problem_code

    my $nodes = $self->{'template'}->getDocumentElement->getElementsByTagName ( $self->conf( 'ModuleReportXMLTags', 'problem' ), 0 );

    my $problem_template;
    for ( my $index = 0; $index < $nodes->getLength; $index++ ) {
	if ( $nodes->item( $index )->getAttributeNode ( $self->conf( 'ModuleReportXMLAttributes', 'problem_code' ) )->getValue eq $problem_code ) {
	    $problem_template = $nodes->item( $index );	    
	    last;
	}
    }

    # Copy the template node and instantiate arguments
    # 1 = deep copy

    my $instantiated_node = $problem_template->cloneNode( 1 ); 

    $instantiated_node->setOwnerDocument( $self->{'report'} );

    &instantiate_problem_arguments( $self, $instantiated_node, @problem_arguments );

    # Append problem at appropriate place in report

    my $doc = $self->{'report'};
    my $root_element = $doc->getDocumentElement;

    # First find the entity node under which this problem belongs

    my $entity_node;

    foreach my $child_node ( $root_element->getChildNodes ) {
	if ( $child_node->getAttributeNode( "name" )->getValue eq $problem_entity ) {
	    $entity_node = $child_node;
	}
    }

    # Make an entity node if there wasn't one for this problem entity
    # already

    if ( ! defined $entity_node ) {
	$entity_node = $doc->createElement( $self->conf( 'ModuleReportXMLTags', 'problem_entity') );

	$entity_node->setAttribute( $self->conf( 'ModuleReportXMLAttributes', 'problem_entity_name' ), $problem_entity );

	$root_element->appendChild( $entity_node );
    }

    # Append the instantiated problem node to the entity we found/made

    $entity_node->appendChild( $instantiated_node );

    $self->{'data'}{'state'}{'problem_count'}++;
}

# Purpose:    get ready to produce a new report
#
sub initialise_report
{
    my $self = shift;

    # Dispose of old report if it exists

    $self->{'report'}->dispose
	if ( defined $self->{'report'} );

    # Make a new report: a DOM document

    $self->{'report'} = new XML::DOM::Document;

    # Create root element and add to document

    my $root = $self->{'report'}->createElement( $self->conf( 'ModuleReportXMLTags', 'report' ) );

    $self->{'report'}->appendChild( $root );
}

# Purpose: fairly self-descriptive I'd say
# 
sub insert_no_problem_element_into_report
{
    my $self = shift;

    # Get the NO_PROBLEM elements which are direct children of
    # the root in the template document

    my $no_problem_nodes = $self->{'template'}->getDocumentElement->getElementsByTagName( $self->conf( 'ModuleReportXMLTags', 'no_problem' ), 0 );

    # If there is one NO_PROBLEM tag, append it's contents into the report right
    # under the root. Method is to clone a new one from the template, then
    # append it to the report.

    if ( $no_problem_nodes->getLength == 1 ) {

	my $copied_node = $no_problem_nodes->item( 0 )->cloneNode( 1 );

	# Need to make the owner the report, otherwise we'll get an
	# exception when trying to insert it into the report

	$copied_node->setOwnerDocument( $self->{'report'} );

	$self->{'report'}->getDocumentElement->appendChild( $copied_node );
    }
}

# Purpose:   fill in the <ARG> elements found in a DOM node with the
#            elements of the supplied list
# In: $node: DOM node containing (possibly) ARG elements
# In: @args: items to replace ARG elements with
#
# Comments: 

# this function is recursive, so that ALL ARG elements within the passed
# node will be replaced.
#
# for nested ARG tags, only the first will be honoured, the
# rest will be removed

# the ARG elements have 'num' attributes indicating
# which item from the list they should be replaced with. Note
# that the 'num' attribute is 1-based, whereas the list index
# is 0-based.
#
sub instantiate_problem_arguments
{
    my ( $self, $node, @args ) = @_;
    
    # Get all the children of this node

    for my $child_node ( $node->getChildNodes ) {

	# We're only interested in elements

	if ( $child_node->getNodeType == XML::DOM::ELEMENT_NODE ) {	

	    # With a tag name matching the one from the conf (default: ARG)

	    if ( $child_node->getTagName eq $self->conf( 'ModuleTemplateXMLArgTag' ) ) {

		# Get the argument number from the template
		# Note: &test_template ensures that it exists

		my $template_arg_no = $child_node->getAttributeNode( $self->conf( 'ModuleTemplateXMLArgNoAttr' ) )->getValue;

		# Create a text element to replace the existing ARG element
		# Note: conversion from 1-based to 0-based.

		my $argument = ( defined $args[$template_arg_no - 1] ) ? $args[$template_arg_no - 1] : 'UNDEFINED';
		
		# Olaf Oct 2004 
		# The SMTP test alows a way to get < and > into the argument
		# We need to escape these.


		$argument=~s/<//g;
		$argument=~s/>//g;

		my $text_elem = $self->{'report'}->createTextNode( $argument );

		# Do the replacement

		eval {
		    $node->replaceChild( $text_elem, $child_node );	
	        };	

		if ( $@ ) {
		    $self->program_error( "failed to replace child node to instantiate argument: $@" );
		}
	    }

	    # Check all elements except ARG ones for ARG children

	    else {
		&instantiate_problem_arguments( $self, $child_node, @args );
	    }	
	}
    }
}

# Purpose:     create a definition of a complex data type representing a report
# In: $writer: ref to an XML::Writer object in which we are creating the schema
#
sub schema_write_report_definition
{
    my ( $self, $writer ) = @_;

    # Shorthand vars for the namespace URI/name

    my $ns = $self->conf( 'ModuleSchemaNamespaceURI' );
    my $nsp = $self->conf( 'ModuleSchemaNamespaceName' );

    $writer->startTag( 
		       [ $ns, 'complexType' ], 
		       name =>  $self->conf( 'ModuleSchemaTypes', 'report' )
    );

    # Version number for the template must be included

    $writer->emptyTag( [ $ns, 'attribute'] , 
		       name => $self->conf( 'ModuleReportXMLAttributes', 'template_version' ), 
		       type => "$nsp:positiveInteger", 
		       use => 'required'
    );

    # A report is a sequence of problem entities

    $writer->startTag( [ $ns, 'sequence' ] );
    $writer->startTag( 
		       [ $ns, 'element' ], 
		       name =>  $self->conf( 'ModuleReportXMLTags', 'problem_entity' ), 
		       type => $self->conf( 'ModuleSchemaTypes', 'problem_entity' ), 
		       minOccurs => 1, 
		       maxOccurs => 'unbounded'  
    );     
    $writer->endTag( [ $ns, 'element' ] );
    $writer->endTag( [ $ns, 'sequence' ] );
    $writer->endTag( [ $ns, 'complexType' ] );
}

# Purpose:     create a definition of a complex data type representing a
#              'problem entity'
# In: $writer: ref to an XML::Writer object in which we are creating the schema
#
sub schema_write_problem_entity_definition
{
    my ( $self, $writer ) = @_;

    # Shorthand vars for the namespace URI/name

    my $ns = $self->conf( 'ModuleSchemaNamespaceURI' );
    my $nsp = $self->conf( 'ModuleSchemaNamespaceName' );

    $writer->startTag( 
		       [ $ns, 'complexType' ], 
		       name =>  $self->conf( 'ModuleSchemaTypes', 'problem_entity' ) 
    );

    # An entity's attribute: it's name

    $writer->emptyTag( [ $ns, 'attribute'] , 
		       name => $self->conf( 'ModuleReportXMLAttributes', 'problem_entity_name' ), 
		       type => "$nsp:string", 
		       use => 'required'
    );

    # A problem entity is a sequence of problem

    $writer->startTag( [ $ns, 'sequence' ] );
    $writer->startTag( 
		       [ $ns, 'element' ], 
		       name => $self->conf( 'ModuleReportXMLTags', 'problem' ),
		       type => $self->conf( 'ModuleSchemaTypes', 'problem' ),
		       minOccurs => 1, 
		       maxOccurs => 'unbounded'  
    );
    $writer->endTag( [ $ns, 'element' ] );
    $writer->endTag( [ $ns, 'sequence' ] );

    $writer->endTag( [ $ns, 'complexType' ] );
}

# Purpose:     create a definition of a simple data type for representing a
#              problem code
# In: $writer: ref to an XML::Writer object in which we are creating the schema
#
sub schema_write_problem_code_definition 
{
    my ( $self, $writer ) = @_;

    # Shorthand vars for the namespace URI/name
	
    my $ns = $self->conf( 'ModuleSchemaNamespaceURI' );
    my $nsp = $self->conf( 'ModuleSchemaNamespaceName' );

    # Problem code is simple type which is an enumeration of
    # all the possible problem codes

    $writer->startTag( 
		       [ $ns, 'simpleType' ], 
		       name => $self->conf( 'ModuleSchemaTypes', 'problem_code' )
    );

    $writer->startTag( 
		       [ $ns, 'restriction' ], 
		       base => "$nsp:string"
    );

    foreach my $problem_code ( keys %PROBLEM_ARG_COUNT ) {
	$writer->emptyTag( 	
		       [ $ns, 'enumeration' ], 
		       value => $problem_code
	);
    }

    $writer->endTag( [ $ns, 'restriction' ] );

    $writer->endTag( [ $ns, 'simpleType' ] );
}

# Purpose:     create a definition of a complex data type representing a problem
# In: $writer: ref to an XML::Writer object in which we are creating the schema
#
sub schema_write_problem_definition 
{
    my ( $self, $writer ) = @_;
	
    my $ns = $self->conf( 'ModuleSchemaNamespaceURI' );
    my $nsp = $self->conf( 'ModuleSchemaNamespaceName' );

    $writer->startTag( 
		       [ $ns, 'complexType' ], 
		       name => $self->conf( 'ModuleSchemaTypes', 'problem' )
    );

    # A problem definition is an attribute code, and can be anything underneath
    # because that part comes from the template file and is user-defined

    # Here's the attribute

    $writer->emptyTag( [ $ns, 'attribute'] , 
		       name => $self->conf( 'ModuleReportXMLAttributes', 'problem_code' ),
		       type => $self->conf( 'ModuleSchemaTypes', 'problem_code' ), 
		       use => 'required'
    );

    # Here's the 'anything goes'

    $writer->emptyTag( [ $ns, 'element'],
		       type => "$nsp:anyType"
    );

    $writer->endTag( [ $ns, 'complexType' ] );
}

# Purpose:				print a diagnostic message
# In:	$text:				message to print
# In:   $header_level (optional):	the strength with which this message should be
#					emphasised. Max is 5, most emphasis is 1
#
sub dprint
{
    my ( $self, $text, $header_level ) = @_;
    return if (
	       ( $self->conf( 'ModuleDiagnosticLevel' ) == 0 ) ||
	       ( 
		  ( defined $header_level ) &&
		  ( $header_level > $self->conf( 'ModuleDiagnosticLevel') )
               )
    );

    $header_level = 5
	if ( ! defined $header_level );

    if ( $self->conf( 'ModuleDiagnosticFormat' ) eq 'txt' ) {
	
	if ( ! defined $text ) {
	    print "\n";
	}
	else {
	    print " " x ( $header_level * 4 )
		if ( defined $header_level );
	    
	    if ( $text =~ /^PROBLEM_/ ) {
		print "$text\n";
	    } else {
		print '>', $text, "\n";
	    }
	}
    }

    else {
	if ( ! defined $text ) {
	    print '<P>';
	}
	else {
	    print "&nbsp;" x ( $header_level * 4 )
		if ( defined $header_level );
	    
	    if ( $text =~ /^PROBLEM_/ ) {
		print "<B>$text</B><BR>";
	    } else {
		print '>', $text, '<BR>';
	    }
	}
    }
}

sub ip_exists
{
  my $ip = shift;
  my $list = shift;
  my %seen;

  %seen = map { $_ => 1 } @$list;
 
  if (exists $seen{$ip}) {
    return 1;
  }
  return 0;
}

# Purpose:			see function name
# In:  @toprocess_ns_list:	list of nameservers to process
# Comments:			an element in @toprocess_ns_list looks like
#				"ns-name[=port]" i.e. has an optional port number
#				attached. The port is split off, the name is
#				standardized, and the duplicates are removed.
#
sub preprocess_ns_list
{
    my $self = shift;

    my @toprocess_ns_list = @_;

    my $last_name;

    my $error;

    $self->dprint( "Split of NS and ports, preprocess NS names", 0 );

    foreach my $ns ( @toprocess_ns_list ) {

      if (&ip_is_ipv6( $ns ) || &ip_is_ipv4( $ns )) {
        # assign IP to the glue hash
        if (!$last_name) { # argument error: IP without a name
          # generate error here
          $error = "nameserver name must precede an IP \'$ns\'";
        }
        else {
          my $ip = lc(&ip_expand_address($ns, &ip_is_ipv4( $ns ) ? 4 : 6));
          push @{$self->{'data'}{'ns'}{'byname'}{$last_name}{'glue'}} , 
            $ip unless (ip_exists ($ip, $self->{'data'}{'ns'}{'byname'}{$last_name}{'glue'}) == 1); 
            # avoid duplicate IPs for glue records
        }
      }
      else {

        my ( $name, $port ) = &split_ns_and_port( $ns, $self->conf( 'ModuleQueryRemoteDefaultPortNumber' ) );

        $self->dprint( "$ns -> $name, $port", 1 );

        # Standardize name

        my $tidy_name = &normalise_domain_name( $name );

        # save for the list parsing

        $last_name = $tidy_name;

        # Include names which are not in the ignore list

        my $ignore_list = $self->conf( 'ModuleNameserverIgnoreList' );

        push @{ $self->{'data'}{'delegation'}{'bare_ns_list'} }, $tidy_name;

        if ( ( grep { &normalise_domain_name( $_ ) eq $tidy_name } @$ignore_list ) ) {
          $self->dprint( "Ignoring nameserver $tidy_name, it's in ModuleNameserverIgnoreList" );
          next;
        }

        # Only $tidy_name s that are not in the ModuleNameserverIgnoreList 
        # Are defined in the 	$self->{'data'}{'ns'}{'byname'} hash.
        $self->{'data'}{'ns'}{'byname'}{$tidy_name}{'port'} = $port;
      }  

    }

    # remove duplicates from bare_ns_list which may be here if we support glue.
    if ($self->check_enabled('CheckGlueIPMissing')) {
      my %unique;
      foreach my $ns (@{ $self->{'data'}{'delegation'}{'bare_ns_list'} }) {
        $unique{$ns} ++;
      }
      @{ $self->{'data'}{'delegation'}{'bare_ns_list'} } = (keys %unique);
    }
    
    return ($error);
}

# Purpose:		make a new resolver object with default settings
#			depending on how it's going to be used
# In:	$type:		'local' or 'remote', defaults to 'local'
# Out:	$res:		a Net::DNS::Resolver object
#
sub get_resolver
{
    my ( $self, $type, $dnssec ) = @_;

    my $res = new Net::DNS::Resolver;

    if ( ! defined $res ) {
    	$self->program_error( "Couldn't create Net::DNS::Resolver object: $?", 1 );
    }

    $res->defnames( 0 );			# no appending of default domain

    # If a nameserver was specified, use the remote nameserver parameters

    if ( $type && $type eq 'remote' ) {
	$res->port( $self->conf( 'ModuleQueryRemoteDefaultPortNumber' ) );
	$res->retry( $self->conf( 'ModuleQueryRemoteRetry' ) );
	$res->retrans( $self->conf( 'ModuleQueryRemoteRetrans' ) );
	$res->udp_timeout( $self->conf( 'ModuleQueryRemoteUDPTimeout' ) );
	$res->recurse( 0 );                     # we don't want NS to recurse if no data
        if ($dnssec == 1) {
          $res->dnssec(1);
        }
    }

    else {

	my $local_nameservers =  $self->conf( 'ModuleLocalResolvers' );

	if ( defined $local_nameservers ) {
	    $res->nameservers( @$local_nameservers );
	}

	$res->port( $self->conf( 'ModuleQueryLocalPortNumber' ) );
	$res->retry( $self->conf( 'ModuleQueryLocalRetry' ) );
	$res->retrans( $self->conf( 'ModuleQueryLocalRetrans' ) );
	$res->udp_timeout( $self->conf( 'ModuleQueryLocalUDPTimeout' ) );
	$res->recurse( 1 );		
    }

    if ($dnssec == -1) {
      $self->dprint("Dnssec skipped");
      $res->dnssec(0);
      $res->udppacketsize(512);
    }
    return $res;
}


# Purpose:		provide a simple interface to Net::DNS::Resolver
# In:	$res		a Net::DNS::Resolver object reference
# In:	$name		the name to query for
# In:   $type		the record type to query for
# In:   $type_only_flag whether to ignore records of another type that come up
# Out:	\@answers	reference to array of references to Net::DNS::RR
#                       in the answer section of the response
# Out:  $error_string	any lookup problems which occured
# Out:  $aa_flag	content of the AA flag in the response header
#			undef, 0, or 1
# Out:  $rcode          The RCODE from the query or emptystring if a loopup problem occured
# Out:  $auth           reference to array of references to Net::DNS::RR
#                       in the authority section of the response
#
sub lookup_records 
{
    my ( $self, $res, $name, $type, $type_only_flag ) = @_;

    my ( @answers, $error_string, $aa_flag, @auths );

    # Check name is not a silly one

    return ( \@answers, "aborting lookup: bad name: $name", undef )
	if ( ! &test_format( $name, 'domain' ) );

    my $packet = new Net::DNS::Packet( $name, $type, 'IN' );

    my $expected_answer_id = $packet->header->id;

    $self->dprint( 
	    "Query: $name, $type, DO=".$res->dnssec.", @(".
	    &list_text( $res->nameservers ).
	    ") port:".
	    $res->port,
	    0 
    );


    $self->print_packet_header( $packet->header, 1 );

    my $query = $res->send( $packet );
    
    if ( $query ) {

	$self->print_packet_header( $query->header, 0, $expected_answer_id );
	
	$aa_flag = $query->header->aa;
	
	if ( $query->header->ancount > 0 ) {

      foreach my $rr ( $query->authority ) {
        push @auths, $rr;
      }
	    
	    foreach my $rr ( $query->answer ) {
		
		my $record_type = $rr->type;
		    
		# Add record if the flag says we don't care about the type,
		# or if the type matches the one requested
		
		if ( ! $type_only_flag || $record_type =~ /^\s*$type\s*$/i ) {
		    push @answers, $rr;
		}
		else {
		    $self->dprint( "Ignoring record type '$record_type'", 0 );
		}
	    }
	}
    }
    else {
	$error_string = $res->errorstring;
    }

    foreach my $rr ( @answers ) {
	$self->print_rr( $rr );
    }
    
    
    my $rcode="";
    
    $rcode=$query->header->rcode unless $error_string;

    return ( \@answers, $error_string, $aa_flag, $rcode, \@auths );
}

# Purpose:		query for data about a zone at an IP address
# In:	$res:		reference to a Net::DNS::Resolver object
# In:	$ip:		IP to query at
# In:   $names_for_ip_text  Comma-separated names of NS which had this IP address
# In:	$port:		port to query at
# In:   $zone:		zone to query about
#
sub get_zone_data_at_ip
{
    my $self = shift;

    my ( $res, $ip, $names_for_ip_text, $port, $zone ) = @_;
    
    $self->dprint( "Query for $zone at $ip:$port", 3 );

    if ( &ip_is_ipv6( $ip ) && ! $self->conf( 'ModuleUseIPv6' ) ) {
	$self->dprint( "Aborting, IPv6 queries disabled.", 4 );
	return;
    }

    # Set IP to query at

    my $query_ip = $self->get_remote_query_ip( $ip );

    $res->nameservers( $query_ip );

    if ( &test_format( $port, 'port' ) ) {
	$res->port( $port );
    }
    else {
	$self->dprint( "Aborting, bad port: $port", 2 );
	
	return;
    }

    my ( $answers, $err, $auths, $rcode );

    my ( $aa_flag, $need_diagnostics_flag ) = ( undef, 0 );


    # Query for SOA RRs


    if ( $ip eq $query_ip ) {
	$self->dprint( "Query for SOA records for name $zone at $ip", 4 );
    }
    else {
	$self->dprint( "Query for SOA records for name $zone at $query_ip (real: $ip)", 4 );
    }

    ( $answers, $err, $aa_flag, $rcode, $auths ) = $self->lookup_records( $res, $zone, 'SOA', 0 );

    # if nameserver has DNSSEC-disabled or query times out 
    # (which may also happen if nameserver has DNSSEC disabled, or firewalled,
    # we resend the query with DNSSEC bit disabled
    if (   ($rcode && ($rcode =~ /^(FORMERR|NOTIMP|SERVFAIL)$/i) ) 
        || ( $err && ( $err =~ /query timed out/ )) )
    {
      my $plain_res = $self->get_resolver('remote',0);
      $plain_res->nameservers( $query_ip );
      $plain_res->port( $port );
      ( $answers, $err, $aa_flag, $rcode, $auths ) = $self->lookup_records( $plain_res, $zone, 'SOA', 0 );
    }

    if ( !$err && $rcode && ($rcode =~ /^(FORMERR|NOTIMP|SERVFAIL)$/i) ) {
      # to report SERVFAIL properly, since $err is not set
      $err = $rcode;
    }


    if ( $err ) {
	$self->p( 'PROBLEM_LOOKUP', $names_for_ip_text, 'SOA', $zone, $ip, $err );
	$need_diagnostics_flag = 1;
    }

    # Now go through what we got...

    foreach my $rr ( @$answers ) {
	if ( $rr->type eq 'SOA' ) {
	    push @{ $self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_list'} }, $rr;
	}
  if ( $rr->type eq 'RRSIG' ) { 
    if ($rr->typecovered eq 'SOA') {
      push @{ $self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_rrsig_list'} }, $rr;
    }
  }

    }

    # Set the aa flag for the SOA query. 
    $self->{'data'}{'zone'}{$ip}{'aa_flag'} = $aa_flag;

    # Get RRSIG RRset from auth section

    foreach my $rr ( @$auths ) {
      if ( $rr->type eq 'RRSIG' ) {
        push @{ $self->{'data'}{'zone'}{$ip}{'rrs'}{'any_rrsig_list'} }, $rr;
      }
    }

    # Query for the NS RR set
    @$answers=(); $err=0;
    
    if ( $ip eq $query_ip ) {
	$self->dprint( "Query for NS records for name $zone at $ip", 4 );
    }
    else {
	$self->dprint( "Query for NS records for name $zone at $query_ip (real: $ip)", 4 );
    }

    ( $answers, $err, $aa_flag, $rcode, $auths ) = $self->lookup_records( $res, $zone, 'NS', 0 );

    # if nameserver has DNSSEC-disabled or query times out 
    # (which may also happen if nameserver has DNSSEC disabled, or firewalled,
    # we resend the query with DNSSEC bit disabled
    if ( ($rcode && ($rcode =~ /^(FORMERR|NOTIMP|SERVFAIL)$/i) )
        || ( $err && ( $err =~ /query timed out/ )) )
    {
      my $plain_res = $self->get_resolver('remote',0);
      $plain_res->nameservers( $query_ip );
      $plain_res->port( $port );
      ( $answers, $err, $aa_flag, $rcode, $auths ) = $self->lookup_records( $plain_res, $zone, 'NS', 0 );
    }

    if ( !$err && $rcode && ($rcode =~ /^(FORMERR|NOTIMP|SERVFAIL)$/i) ) {
      # to report SERVFAIL properly, since $err is not set
      $err = $rcode;
    }

    if ( $err ) {
	$self->p( 'PROBLEM_LOOKUP', $names_for_ip_text, 'NS', $zone, $ip, $err );
	$need_diagnostics_flag = 1;
    }

    # Now go through what we got...

    foreach my $rr ( @$answers ) {
	if ( $rr->type eq 'NS' ) {
	    push @{ $self->{'data'}{'zone'}{$ip}{'rrs'}{'ns_list'} }, $rr;
	}
    }

    
    # Query for the DNSKEY RR set
    
    if ( $ip eq $query_ip ) {
	$self->dprint( "Query for DNSKEY records for name $zone at $ip", 4 );
    }
    else {
	$self->dprint( "Query for DNSKEY records for name $zone at $query_ip (real: $ip)", 4 );
    }

    ( $answers, $err, $aa_flag, undef ) = $self->lookup_records( $res, $zone, 'DNSKEY', 0 );

    # Now go through what we got...

    foreach my $rr ( @$answers ) {
	    if ( $rr->type eq 'DNSKEY' ) {
	      push @{ $self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_list'} }, $rr;
	    }
      if (($rr->type eq 'RRSIG') && ($rr->typecovered eq 'DNSKEY')) {
        push @{ $self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_rrsig_list'} }, $rr;
      }
    }

    # now query for all possible glue records attached to this IP 

    my @glue_names = $self->get_glue_name_list ($ip);

    foreach my $ns_name (@glue_names) {

      ( $answers, $err, $aa_flag, $rcode, $auths ) = $self->lookup_records( $res, $ns_name, 'A', 0 );

      # if nameserver has DNSSEC-disabled or query times out 
      # (which may also happen if nameserver has DNSSEC disabled, or firewalled,
      # we resend the query with DNSSEC bit disabled
      if (   ($rcode && ($rcode =~ /^(FORMERR|NOTIMP|SERVFAIL)$/i) ) 
          || ( $err && ( $err =~ /query timed out/ )) )
      {
        my $plain_res = $self->get_resolver('remote',0);
        $plain_res->nameservers( $query_ip );
        $plain_res->port( $port );
        ( $answers, $err, $aa_flag, $rcode, $auths ) = $self->lookup_records( $plain_res, $ns_name, 'A', 0 );
      }

      if ( $err ) {
        $self->p( 'PROBLEM_LOOKUP', $names_for_ip_text, 'A', $ns_name, $ip, $err );
        $need_diagnostics_flag = 1;
      }

      # Now go through what we got...

      foreach my $rr ( @$answers ) {
        if ( $rr->type eq 'A') {
          push @{$self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'glue_a_list'}}, $rr;
        }
      }

      # check AAAA record

      ( $answers, $err, $aa_flag, $rcode, $auths ) = $self->lookup_records( $res, $ns_name, 'AAAA', 0 );

      # if nameserver has DNSSEC-disabled or query times out 
      # (which may also happen if nameserver has DNSSEC disabled, or firewalled,
      # we resend the query with DNSSEC bit disabled
      if (   ($rcode && ($rcode =~ /^(FORMERR|NOTIMP|SERVFAIL)$/i) ) 
          || ( $err && ( $err =~ /query timed out/ )) )
      {
        my $plain_res = $self->get_resolver('remote',0);
        $plain_res->nameservers( $query_ip );
        $plain_res->port( $port );
        ( $answers, $err, $aa_flag, $rcode, $auths ) = $self->lookup_records( $plain_res, $ns_name, 'AAAA', 0 );
      }

      if ( $err ) {
        $self->p( 'PROBLEM_LOOKUP', $names_for_ip_text, 'AAAA', $ns_name, $ip, $err );
        $need_diagnostics_flag = 1;
      }

      # Now go through what we got...

      foreach my $rr ( @$answers ) {
        if ( $rr->type eq 'AAAA') {
          push @{$self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'glue_aaaa_list'}}, $rr;
        }
      }

    } # end of loop 'for glues only'

    # TODO: Check the authority bit on both queries. A problem should
    # Be flagged if both queries do not return the same authority flag
}

# Purpose:		given a remote nameserver address, replace it with the
#                       module-config-defined actual IP to query if specified.
# In: $ip               remote nameserver IP 
# Out: $query_ip:       actual IP that should be queried
#
# Comment:
# in actual use, ModuleQueryRemoteNameserverV4 and ModuleQueryRemoteNameserverV6
# should not be specified, they are only there for testing purposes.
#
sub get_remote_query_ip
{
    my $self = shift;

    my $query_ip = shift;

    my $protocol_version = &get_ip_version( $query_ip );

    if ( 
	 ( $protocol_version == 4 ) &&
	 ( defined $self->conf( 'ModuleQueryRemoteNameserverV4' ) )
    ) {
	$query_ip = $self->conf( 'ModuleQueryRemoteNameserverV4' );
    }
    elsif ( 
	 ( $protocol_version == 6 ) &&
	 ( defined $self->conf( 'ModuleQueryRemoteNameserverV6' ) )
    ) {
	$query_ip = $self->conf( 'ModuleQueryRemoteNameserverV6' );
    }

    return $query_ip;
}

# Purpose:		print a resource record
# In:	$rr:		ref to a Net::DNS:RR object
#
sub print_rr
{
    my ( $self, $rr )  = @_;

    return if ( $self->conf( 'ModuleDiagnosticLevel' ) == 0 );

    my $rr_string = $rr->string;

    if  ( $self->conf( 'ModuleDiagnosticFormat' ) eq 'html' ) {
	print "&nbsp;&nbsp;&nbsp;&nbsp;" if  ( $self->conf( 'ModuleDiagnosticFormat' ) eq 'html' );	
	

	# If it's an SOA record, make the format a bit more
	# HTML friendly. Maybe would be better to print the values
	# manually, ah well
	if ( $rr->type eq 'SOA' ) {
	    print "<b><table>\n";
	    print "<tr>".
		"<td>&nbsp;&nbsp;&nbsp;&nbsp;".$rr->name."</td>".
		"<td>".$rr->ttl."</td>".
		"<td>".$rr->class."</td>".
		"<td>".$rr->type."</td>".
		"<td>".$rr->mname."</td>".
		"<td>".$rr->rname."</td>".
		"<td> ( ".$rr->serial."</td><td>; serial</td>".
		"</tr>\n";
	    print "<tr><td colspan=6></td>".
		"<td>  ".$rr->refresh."</td><td>; refresh</td></tr>\n";
	    print "<tr><td colspan=6></td>".
		"<td>  ".$rr->retry."</td><td>; retry</td></tr>\n";
	    print "<tr><td colspan=6></td>".
		"<td>  ".$rr->expire."</td><td>; expire</td></tr>\n";

	    print "<tr><td colspan=6></td>".
		"<td>  ".$rr->minimum."</td><td>; minimum )</td></tr>\n";

	    print "</table></b>\n";
	    $rr_string =~ s/\)//;        # delete the closing bracket
	    $rr_string =~ s/$/ \)/;       # put new closing bracket at end
	    
	    $rr_string =~ s/\s*;\s*/;/g; # move labels closer to their associated values
	}else{
	    # Anything else than a SOA RR.
	    print "<B>$rr_string</B><BR>";
	}
    }
    else {
	print $rr_string, "\n";
    }
}

# Purpose:		choose an IP from the AAAA and A records returned
#			for a name
# In:	$ns_name	name we're picking an IP for
# Out:  $ip		IP number we chose
# Comments:		we give preference to IPv6 addresses.
#			we use the first valid address that we find,
#                       meaning the choice is random if there are
#			multiple equivalent records
#
sub pick_an_ip_for_ns
{
    my ( $self, $ns_name ) = @_;

    my ( $ip4, $ip6 );

    my @problems;

    my $found_a_record_flag = 0;

    # Try for AAAA records first...

    my $aaaa_rr_list = $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'aaaa_list'};

    if ( defined $aaaa_rr_list ) {

	foreach my $aaaa_rr ( @{ $aaaa_rr_list } ) {

	    $found_a_record_flag = 1;

	    my @ip_problems = $self->check_for_ip_problems( $aaaa_rr, 'Net::DNS::RR::AAAA');

	    if ( @ip_problems ) {
		push @problems, @ip_problems;
	    }
	    else {
		# We can only work with a full prefix
		
		$ip6 = &ip_expand_address( $aaaa_rr->address, 6 );

		$self->dprint( "Got IPv6 address $ip6", 2 );		
		last;
	    }
	}
    }

    # If we don't already have any address yet, or we have an IPv6
    # address but we can't make IPV6 queries, check for an IPv4 address

    my $a_rr_list = $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'a_list'};

    if ( 
	 (
	  ( ! $ip4 && ! $ip6 ) ||
	  ( $ip6 && ! $self->conf( 'ModuleUseIPv6' ) )
	 ) &&
	 defined $a_rr_list
    ) {
	foreach my $a_rr ( @{ $a_rr_list } ) {

	    $found_a_record_flag = 1;

	    # Net::DNS prints lots of debug garbage if you
	    # call method &address on a record with an empty
	    # data field.

	    my @ip_problems = $self->check_for_ip_problems( $a_rr, 'Net::DNS::RR::A');

	    if ( @ip_problems ) {
		push @problems, @ip_problems;
	    }
	    else {
		$ip4 = $a_rr->address;

		$self->dprint( "Got IPv4 address $ip4", 2 );

		last;
	    }
	}
    }

    # if this is a glue record with (multiple) ipv4/ipv6 addresses, assign ip from glue
    # prefer glue ipv4/ipv6 over whatever was resolved
    my ($ip4_glue_list, $ip6_glue_list) = $self->get_glue_ip_list($ns_name);
    my $ip6_glue;
    my $ip4_glue;

    if ($ip4_glue_list && @{$ip4_glue_list}) {
      $ip4_glue = $ip4_glue_list->[0]; # take first IP;
      $self->dprint( "Got GLUE IPv4 address $ip4_glue", 2 );
    }

    if ($ip6_glue_list && @{$ip6_glue_list}) {
      $ip6_glue = $ip6_glue_list->[0]; # take first IP;
      $self->dprint( "Got GLUE IPv6 address $ip6_glue", 2 );
    }

    # Give preference to IPV6 if we're IPV6 capable, otherwise
    # to IPv4

    my $picked_ip;

    if ( $self->conf( 'ModuleUseIPv6' ) )	{ $picked_ip = $ip6_glue || $ip4_glue || $ip6 || $ip4 || undef; }
    else			        { $picked_ip = $ip4_glue || $ip6_glue || $ip4 || $ip6 || undef; }

    $self->{'data'}{'ns'}{'byname'}{$ns_name}{'ip'} = $picked_ip;

    # Maybe we didn't find any records at all

    $self->p( 'PROBLEM_COULDNT_LOOKUP_NS_ADDRESS', $ns_name, $ns_name )
	if ( ! $found_a_record_flag && ! $picked_ip);

    return @problems;
}

# Purpose:		check for problems with an IP we found
# In:	$name	        hostname we're checking IP of
# Out:  $ip		IP number it resolved to
# if second argument is undef (no rr), then first argument is used as text ip
#
sub check_for_ip_problems
{
    my $self = shift;

    my ( $rr, $rr_type ) = @_;

    my $hostname = $rr->name;

    my @problems;

    my $ip;

    # Net::DNS prints lots of debug garbage if you
    # call method &address on an AAAA RR with an empty
    # data field.
    # Ah well.

    if ( 
	 ( defined $rr ) &&
	 ( ref( $rr ) eq $rr_type )
    ) {
	$ip = $rr->address;
    }

    my $printable_ip = $ip || '*undefined*';
    
    $self->dprint( "Check IP $printable_ip is valid and usable", 2 );

    # IP must be syntactically valid
	    
    if (  
	  ( ! defined $ip ) ||
	  ( ! &is_valid_prefix( $ip ) )
    ) {
	push @problems, [
			 'PROBLEM_BAD_IP', $hostname, 
			 $printable_ip, $hostname
	];
		
	$self->dprint( "Ignoring invalid IP address: $printable_ip for name $hostname", 3 );
    }

    # We don't like stuff broadcast addresses etc either

    elsif ( my $reason = $self->is_unusable_prefix( $ip ) ) {
	push @problems, [
			 'PROBLEM_UNUSABLE_IP', $hostname, 
			 $printable_ip, $hostname, $reason
	];
	
	$self->dprint( "Ignoring unusable prefix '$printable_ip' for name $hostname: $reason", 3 );
    }

    return @problems;
}


# Purpose:		query for the set of existing PTR records mapping from $ip
# In:	$res:		reference to a Net::DNS::Resolver object
# In:   $ip:		the IP we got from an A/AAAA record for $name
# In:	$name:		the name which had an A/AAAA record pointing to $ip
# Comments: 
#
sub get_ptr_rr_data
{
    my $self = shift;

    my ( $res, $ip, $name ) = @_;

    $self->dprint( "Looking up PTR for $ip ($name)", 2 );
    
    # We need to find the IP version of $ip first, so we know
    # which reverse domain it's in

    my $reverse_name;
    my $ip_version = &get_ip_version( $ip );

    if ( $ip_version ) {
	$reverse_name = &ip_reverse( $ip, ( $ip_version == 4 ? 32 : 128 ), $ip_version );
    }
    else {
	$self->program_error( "couldn't figure out IP version of $ip, not looking up PTR records" );
    }

    my ( $ptr_rrs, $ptr_err, undef, undef ) = $self->lookup_records( $res, $reverse_name, 'PTR', 1 );

    $self->p( 'PROBLEM_LOOKUP', $name,
	     'PTR', $reverse_name, &list_text( $res->nameservers ), $ptr_err
    ) if ( $ptr_err );

    $self->{'data'}{'ns'}{'byname'}{$name}{'rrs'}{'ptr_list'} = $ptr_rrs
	if ( $ptr_rrs && @$ptr_rrs );
    
    # 20011127 paulv
    # fix to obtain .int reverse name as well (&get_reverse_name returns arpa)

    #if ( $ip_version == 6 ) {

	#$reverse_name =~ s/arpa(\.?)$/int$1/;

   # 	my ( $int_ptr_rrs, $int_ptr_err, undef, undef ) = $self->lookup_records( $res, $reverse_name, 'PTR', 1 );

    #    push @{ $self->{'data'}{'ns'}{'byname'}{$name}{'rrs'}{'ptr_list'}}, $int_ptr_rrs
	  #  if ( $int_ptr_rrs && @$int_ptr_rrs );

    #	 $self->p( 'PROBLEM_LOOKUP', $name,
	  #      'PTR',	$reverse_name, &list_text( $res->nameservers ), $int_ptr_err 
    # ) if ( $int_ptr_err );

    #}
}

# Purpose:		pretty print a DNS packet header and check for some
#			bad values
# In:	$header:	ref to a Net::DNS:Header object
# In:	$query_flag:	0 = this is a query packet 1 = response packet
# In:	$expected_id:	For an answer packet, the ID number under which the
#			query packet was sent out, otherwise undef
#
sub print_packet_header
{
    my ( $self, $header, $query_flag, $expected_id ) = @_;

    return if ( 
		( $self->conf( 'ModuleDiagnosticLevel' ) == 0 ) || 
		( ! $self->conf( 'ModuleDiagnosticPacketHeaders' ) )
    );

    # Pick out the values we're interested in

    my $aa = $header->aa;			# Auth answer flag
    my $qr = $header->qr;			# Response flag
    my $tc = $header->tc;			# Truncated packet
    my $rd = $header->rd;			# Recursion desired
    my $ra = $header->ra;			# Recursion available
    my $id = $header->id || '?';		# Packet ID
    my $rcode = $header->rcode || '?';		# Response code
    my $qcount = $header->qdcount || '?';	# no question records
    my $ancount = $header->ancount || '?';	# no answer records
    my $aucount = $header->nscount || '?';	# no auth records
    my $adcount = $header->arcount || '?';	# no additional records

    # Specify which header fields are flags

    my %is_flag = ( 'QR' => 1, 'TC' => 1, 'AA' => 1, 'RA' => 1, 'RD' => 1 );

    # The fields we want to print depend on whether this is a query or
    # an answer packet
   
    my @relevant_fields;
    print "&nbsp;&nbsp;" if  ( $self->conf( 'ModuleDiagnosticFormat' ) eq 'html' );	

    if ( $query_flag ) {
	@relevant_fields = ( 'ID', $id, 'QR', $qr, 'RD', $rd, 'Questions', $qcount );
	print ">Q Header: ";
    }
    else {
	@relevant_fields = (
		'ID', $id, 'QR', $qr, 'AA', $aa, 'TC', $tc, 'RA', $ra, 'RCode:', $rcode,
		'Ans', $ancount, 'Auth', $aucount, 'Add', $adcount
	);
	print ">A Header: ";
    }

    # Print the relevant fields

    while ( my $label = shift @relevant_fields ) {

	my $value = shift @relevant_fields;

	if  ( $self->conf( 'ModuleDiagnosticFormat' ) eq 'html' ) {	

	    # For flags, we just print the label if it's set to save space

	    if ( $is_flag{$label} ) {
		print '<I>', $label, '</I> ' if ( $value );
	    }
	    else {
		print join '', $label, " ", '<I>', $value, '</I>', ' ';
	    }
	}
	else {
	    if ( $is_flag{$label} ) {
		print $label, ' ' if ( $value );
	    }
	    else {
		print $label, ' ', $value, ' ';
	    }
	}
    }

    # End the line

    print "<br/>" if ( $self->conf( 'ModuleDiagnosticFormat' ) eq 'html' );
    print "\n";

    # Check for some absurdities

    if ( $query_flag ) {

	$self->dprint( "!!! qr set in a query packet" ) if ( $qr );
	$self->dprint( "!!! aa set in query packet" ) if ( $aa );
	$self->dprint( "!!!  ra in query packet", 5 ) if ( $ra && $query_flag );
    }

    else {
	$self->dprint( "!!!  query id ($expected_id) different from answer id($id)", 5 )
	    if ( defined $expected_id && $id != $expected_id );

	$self->dprint( "!!! qr not set in a response packet" ) if ( ! $qr );
    }

    $self->dprint( "!!!  packet ID less than 0", 5 )         if ( $id < 0 );
    $self->dprint( "!!!  packet ID greater than 65535", 5 )  if ( $id > 65535 );
    $self->dprint( "!!!  truncated packet", 5 )              if ( $tc );

}

sub check_ns_set
{
    my $self = shift;

    $self->dprint( "Check NS set", 0 );

    $self->check_for_duplicate_delegated_ns_names()
	if ( $self->check_enabled( 'CheckForDuplicateDelegatedNameserverNames' ) );

    $self->check_number_of_nameserver_names() 
	if ( $self->check_enabled( 'CheckNumberOfNameserverNames' ) );


    $self->check_for_duplicate_ns_ips()
	     if ( $self->check_enabled( 'CheckForDuplicateNameserverIPs' ) );

    $self->check_ip_proximity()
	if ( $self->check_enabled( 'CheckIPProximity' ) );

    $self->check_compare_soas_returned()
		  if ( 
		       $self->check_enabled( 'CheckForDifferingSerialNumbers' ) ||
		       $self->check_enabled( 'CheckForDifferingSOAValues' )
	          );

    $self->compare_ns_sets_returned()
		  if ( $self->check_enabled( 'CheckForInconsistentNSListings' ) );


    $self->check_for_required_nameservers()
		  if ( $self->check_enabled( 'CheckForRequiredNameservers' ) );

    $self->check_for_prohibited_nameservers()
		  if ( $self->check_enabled( 'CheckForProhibitedNameservers' ) );
}

sub check_for_required_nameservers
{
    my $self = shift;

    if ( 
	 ( defined $self->conf( 'ParamRequiredNameserverList' ) ) &&
	 @{ $self->conf( 'ParamRequiredNameserverList' ) }  
    ) {

	my @delegated_list = @{ $self->{'data'}{'delegation'}{'bare_ns_list'} };    
	my @required = @{ $self->conf( 'ParamRequiredNameserverList' ) };
	my @missing_required;

	# Find missing ones

	foreach my $required_server ( @required ) {
	    my $normal_required_server = &normalise_domain_name( $required_server );

	    push @missing_required, $required_server
		if ( ! grep { $normal_required_server eq $_ } @delegated_list );
	}

	# Compare what we found to the spec.

	if ( 
	     ( 
	       ( $self->conf( 'ParamRequiredNameserverSpec' ) eq 'exactly-one' ) &&
	       ( ( @required - @missing_required ) != 1  )
	     ) ||
	     (
	        ( $self->conf( 'ParamRequiredNameserverSpec' ) eq 'all' ) &&
	        ( @missing_required > 0 )
	     ) ||
	     ( 
	       ( $self->conf( 'ParamRequiredNameserverSpec' ) eq 'at-least-one' ) &&
	       ( scalar( @missing_required ) == scalar( @required ) )
	     ) 
	) {
	    # Make the spec more readable
	    
	    my $spec_descr = $self->conf( 'ParamRequiredNameserverSpec' );
	    $spec_descr =~ s/-/ /;
	    
	    $self->p( 'PROBLEM_REQUIRED_NAMESERVER_MISSING', 'set',
		      $spec_descr, &list_text( @required ), 
		      &list_text( @missing_required ), &list_text( @delegated_list )
	    );
	}
    }
}

sub check_for_prohibited_nameservers
{
    my $self = shift;

    my @delegated_list = @{ $self->{'data'}{'delegation'}{'bare_ns_list'} };    

    if ( defined $self->conf( 'ParamProhibitedNameserverList' ) ) {
	foreach my $prohibited_server ( @{ $self->conf( 'ParamProhibitedNameserverList' ) } ) {
	    
	    my $normal_prohibited_server = &normalise_domain_name( $prohibited_server );
	    
		$self->p( 'PROBLEM_PROHIBITED_NAMESERVER_PRESENT', 'set',
			  $prohibited_server, &list_text( @delegated_list )
		  )
		    if ( grep { $prohibited_server eq $_ } @delegated_list );
	}
    }
}

# Purpose:		check there were the required number of NS delegated
# Comments:		we could use the 'unique_pruned_ns_list', but
#			duplicate nameserver names in the delegated list are
#			already reported elsewhere
#
sub check_number_of_nameserver_names
{
    my $self = shift;
    my @the_list = @{ $self->{'data'}{'delegation'}{'bare_ns_list'} };    

    my $min = $self->conf( 'ParamMinimumNumberNameservers' );

    $self->dprint( "Check number of nameserver names", 1 );
    
    $self->p( 'PROBLEM_TOO_FEW_DELEGATED_NAMESERVERS', 'set',
	      scalar (@the_list), &list_text( @the_list ), $min
    )
	if ( @the_list < $min );
}

# Purpose:		look for two nameserver names having resolved to
#			same IP address.
# Comments:		as noted previously, if an NS name resolves to several
#			addresses, we pick a random one, so this test isn't
#			very reliable
#
sub check_for_duplicate_ns_ips
{
    my $self = shift;

    $self->dprint( "Check for duplicate IP's", 1 );

    while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'ns'}{'byip'} } ) {

	my @name_list = @{ $ip_ref->{'name_list'} };

	$self->dprint( "IP $ip: ". &list_text( @name_list ), 2 );
	
	$self->p(  'PROBLEM_DUPLICATE_DELEGATED_NS_IP', 'set', $ip, &list_text( @name_list ) )
	    if ( @name_list > 1 );
    }
}

# Purpose:		check for NS IP's being closeish numerically
# Comments:		'close' for IPv4 we interpret as being the same /24
#			'close' for IPv6 we interpret as being the same subnet (SLA)
#
sub check_ip_proximity
{
    my $self = shift;

    $self->dprint( "Check IP proximity of NSs", 1 );
    
    my @ips = keys %{ $self->{'data'}{'ns'}{'byip'} };

    # If there's only one IP left, nothing to compare it with :-)

    while ( 
	    ( @ips > 1 ) &&
	    ( my $outer_ip = shift @ips )
    ) {

	# Get names associated with outer IP

	my $outer_names = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$outer_ip}{'name_list'} } );

	$self->dprint( "Comparing: $outer_ip ($outer_names)", 2 );

	# Check for an embedded IPv4 addressi in $outer_ip. In this case, we
	# want to use the embedded value.

	# Lee 20030930
	# Net::IP can't handle this stuff

	#if ( &ip_is_ipv6( $outer_ip ) && defined ( my $affix = &get_ipv4_affix_6( $outer_ip ) ) ) {
	#   $self->dprint( "(embedded IPv4: treating as $affix)" );
	#  $outer_ip = $affix;
	#}

	my @inner_ips = @ips;

	# Compare $outer_ip with all the other IP's we havn't already
	# compared it to

	foreach my $inner_ip ( @inner_ips ) {

	    # Get names associated with inner IP
	
	    my $inner_names = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$inner_ip}{'name_list' } } );

	    # Check for an embedded IPv4 address in $inner_ip. In this case, we
	    # want to use the embedded value.

	    # Lee 20030930
	    # Net::IP can't handle this stuff

#	    if ( &ip_is_ipv6( $inner_ip ) && defined ( my $affix = &get_ipv4_affix_6( $inner_ip ) ) ) {
#		$self->dprint( "(extracted embedded IPv4 $affix from $inner_ip )" );
#		$inner_ip = $affix;
#	    }

	    if ( &ip_is_ipv4 ( $inner_ip ) ) {

	        # Compare only IPv4 with IPv4...

    		next if ( ! &ip_is_ipv4( $outer_ip ) );

		$self->dprint( "...with $inner_ip ($inner_names)" );

		my @outer_octets = split /\./, $outer_ip;
		my @inner_octets = split /\./, $inner_ip;

		# 'Close' if first three octets match (i.e. in same /24)

		if ( ( join '.', @outer_octets[0..2] ) eq ( join '.', @inner_octets[0..2] ) ) {

		    # Make sure names come out in a predictable order for
		    # testing purposes

		    if ( $inner_names cmp $outer_names ) {
			$self->p( 'PROBLEM_DELEGATED_NS_ON_SAME_SUBNET', 'set', 
				  $outer_ip, $outer_names,
				  $inner_ip, $inner_names
			  );
		    }
		    else {
			$self->p( 'PROBLEM_DELEGATED_NS_ON_SAME_SUBNET', 'set', 
				  $inner_ip, $inner_names,
				  $outer_ip, $outer_names

			  );
		    }
		}

	    }
	    
	    elsif ( &ip_is_ipv6( $inner_ip ) ) {

  	        # ...or IPv6 with IPv6

		# next if ( ! &ip_is_ipv6( $outer_ip ) );

		#$self->dprint( "...with $inner_ip ($inner_names)" );

		#my $outer_binary = &ip62bin( $outer_ip );
		#my $inner_binary = &ip62bin( $inner_ip );

		#my $inner_sla_binary = &get_field_value_6( $inner_binary, 'SLA' );
		#my $outer_sla_binary = &get_field_value_6( $outer_binary, 'SLA' );		    

		# IPv6's are considered close if they both have an SLA field,
		# and it's the same

		#$self->p( 'PROBLEM_DELEGATED_NS_ON_SAME_SUBNET', 'set', 
		#	 $inner_ip, $inner_names,
		#	 $outer_ip, $outer_names
		#)
		#    if ( $outer_sla_binary && $inner_sla_binary &&
		#	$outer_sla_binary eq $inner_sla_binary
  		#    );
	    }
	    else {
		&program_error( "IP $inner_ip couldn't be classified as either IPv4 or IPv6" );
	    }
	}
    }
}

# Purpose:		compare the SOA's we got back for differing
#			serial numbers anf mnames
#
sub check_compare_soas_returned
{
    my $self = shift;

    $self->dprint( "Comparing primary names and serial numbers on all nameservers", 1 );

    # Basically, we report if the serial or mname we're currently
    # looking at is different from the previous ones we looked at

    my ( $previous_ip, $previous_names_text, %previous_values );

    my $soa_keys = $self->conf( 'ModuleSOAKeys' );

    while (  my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

	my $ns_names_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );

	my $soa_rrs = $ip_ref->{'rrs'}{'soa_list'};

	# Ignore nameservers we couldn't get an SOA for

	if ( ! defined $soa_rrs || ! @$soa_rrs ) {
	    $self->dprint( "$ip ($ns_names_text): ignoring, couldn't get an SOA.", 2 );
	    next;
	}

	# Ignore nameservers with multiple SOA's

	elsif ( @$soa_rrs > 1 ) {
	    $self->dprint( "$ip ($ns_names_text): ignoring, got multiple SOA's.", 2 );
	    next;
	}
	
        # Ignore nameservers that we couldn't zone transfer an SOA, or
	# query for the 'SOA' query didn't give an AA answer

	elsif ( ! $ip_ref->{'aa_flag'} ) {
	    $self->dprint( "$ip ($ns_names_text): ignoring havn't got authoritative SOA", 2 );
	    next;
	}

	my $soa_rr = $soa_rrs->[0];

	# If this is the first one we're looking at...

	if ( ! defined $previous_ip ) {

	    $previous_ip = $ip;
	    $previous_names_text = $ns_names_text;
		
	    foreach my $key ( @$soa_keys ) {
		$previous_values{$key} = $soa_rr->{$key};
	    }
	}

	# Otherwise mention what we're comparing in the diagnostics

	else {
	    $self->dprint( "Comparing $ip ($ns_names_text) with $previous_ip ($previous_names_text)",  2 );
	}
	
        # Check to see if any SOA values are different from in the
        # previous SOA record we looked at

	my %differing_keys;

	foreach my $key ( @$soa_keys ) {
		# paulv 20030211 changed to use case insensitive regexp instead
		if ( 
		    defined $previous_values{$key} &&
		    defined $soa_rr->{$key} &&
		    ( (lc $previous_values{$key}) ne (lc $soa_rr->{$key}) )
		) {
			$differing_keys{$key} = 1;

			$self->dprint( "previous_values{$key} = *". $previous_values{$key} ."*", 3);
			$self->dprint( "soa_rr->{$key} = *". $soa_rr->{$key} ."*", 3);
		}
	}

	# Report only the serial number difference, if it is different.
	# It's no surprise that other values are different if the serial is
	
	if ( $differing_keys{'serial'} ) {

	    # We need to do this even if the serial number check is disabled,
	    # so we can decide whether to check the other SOA values

	    $self->p( 'PROBLEM_SERIAL_NUMBERS_DIFFER', 'set',
		     $ip,
		     $ns_names_text,
		     $soa_rr->serial,
		     $previous_ip,
		      $previous_names_text,
		     $previous_values{'serial'}
	    )
		if ( $self->check_enabled( 'CheckForDifferingSerialNumbers' ) );
	}

	# If serial number is the same but other values are different
	# they've probably forgotten to update the serial on the primary

  # this is added on request of Brett by katie 20060210
  if ($differing_keys{'mname'} && ! $self->check_enabled('CheckSOAMnameDiffer') ) {
     delete $differing_keys{'mname'};
  }
	
	if (  
	      ( !$differing_keys{'serial'} && keys %differing_keys > 0) ||
	      ( $differing_keys{'serial'} && keys %differing_keys > 1)
	      ) {

	    $self->p( 'PROBLEM_SOA_VALUES_DIFFER', 'set',
		     $ip,
		      $ns_names_text,
		     $previous_ip,
		      $previous_names_text,
		     join ', ', keys %differing_keys
	   )
		if ( $self->check_enabled( 'CheckForDifferingSOAValues' ) );
	}

	# Record current values for comparing with next values

	$previous_ip = $ip;
	$previous_names_text = $ns_names_text;

	foreach my $key ( @$soa_keys ) {
	    $previous_values{$key} = $soa_rr->{$key};
	}
    }
}

# Purpose:		compare the NS lists returned by each NS
# Comments:		we get a list of NS names referenced with the
#			IP's that referenced them. Then we compare the
#			IP's that referenced them with the full set of IP's
#
sub compare_ns_sets_returned
{
    my $self = shift;

    $self->dprint( "Checking consistency of NS lists returned", 1 );

    # Get a hash with ip->@names_listed_there

    my %listed_names_by_ip;

    while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

	# If we didn't get an SOA for the zone at the IP, don't bother
	# including it in this check: we will already have failed the
	# thing based on that.

	# Note that a nameserver may have an empty NS list: we can't
	# exclude it based on that (because it's -totally' inconsistent)

	if ( exists $ip_ref->{'rrs'}{'soa_list'} ) {
	    @{ $listed_names_by_ip{$ip} } = &ns_rrs_to_ns_names( $ip_ref->{'rrs'}{'ns_list'} );
	}
    }

    # Get a hash with listed_name->@ips_listing_it

    my %ips_by_listed_name;

    while ( my ( $ip, $name_list ) = each %listed_names_by_ip  ) {
	foreach my $name ( @$name_list ) {
	    push @{ $ips_by_listed_name{$name} }, $ip;
	}
    }

    # For each name that is listed somewhere at one of the IPs...

    while ( my ( $ns_listed, $ip_listing_ref ) = each %ips_by_listed_name ) {

	$self->dprint( "Checking where $ns_listed *isn't* listed", 2 );

	my @not_listed_by;
	foreach my $ip_listing ( keys %listed_names_by_ip ) {

	    # add an NS as not listing another NS if we got and
	    # answer from it, but the other NS isn't listed in the
	    # answer

	    push @not_listed_by, $ip_listing
		if ( 
		    defined $listed_names_by_ip{$ip_listing} &&
		    ! grep { $_ eq $ns_listed } @{ $listed_names_by_ip{$ip_listing} } 
		);
	}

	my @ip_listing_names = &ip_list_to_name_list( 
			$self->{'data'}{'ns'}{'byip'}, @$ip_listing_ref 
	);
	my @not_listed_by_names = &ip_list_to_name_list( 
			$self->{'data'}{'ns'}{'byip'}, @not_listed_by
	);


	if ( @not_listed_by ) {
	    $self->p( 'PROBLEM_INCONSISTENT_NS_LISTING', 'set',
		     $ns_listed,
		     &list_text( @$ip_listing_ref ),
		     &list_text( @ip_listing_names ),
		     &list_text( @not_listed_by ),
		     &list_text( @not_listed_by_names )
	    );
	}
	else {
	    $self->dprint( "OK, it's listed at all servers we got an answer from.", 3 );
	}
    }
}

# Purpose:		check for a name appearing twice in the submmitted
#			NS list
#
sub check_for_duplicate_delegated_ns_names
{
    my $self = shift;

    # Check all. Otherwise, say, ns.ripe.net is included twice it's not noticed

    $self->dprint( "Check for duplicate nameserver names", 1 );
    
    my ( %seen, %duplicate_count );

    # Count duplicates

    foreach my $item ( @{ $self->{'data'}{'delegation'}{'bare_ns_list'} } ) {
	$duplicate_count{$item}++
	    if ( exists $seen{$item} );

	$seen{$item}++;
    }

    # Report all duplicates found

    while ( my ( $duplicate_ns, $count ) = each %duplicate_count ) {
	$self->p( 'PROBLEM_DUPLICATE_DELEGATED_NS_NAME', 'set',  $duplicate_ns, $count );
    }
}

# Purpose:		check for a name appearing twice in the NS list at a nameserver
#
sub check_ns_list_for_duplicates
{
    my ( $self, $ip, $names_for_ip_text ) = @_;

    my @ns_names_at_this_ip = &ns_rrs_to_ns_names( $self->{'data'}{'zone'}{$ip}{'rrs'}{'ns_list'} );

    $self->dprint( "Check for duplicate nameserver names listed at $ip ($names_for_ip_text)", 2 );
    
    my ( %seen, %duplicate_count );

    # Count duplicates

    foreach my $item ( @ns_names_at_this_ip ) {
	$duplicate_count{$item}++
	    if ( exists $seen{$item} );

	$seen{$item}++;
    }

    # Report all duplicates found

    while ( my ( $duplicate_ns, $count ) = each %duplicate_count ) {
	$self->p( 'PROBLEM_DUPLICATE_NS_NAME_LISTED_AT_ZONE', $names_for_ip_text,
		  $duplicate_ns, $ip, $names_for_ip_text, $count 
	);
    }
}

# Purpose:			compare the list of nameservers for the zone that a
#				NS returned, with the list that was delegated
# In:	$ip			the IP of the relevant nameserver
# In:   $names_for_ip_text	comma-separated names of NS which had this IP address
#
sub check_ns_list_against_delegated
{
    my ( $self, $ip, $names_for_ip_text ) = @_;

    my $ns_rrs = $self->{'data'}{'zone'}{$ip}{'rrs'}{'ns_list'};

    my $bare_delegated_ns_list = $self->{'data'}{'delegation'}{'bare_ns_list'};



    my @ns_names_at_this_ip = &ns_rrs_to_ns_names( $ns_rrs );

    if ( $self->check_enabled( 'CheckAllDelegatedNameserversListedAtZone' ) ) {
	$self->dprint( "Check for delegated NS not at this IP", 2 );
	
	$self->dprint( "Delegated list: ". &list_text( @$bare_delegated_ns_list ), 3 );
	$self->dprint( "Listed by $ip: ". &list_text( @ns_names_at_this_ip ), 3 );
       
	foreach my $delegated_ns ( @$bare_delegated_ns_list ) {

	    $self->p( 'PROBLEM_DELEGATED_NS_NOT_LISTED_AT_ZONE', $names_for_ip_text,
		      $delegated_ns, $ip, $names_for_ip_text
		      )
		if ( ! grep { $_ eq $delegated_ns } @ns_names_at_this_ip );

		     
;
	}
    }

    if ( $self->check_enabled( 'CheckListedNameserversInDelegatedList' ) ) {

	$self->dprint( "Check for NS at this IP not in delegated list", 2 );
	
	$self->dprint( "Delegated list: ". &list_text( @$bare_delegated_ns_list ), 3 );
	$self->dprint( "Listed by $ip: ". &list_text( @ns_names_at_this_ip ), 3 );
	
	foreach my $listed_ns ( @ns_names_at_this_ip ) {
	    
	    $self->p( 'PROBLEM_LISTED_NAMESERVER_NOT_IN_DELEGATED_LIST', $names_for_ip_text,
		 $listed_ns, $ip, $names_for_ip_text, 
		      &list_text( @$bare_delegated_ns_list )
		      )
		if ( ! grep { $_ eq $listed_ns } @$bare_delegated_ns_list );
	}
    }
}

# Purpose: check_glue_rrs
# This is to implement the following check:
# take glue name, query for A/AAAA glue_name @ glue_ip (first one)
# check that at least one of the glue IPs has matching A/AAAA record.
# this might need to change...
#
sub check_glue_rrs
{

    my $self = shift;

    return unless ($self->check_enabled( 'CheckGlueIPRRMissing' ) );

    $self->dprint( "Check A/AAAA glue RRs for each glue name", 0 );

    # analyze the A/AAAA rrs that we have got from the glue servers

    foreach my $ns_name ( keys %{$self->{'data'}{'ns'}{'byname'} } ) {
      my $name_ok;
 
      if ( $self->is_glue($ns_name)) {
        my ($glue_ip4_list, $glue_ip6_list) = $self->get_glue_ip_list($ns_name);

        if ($glue_ip4_list && @{$glue_ip4_list} &&
            exists $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'glue_a_list'} ) {

          foreach my $ip (@{$glue_ip4_list}) {
            foreach my $rr (@{$self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'glue_a_list'}}) {
              if ($rr->type eq 'A'
                  && ( lc(&ip_expand_address($rr->address, 4)) eq $ip) ) {
                  $name_ok = 1;
                  last;
              }
            }
          }
        }

        if ($glue_ip6_list && @{$glue_ip6_list} &&
            exists $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'glue_aaaa_list'} ) {

          foreach my $ip (@{$glue_ip6_list}) {
            foreach my $rr (@{$self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'glue_aaaa_list'}}) {
              if ($rr->type eq 'AAAA'
                  && ( lc(&ip_expand_address($rr->address, 6)) eq $ip) ) {
                  $name_ok = 1;
                  last;
              }
            }
          }
        }
        if (! $name_ok) {
          $self->p('PROBLEM_GLUE_IP_RR_MISSING', $ns_name, $ns_name);
        }
      }
    }
}

sub check_zone_at_ips
{
    my $self = shift;


    $self->dprint( "Check zone at each IP", 0 );

    my %serials_seen;
    
    my %skipped_serials;

    while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

        my $names_for_ip_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );
	
	$self->dprint( "Check at $ip ($names_for_ip_text)", 1 );

	# If we didn't get records for this zone for whatever reason, quit now
	# We'll have complained about it earlier ( BAD_PORT, whatever)


 	if ( ! defined $ip_ref ) {
	    $self->dprint( "Aborting, we couldn't query for data at this IP.", 2 );
	    next;
	}

	if ( 
	     ( &ip_is_ipv6( $ip ) ) && 
	     ( ! $self->conf( 'ModuleUseIPv6' ) )
	) {
	    $self->dprint( "Aborting, it's an IPv6 IP and ModuleUseIPv6 is off", 2 );
	    next;
	}

	my $ok_soa = $self->check_soa_valid( $ip, $names_for_ip_text );

	# If we got an OK SOA

        if ( defined $ok_soa ) {

	    # Check it unless we've seen it before

	    if ( defined $ok_soa->serial && exists $serials_seen{$ok_soa->serial} ) {

		# If we've seen this serial before, we don't want to do the
		# checks, since we can probably assume that the values are the
		# same as for the other nameserver that had this serial
		
		# If they aren't, they'll get picked up by the comparison of
		# the SOA records. This is better than repeating the same
		# problem reports for each nameserver when the primary has
		# a bad value in the SOA record
    
		my $previous_ip = $serials_seen{$ok_soa->serial};


		
		my $previous_ip_ref = $self->{'data'}{'ns'}{'byip'}{$previous_ip};

		my $names_for_previous_ip_text = &list_text( @{ $previous_ip_ref->{'name_list'} } );
		    

		# Check the TTL on all aquired RRsets


		if (defined ($self->{'data'}{'zone'}{$ip}{'rrs'}{'ns_list'}) &&
		    $self->check_enabled( 'CheckRRsetTTL' )){
		    my @ns = @{$self->{'data'}{'zone'}{$ip}{'rrs'}{'ns_list'}};
		    $self->check_rrset_ttl(  $names_for_ip_text, @ns ) if @ns>1;
		}

		$self->dprint( "Aborting further checks at this NS: seen this serial ".
			"already at $previous_ip ($names_for_previous_ip_text)"
			, 2
		);

		
		$skipped_serials{$ok_soa->serial}++;
	    }
	    else {
		$self->check_serial( $names_for_ip_text, $ok_soa->serial )
		    if ( 
			 $self->check_enabled( 'CheckSerialRecommendedFormat' )
		    );
		
		
		$self->check_soa_values( $ip, $names_for_ip_text );


		if (defined ($self->{'data'}{'zone'}{$ip}{'rrs'}{'ns_list'}) &&
		    $self->check_enabled( 'CheckNameserverMinimumTTL' )){
		    my @ns = @{$self->{'data'}{'zone'}{$ip}{'rrs'}{'ns_list'}};
		   $self->check_nameserver_minimum_ttl(  $names_for_ip_text, $ns[0] ) if @ns>0;
		}



			
		$self->check_ns_list_against_delegated( $ip, $names_for_ip_text )
		    if ( 
			 $self->check_enabled( 'CheckAllDelegatedNameserversListedAtZone' ) ||
			 $self->check_enabled( 'CheckListedNameserversInDelegatedList' )

		    );

		$self->check_ns_list_for_duplicates( $ip, $names_for_ip_text )
		    if ( $self->check_enabled( 'CheckForDuplicateNameserverNamesAtZone' ) );

		# Since we havn't checked this serial before, we should
		# record that it's now been checked

		$serials_seen{$ok_soa->serial} = $ip
		    if ( defined $ok_soa->serial );
	    }
	}
    }

    # If we skipped any zone checks due to identical serials, record here
    # Otherwise it's confusing if we just report an error at one server.

    if ( $self->check_enabled( 'CheckSkippedSerial' ) ) {
	foreach my $serial_no ( keys %skipped_serials ) {
	    $self->p( 'PROBLEM_SKIPPED_SERIAL_NUMBER', 'set', $serial_no );
	
	}
    }
}

# Purpose:			check an SOA serial field
# In:	$problem_entity:	which key to record the problem under
# In:	$serial:		the value to check
#
sub check_serial
{
    my ( $self, $problem_entity, $serial ) = @_;

    $self->dprint( "Check SOA serial", 2 );

    if ( ! defined $serial ) {
	$self->p( 'PROBLEM_SOA_RECORD_FIELD_EMPTY', $problem_entity, 'serial' );
    }
    else {
	$self->dprint( "Checking serial '$serial'", 3 );
    
	# Not all digits

	if ( $serial =~ /^(\d+)$/ ) {

	    # Digits but not recommended format

	    if ( $self->check_enabled( 'CheckSerialRecommendedFormat' ) ) {

		if ( 
		    $serial =~ /
			      ^
			      (\d{4})		# year
			      (\d{2})		# month
			      (\d{2})		# day
			      (\d{2})		# counter
			      $
			   /x 
		) {
		    my ( $year, $month, $day, $counter )= ( $1, $2, $3, $4 );

		    if ( 
			 $year <  $self->conf( 'ParamFirstReasonableSerialYear' ) ||
			 $month < 1 ||       # first month
			 $month > 12 ||      # last month
			 $day < 1 ||       # first day
			 $day > 31         # last day
		    ) {
			$self->p( 'PROBLEM_SERIAL_NOT_RECOMMENDED_FORMAT', $problem_entity, $serial );
		    }
		    else {

			# We need localtime secs in order to be able to see whether the
			# SOA date is too far in the future

			my $localtime_seconds;

			eval {			
			    $localtime_seconds = timelocal( 0, 0, 0, $day, $month - 1, $year);
			};

			# Timelocal might croak here.
			# Most likely reason at this point is e.g. 30th day of February specified

			if ( $@ ) {
			    $self->p( 'PROBLEM_SERIAL_NOT_RECOMMENDED_FORMAT', $problem_entity, $serial );
			}

			elsif ( $self->check_enabled( 'CheckSerialNotInFuture' ) ) {
			    my $allowance_time_seconds = $self->conf( 'ParamSerialFutureAllowanceHours' ) * 3600;
			    if ( $localtime_seconds > ( time + $allowance_time_seconds ) ) {
				$self->p( 'PROBLEM_SERIAL_IN_FUTURE', $problem_entity, $serial );
			    }
			}
		    }
		}
		else {	
		    $self->p( 'PROBLEM_SERIAL_NOT_RECOMMENDED_FORMAT', $problem_entity, $serial );
		}
	    }

	}

	else {
	    &program_error("Serial number $serial matching failed");
	}
    }
}

# Purpose:			check the SOA record we found an OK SOA for the zone
#				at a particular IP
# In:	$ip:			IP source of SOA
# In:   $names_for_ip_text	comma-separated names of NS which had this IP address
#
sub check_soa_valid
{
    my ( $self, $ip, $names_for_ip_text ) = @_;

    $self->dprint( "Check SOA valid", 2 );

    # If we didn't get an SOA there's nothing to check

    my $soa_rrs = $self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_list'};

    if ( ! $soa_rrs || @$soa_rrs == 0 ) {
	$self->p( 'PROBLEM_NO_SOA_AT_SERVER', $names_for_ip_text,
		  $ip, $names_for_ip_text 
	) if ( $self->check_enabled( 'CheckNoSOARecordAtServer' ) );
    }

    # If there was more than one SOA returned, don't do any checks there's
    # something badly wrong :-) Note: Net::DNS::Resolver only returns one
    # of the SOA records which is included in a zone transfer.

    elsif ( @$soa_rrs > 1 ) {
	$self->p( 'PROBLEM_MULTIPLE_SOA_AT_SERVER', $names_for_ip_text, 
		  $ip, $names_for_ip_text
	)
	    if ( $self->check_enabled( 'CheckMultipleSOARecordsAtServer' ) );
    }		

    # If we didn't manage a zone transfer and the nameserver returned a non
    # authoritative answer, there's no point checking it

    elsif ( ! $self->{'data'}{'zone'}{$ip}{'aa_flag'} ) {
	$self->p( 'PROBLEM_NON_AA_NAMESERVER', $names_for_ip_text, 
		  $ip, $names_for_ip_text )
	    if ( $self->check_enabled( 'CheckSOARecordAtServerAuthoritatative' ) );
    }
 
    # Otherwise, we have a single authoritative SOA, return it's serial number

    else {
	return $soa_rrs->[0];
    }

    return undef;
}

# Purpose:			check the SOA record we found at $ip for $zone
# In:	$ip:			IP source of SOA
# In:   $names_for_ip_text	comma-separated names of NS which had this IP address
#
sub check_soa_values
{
    my ( $self, $ip, $names_for_ip_text ) = @_;

    $self->dprint( "Check SOA values", 2 );

    my $soa = $self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_list'}[0];
    my $mname_rrs = $self->{'data'}{'zone'}{$ip}{'mname_rrs'};

    $self->check_mname( 
			$names_for_ip_text, $soa->mname, 
			$mname_rrs,
			&ns_rrs_to_ns_names( $self->{'data'}{'zone'}{$ip}{'rrs'}{'ns_list'} )
    );
    
    $self->check_rname( $names_for_ip_text, $soa->rname );
    
    $self->check_refresh( $names_for_ip_text, $soa->refresh );
    
    $self->check_retry( $names_for_ip_text, $soa->retry, $soa->refresh );
    
    $self->check_expire( $names_for_ip_text, $soa->expire, $soa->refresh, $soa->retry );
    
    $self->check_minimum( $names_for_ip_text, $soa->minimum );
}

# Purpose:			check an SOA mname field
# In:	$problem_entity:	which key to record the problem under
# In:	$mname:			the value to check
# In:   @ns_list:		the nameservers  found listed for the zone at the
#				same nameserver
# Comments:			mname is supposed to be the name of the primary master
#
sub check_mname
{
    my ( $self, $problem_entity, $mname, $mname_rrs, @ns_list ) = @_;

    $self->dprint( "Check SOA mname", 3 );

    $mname = &normalise_domain_name( $mname ); 

    # Added Lee 20001116
    # Check mname defined

    if ( ! $mname ) {
	$self->p( 'PROBLEM_SOA_RECORD_FIELD_EMPTY', $problem_entity, 'mname' );
    }

    # If so, do required checks on it

    else {
	$self->check_hostname( $problem_entity, $mname, 'the MNAME field of the SOA record' );

	if ( $self->check_enabled( 'CheckMnameInNSSetAtServer' ) ) {
	
	    if ( @ns_list ) {

		$self->dprint( 
			       "Looking for mname '$mname' in the NS records listed on this server (".
			       &list_text( @ns_list ).
			       ')',
			       3
			       );

		$self->p( 'PROBLEM_MNAME_NOT_LISTED_AS_NS', $problem_entity,
			  $mname, &list_text( @ns_list ) 
			 )	
		    if ( ! grep { $_ eq $mname } @ns_list );
	    }
	    else {
		$self->dprint( "Not comparing mname with NS list, no NS listed", 4 );
	    }
	}

	# See if we could get an A or AAAA record for the mname

	if ( $self->check_enabled( 'CheckMnameResolvable' )) {

	    $self->dprint( "Check whether $mname resolved", 3 );

	    if ( ! (
		      ( defined $mname_rrs->{'a_list'} && @{ $mname_rrs->{'a_list'} } ) ||
		      ( defined $mname_rrs->{'aaaa_list'} && @{ $mname_rrs->{'aaaa_list'} } )
	           ) 
	    ) {
		$self->p( 'PROBLEM_MNAME_DOESNT_RESOLVE', $problem_entity, $mname );
	    }
	}
    }     
}

# Purpose:			check an SOA rname field
# In:	$problem_entity:	which key to record the problem under
# In:	$rname:			the value to check
# In:   $zone:			the zone the SOA record marks
# Comments:			rname is the email contact address for the zone
#
sub check_rname
{
    my ( $self, $problem_entity, $rname ) = @_;

    $self->dprint( "Check SOA rname", 3 );

    my $zone = $self->{'data'}{'delegation'}{'zone'};

    if ( ! $rname ) {
	$self->p( 'PROBLEM_SOA_RECORD_FIELD_EMPTY', $problem_entity, 'rname' );
    }

    else {
	$self->dprint( "Checking rname '$rname'", 4 );

	# Check duplicate domain in rname,
	# usually signifies they forgot to put a dot on the end of the
	# name in the zonefile

	if ( 
	    $self->check_enabled( 'CheckRnameNoDuplicateZone' ) &&
	    $rname =~ /\.$zone\.$zone\.?$/i
	){
	    $self->p( 'PROBLEM_RNAME_DUPLICATE_ZONE', $problem_entity, $rname );
	    $self->disable_check("CheckRnameMail");
	    }

	# Check if it's reverse domain


	if ( 
	    $self->check_enabled( 'CheckRnameNotReverseZone' ) &&
	    &is_reverse_zone( $rname ) 
	){
	    $self->p( 'PROBLEM_REVERSE_DOMAIN_IN_RNAME', $problem_entity, $rname );
	    $self->disable_check("CheckRnameMail");
	}
	# Find and check the domain name

	my $tidy_domainname;

	# Either if it's got an @, which case domain
	# name assumed to be on RHS of @

 	if ( $rname =~ /
	    (.+?)
	    \@
	    (.+)          # first part of domain name
	    $
	    /x
	) {
	    my $domain_name = $2;
	    
	    if ( $self->check_enabled( 'CheckRnameHasNoAtSign' ) ){
		$self->p( 'PROBLEM_RNAME_HAS_AT_SIGN', $problem_entity, $rname );
		$self->disable_check("CheckRnameMail");
	    };
	    
	    if ( $self->check_enabled( 'CheckRnameAgainstDomainNameSyntax' ) ){	    
		$self->check_rname_host_syntax( $problem_entity, $domain_name );
		$self->disable_check("CheckRnameMail");
	    }
		    

	}

	# Or it hasn't, in which case the domain name is
	# the thing on the RHS of the first non-commented
	# '.'

	elsif ( $rname =~ 
	      /^
		 (.*?[^\\])    # anything preceeding not a \.
		  \.           # then a .
		 (.+)          # then anything
	      $/x
	) {
	    my $hostname=$2;
	    my $username=$1;
	    $username =~ s/\\\./\./g;
	    my $emailaddress = $username.'@'.$hostname;

	    $self->check_rname_host_syntax( $problem_entity, $hostname )
		if ( $self->check_enabled( 'CheckRnameAgainstDomainNameSyntax' ) );


	    $self->check_rname_mail(  $emailaddress, $zone, $problem_entity,$rname )
		if ( $self->check_enabled ( 'CheckRnameMail' ));
	}
	elsif ( $self->check_enabled( 'CheckRnameMiscSyntax' ) ) {
	    $self->p( 'PROBLEM_RNAME_BAD_SYNTAX', $problem_entity, $rname );
	}
    }
}

sub check_rname_host_syntax
{
    my ( $self, $problem_entity, $domain_name ) = @_;

    my $tidy_domainname = &normalise_domain_name( $domain_name );

    $self->dprint( "Checking $tidy_domainname against domainname syntax", 5 );
		
    $self->check_hostname( $problem_entity, $domain_name, 'the RNAME field of the SOA record' );
}

# Purpose:			check an SOA refresh value
# In:	$problem_entity:	which key to record the problem under
# In:	$refresh		the value to check
#
sub check_refresh
{
    my ( $self, $problem_entity, $refresh ) = @_;

    $self->dprint( "Check SOA refresh", 3 );

    if ( ! defined $refresh ) {
	$self->( 'PROBLEM_SOA_RECORD_FIELD_EMPTY', $problem_entity, 'refresh' );
    }

    elsif ( $self->check_enabled( 'CheckRefreshRange' ) ) {

	my $min = $self->conf( 'ParamRefreshMin' );
	my $max = $self->conf( 'ParamRefreshMax' );

	$self->dprint( "Checking refresh ($refresh) against min $min, max $max", 4 );

	if ( ( $refresh < $min ) || ( $refresh > $max ) ) {
	    $self->p( 'PROBLEM_REFRESH_OUT_OF_RANGE', $problem_entity,
		      $refresh, &seconds_to_descr_string( $refresh ),
		      $min, $max
            );
	}
    }
}

# Purpose:			check an SOA retry value
# In:	$problem_entity:	which key to record the problem under
# In:	$retry:			the value to check
# In:   $refresh:		the refresh value from the same SOA record
#
sub check_retry
{
    my $self = shift;
    my ( $problem_entity, $retry, $refresh ) = @_;

    $self->dprint( "Check SOA retry", 3 );

    if ( ! defined $retry ) {
	$self->( 'PROBLEM_SOA_RECORD_FIELD_EMPTY', $problem_entity, 'retry' );
    }
    elsif ( $self->check_enabled( 'CheckRetryRange' ) ) {

	my $min_fraction = $self->conf( 'ParamRetryMinFractionRefresh' );
	my $max_fraction = $self->conf( 'ParamRetryMaxFractionRefresh' );

	if ( defined $refresh ) {

	    $self->dprint( "Checking retry ($retry) against refresh ($refresh)", 4 );

	    my $min_value = $min_fraction * $refresh;
	    my $max_value = $max_fraction * $refresh;

	    $self->dprint( "Lower boundary $min_value, upper boundary $max_value", 5 );

	    if ( $retry < $min_value || $retry > $max_value ) {

		$self->p( 'PROBLEM_RETRY_OUT_OF_RANGE', $problem_entity,
			  $retry, &seconds_to_descr_string( $retry ),
			  $min_fraction, $max_fraction, 
			  &seconds_to_descr_string($min_value),
			  &seconds_to_descr_string($max_value)
		);
	    }
	}
	else {
	    $self->dprint( "Not comparing retry to refresh: refresh is undefined" );
	}
    }
}

# Purpose:			check an SOA expire value
# In:	$problem_entity:	which key to record the problem under
# In:	$expire:		the value to check
# In:   $retry:		        the retry value from the same SOA record
#
sub check_expire
{
    my ( $self, $problem_entity, $expire, $refresh, $retry ) = @_;

    $self->dprint( "Check SOA expire", 3 );

    if ( ! defined $expire ) {
	$self->( 'PROBLEM_SOA_RECORD_FIELD_EMPTY', $problem_entity, 'expire' );
    }
    else {
	my $min = $self->conf( 'ParamExpireMin' );
	my $max = $self->conf( 'ParamExpireMax' );

	if ( $self->check_enabled( 'CheckExpireRange' ) ) {

	    $self->dprint( "Checking expire ($expire) against min $min, max $max", 4 );

	    $self->p( 'PROBLEM_EXPIRE_OUT_OF_RANGE', $problem_entity,
		      $expire, &seconds_to_descr_string( $expire ),
		      $min, $max
	    )
		if ( $expire < $min || $expire > $max );
	}

	if ( defined $refresh ) {

	    if ( $self->check_enabled( 'CheckExpireAgainstRefresh' ) ) {
		$self->dprint( "Checking expire ($expire) against refresh ($refresh)", 4 );

		$self->p( 'PROBLEM_EXPIRE_LESS_THAN_REFRESH', $problem_entity,
			  $expire, $refresh
		)
		    if ( $expire < $refresh );
	    }

	    if ( 
		 $self->check_enabled( 'CheckExpireAgainstRefreshRetrySum' ) &&
		 defined $retry 
	    ) {
		$self->dprint( "Checking expire ($expire) against sum of refresh, retry ($refresh, $retry)", 4 );
		$self->p( 'PROBLEM_EXPIRE_LESS_THAN_REFRESH_RETRY_SUM', $problem_entity,
			 $expire, $refresh, $retry 
		)
		    if ( $expire < ( $refresh + $retry ) );
	    }
	    else {
		$self->dprint( "Not checking expire against sum of refresh, retry, because retry is is undefined" );
	    }

	}
	else {
	    $self->dprint( "Not checking expire against refresh, refresh is undefined" );
	}
    }
}

# Purpose:			check an SOA minTTL value
# In:	$problem_entity:	which key to record the problem under
# In:	$minimum:		the value to check
#
sub check_minimum
{
    my ( $self, $problem_entity, $minimum ) = @_;

    $self->dprint( "Check SOA minimum", 3 );

    if ( ! defined $minimum ) {
	$self->( 'PROBLEM_SOA_RECORD_FIELD_EMPTY', $problem_entity, 'minTTL' );
    }
    elsif ( $self->check_enabled( 'CheckMinimumRange' ) ) {

	my $min = $self->conf( 'ParamMinimumMin' );
	my $max = $self->conf( 'ParamMinimumMax' );

	$self->dprint( "Checking minTTL ($minimum) against bounds $min, $max", 4 );

	if ( $minimum < $min || $minimum > $max ) {
	    $self->p( 'PROBLEM_MINIMUM_OUT_OF_RANGE', $problem_entity, 
		      $minimum, &seconds_to_descr_string( $minimum ),
		      $min, $max
	    );
	}
    }
}

# Purpose:			signal a problem which the program maintainer
#				should be made aware of.
# In:	$text:			description of the problem
# Out:	$exit_code (optional):	if defined, causes the program running to exit
#				with that code.
# SEF:				writes to STDERR
# SEF:				generates __DIE__ if $fatal_flag defined
#
sub program_error
{
    my ( $self, $text, $fatal_flag ) = @_;

    # Make up a stack trace

    my ($lvl, $stack_trace ) = ( 0, "" );

    while ( my ($package, $file, $line, $caller, @dummy ) = caller( $lvl++ ) ) {
        $stack_trace .= "\tPackage $package($file:line $line) called &$caller\n";
    }

    # print out error and trace

    if ( $fatal_flag ) {
	print  "$0: error: $text, continuing.";

    }
    else {

	print STDERR "$0: error: $text, continuing.";
	print STDERR "$stack_trace\n";

	$self->dprint( "Program error: '$text', non-fatal. Maintainer will be notified." );
    }
}

# Purpose: check the details returned about a nameserver
#
sub check_ns_details
{
    my $self = shift;

    $self->dprint( "Check NS name details", 0 );

    # Go through each unique name

    while (  my ( $ns_name, $ns_ref ) = each %{ $self->{'data'}{'ns'}{'byname'} } ) {

	my $ip = $ns_ref->{'ip'};

	$self->dprint( "Check $ns_name details", 1 );

	$self->check_hostname( $ns_name, $ns_name, 'this nameserver' );

	$self->check_ns_name_canonical( $ns_name, $ns_name )
	    if ( $self->check_enabled( 'CheckNameserverNameCanonical' ) );

	if ( 
	     $ip &&
	     (
	        $self->check_enabled( 'CheckNameserverHasReverse' ) ||
	        $self->check_enabled( 'CheckNameserverHasMatchingReverse' )
             )
	) {
	    $self->check_nameserver_reverse_mapping( $ns_name, $ip );
	}
    }
}

# Purpose:		Check that we were able to look up a valid IP
#			address for a nameserver (see &pick_an_ip)
# In:   $ns_name	Relevant nameserver
# In:   $ip		The address we got, or undef
#
#sub check_ns_had_ip
#{
#    my ( $self, $ns_name, $ip ) = @_;
#
#    $self->dprint( "Check we got an IP for $ns_name", 2 );
#    $self->p( 'PROBLEM_COULDNT_LOOKUP_NS_ADDRESS', $ns_name, $ns_name )
#	if ( ! $ip );
#    
#    return ( defined $ip );
#}




# Purpose:               enable or disable checks.
#                        
# In: $check_name:       the relevant configuration variable name
# Out:                   boolean yes/no
#
sub disable_check
{
    my ( $self, $check_name ) = @_;
    $self->dprint( "disabling $check_name", 5 );
    $self->{'config'}->$check_name(0);
}


sub enable_check
{
    my ( $self, $check_name ) = @_;
    $self->dprint( "enabling $check_name", 5 );

    $self->{'config'}->$check_name(1);
 }



# Purpose:               report whether a particular test is enabled in
#                        the config
# In: $check_name:       the relevant configuration variable name
# Out:                   boolean yes/no
#
sub check_enabled
{
    my ( $self, $check_name ) = @_;

    return ( $self->{'config'}->$check_name );
}

# Purpose:		check that the appropriate PTR record is set up
#			for an NS
# In:	$ns_name	name of the NS
# In:	$ip		IP we picked for it
# Comments:             if there are multiple PTR's set up, we don't care,
#			as long as *one* of them maps back to the NS name
#
# Lee 20001110
# Amended so that if we're checking a reverse zone, we don't complain if the
# address of the nameserver lies within the reverse zone and we can't find
# a PTR record. They obviously can't setup reverse for the NS IP if the
# relevant reverse zone doesn't exist yet.
# 
sub check_nameserver_reverse_mapping
{
    my ( $self, $ns_name, $ip ) = @_;

    $self->dprint( "Checking for PTR record mapping back to name", 2 );

    my $ns_ip = $self->{'data'}{'ns'}{'byname'}{$ns_name}{'ip'};

    my $zone = $self->{'data'}{'delegation'}{'zone'};

    # Lee 20010115
    # If the NS IP we're checking is in the zone we're checking,
    # don't do this check

    # Only for IPv4, 6 gets too complicated
    # Doesn't work for /16 reverses

    if ( &is_reverse_zone( $zone ) == 4 && &ip_is_ipv4( $ns_ip ) ) {

	# Turn the IP into a reverse zone by reversing it, adding in-addr.arpa
	# and chopping off the LHS

	my $reverse_ip = join '.', ( reverse split /\./, $ns_ip ), 'in-addr.arpa';

	$reverse_ip = &normalise_domain_name( $reverse_ip );

	$reverse_ip =~ s/^\d+\.//;

	# Compare to the zone we're testing

	if ( $reverse_ip eq $zone ) {
	    $self->dprint( "IP $ns_ip part of reverse zone we're checking, abort check", 3 );
	    return;
	}
    }

    # If we got any PTR RR's back...

    my $ptr_rrs;

    if ( &hash_key_exists( $self->{'data'}{'ns'}, 'byname', $ns_name, 'rrs', 'ptr_list' ) ) {
	$ptr_rrs = $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'ptr_list'};
    }

    my ( $found_count, $matched_flag ) = ( 0, 0 );


    if ( $ptr_rrs && @$ptr_rrs ) {

	my $no_ptr_records = scalar( @$ptr_rrs );

	
	$self->dprint( "Found $no_ptr_records PTR record(s), trying to match with hostname $ns_name", 3 );
	
	# Go through them to try and match with the hostname
	
	foreach my $ptr_rr ( @$ptr_rrs ) {

	    $found_count++;
	    
	    # Net::DNS often has problems with weird data
	    # with regard to PTR records, we better be sure that
	    # what we're about to look at is what we expect it
	    # to be
	    
	    if ( 
			  ( ! defined $ptr_rr ) || 
			  ( ref($ptr_rr) ne 'Net::DNS::RR::PTR' ) ||
			  ( ! defined $ptr_rr->ptrdname )
	     ) {
		$self->dprint( "Ignoring undefined or bad PTR reference..", 4 );
		next;
	    }
		
	    my $ptr_name = &normalise_domain_name( $ptr_rr->ptrdname );
	    
	    if ( ! defined $ptr_name ) {
		$self->dprint( "Ignoring PTR data undefined after normalisation for ". $ptr_rr->ptrdname, 4 );
		next;
	    }
   
	    $self->dprint( "ptr_name = $ptr_name", 4 );
	    
	    # If it's data section matches the NS name, bingo
	    
	    if ( $ptr_name eq $ns_name ) {
		$self->dprint( "OK" );
		$matched_flag = 1;
		last;
	    }
	}
    }
    
    # See what we got
    # Note: CheckNameserverHasMatchingReverse will not produce a problem
    # if there are no OK records found

    if ( $found_count == 0 ) {
	if ( $self->check_enabled( 'CheckNameserverHasReverse' ) ) {
	    $self->p( 'PROBLEM_NO_REVERSE_MAPPING', $ns_name, $ip, $ns_name );
	}
    }
    elsif ( ! $matched_flag ) {
	if ( $self->check_enabled( 'CheckNameserverHasMatchingReverse' ) ) {
	    $self->p( 'PROBLEM_WRONG_REVERSE_MAPPING', $ns_name, $ip, $ns_name );
	}
    }
}


sub check_rrset_ttl
{
    my ($self, $problem_entity, @rrset)=@_;
    return unless @rrset;
    $self->dprint( "Checking RRSET TTL for ".$rrset[0]->name." ".$rrset[0]->type, 2 );
    my $ttl=$rrset[0]->ttl;
    foreach my $rr (@rrset) {
	if ($rr->ttl != $ttl ){
	    $self->p('PROBLEM_RRSET_TTLS_NOT_EQUAL',$problem_entity,$ttl,$rr->ttl,$rrset[0]->string,$rr->string);
	    return;
	}
    }
}
    
# Purpose:		
#			
# In:	$ns_name	
#

sub check_nameserver_minimum_ttl
{
    my ( $self, $problem_entity, $ns ) = @_;
    $self->dprint( "Checking TTL of NS RR against minimum ttl", 2 );
    my $ttl=$ns->ttl;;
    
    if ( $ttl < $self->conf( 'ParamMinimumNameserverTTL' ) ) {
	
	$self->p( 'PROBLEM_NS_TTL_TOO_SHORT', $problem_entity,
		  $ttl , $self->conf( 'ParamMinimumNameserverTTL' )   
		  );
	
	
    }
    
}
		    

		    
sub check_ns_name_canonical
{
    my ( $self, $problem_entity, $ns_name ) = @_;

    $self->dprint( "Check nameserver name canonical", 2 );

    $self->p( 'PROBLEM_NS_NAME_NOT_CANONICAL', $problem_entity, $ns_name )
	if ( defined $self->{'data'}{'ns'}{'byname'}{$ns_name}{'rrs'}{'cname_list'} );
}

# Purpose:			apply domainname syntax restrictions to a name
# In:	$problem_entity:	which key to record the problem under
# In:	$name:			the name to check
# Out:  $problem_count:         number of problems found
#
sub check_domainname_syntax
{
    my ( $self, $problem_entity, $name, $domain_name_source ) = @_;

    my $problem_count = 0;

    # Total length of name

    if ( length $name > $self->conf( 'ParamMaxDomainNameLength' ) ) {
	$self->p( 'PROBLEM_DOMAIN_NAME_TOO_LONG', $problem_entity,
	     $name, $domain_name_source, length $name, $self->conf( 'ParamMaxDomainNameLength' )
	);
	$problem_count++;
    }

    # 0 length labels

    if ( $name =~ /\.\./ ) {
	$self->p( 'PROBLEM_DOMAIN_NAME_LABEL_EMPTY', $problem_entity, $name, $domain_name_source );
	$problem_count++;
    }

    # Labels too long

    my @labels = split /\./, $name;

    foreach my $label ( @labels ) {

	my $max_label_length =  $self->conf( 'ParamMaxDomainNameLabelLength' );

	if ( 
		 $label && 
		 ( length $label > $max_label_length )
	) {
	    $self->p( 'PROBLEM_DOMAIN_NAME_LABEL_TOO_LONG', $problem_entity,
		      $label, $name, $domain_name_source, 
		      length $label, $max_label_length
		      );
	    $problem_count++;
	}
    }

    return $problem_count;
}

# Purpose:			apply hostname character restrictions to a name
# In:	$problem_entity:	which key to record the problem under
# In:	$name:			the name to check
# Comments: 
#
# This check is only for the actual characters used, not for stuff
# like label length etc

sub check_hostname
{
    my $self = shift;

    my ( $problem_entity, $hostname, $where_found ) = @_;

    $self->dprint( "Checking $hostname (from: $problem_entity, $where_found) syntax", 2 );

    if ( $self->check_enabled( 'CheckHostnamesAgainstDomainNameSyntax' ) ) {

	$self->dprint( "Checking $hostname against domainname syntax", 3 );
	
	my $problem_count = $self->check_domainname_syntax( $problem_entity, $hostname, $where_found );

	if ( $problem_count == 0 ) {
	
	    $self->dprint( "Checking $hostname against hostname syntax as required", 3 );

	    my @labels = split ( /\./, $hostname );

	    # Check also for one label if everything else OK

	    $self->p( 
		  'PROBLEM_HOSTNAME_ONE_LABEL', $problem_entity,
		  $hostname, $where_found 
	    )
		if ( 
		     $self->check_enabled( 'CheckHostnamesHaveMultipleLabels' ) &&
		     ( @labels == 1 )
	        );

	    foreach my $label ( @labels ) {

		# Lee 20040128
		# Corrected ignoring of zero-length labels

		next if ( length $label == 0 );

		# Lee 20010831
		# Removed error for all-numeric labels

		# RFC952 stipulates [let]([let|digit|-]*)[let|dig]
		# RFC1123 relaxes to allow digit as first
		# Therefore all-numeric labels are not excluded. 
		
		# As far as I can see, RFC1912 is mistaken in it's claim
		# that 'Labels may not be all numbers,'. Since it's
		# only informational, I think the two standards-track
		# RFCs above should be adhered to.

		$self->p( 
			  'PROBLEM_HOSTNAME_LABEL_ILLEGAL_CHARACTERS', $problem_entity, 
			  $label, $hostname, $where_found
                )
		    if ( 
		      $self->check_enabled( 'CheckHostnamesForIllegalCharacters' ) &&
		      ( $label !~ /^[a-zA-Z0-9-]+$/ )
		    );
	
		$self->p( 
			  'PROBLEM_HOSTNAME_LABEL_ILLEGAL_START_OR_END', $problem_entity,
			  $label, $hostname, $where_found
                )
		    if ( 
			 $self->check_enabled( 'CheckHostnamesForIllegalStartEndCharacters' ) &&
			 ( 
			   $label !~ /^[a-zA-Z0-9]/ || 
			   $label !~ /[a-zA-Z0-9]$/
			 )
		    );
	    }
	}
    }
}



# Purpose:			Check if there is a "listening" MTA
# In:	$email_address		e-mail address as parsed from the rname
#                               including the substitution of the first
#                               non escaped dot into an @ sign.

# In;   $zone                   from which zone does the SOA originate
# In:	$problem_entity:	which key to record the problem under
# In:   $rname                  The original SOA field.
# Comments: 
#

# Assumes that it is not called if PROBLEM_RNAME_DUPLICATE_ZONE,
# PROBLEM_RNAME_HAS_AT_SIGN, PROBLEM_RNAME_BAD_SYNTAX and
# PROBLEM_REVERSE_DOMAIN_IN_RNAME have not been flagged.



sub check_rname_mail {
    my $self=shift;
    my $email_address=shift;
    my $zone=shift;
    my $problem_entity=shift;
    my $rname=shift;

    # e-mail cache related work.
    my $cachedir=$self->conf("ModuleMailCache");
    
    if (defined($cachedir) ){
	# REGEX will make it taint safe and remove any trailing /
	$cachedir =~ s!(.*?)/?$!$1!;

	$self->dprint( "MailCheck: is the rname cached?", 4 );    	
	if (-f "$cachedir/delcheck".md5_hex($rname)){

	    $self->dprint("MailCheck:The rname ($rname) address was cached previously, no Mail checks performed",1);	    return;
	    
	}
	$self->dprint( "MailCheck:The email address was not found in the cache", 5 );    	
    }

    # We allready did test the format so this can be kept simple.
    $email_address=~/@(.*)/;
    
    my $domain=$1;
    $self->program_error( "Domain could not be extracted from $email_address", 1 ) unless $domain;
    
    my $mailfrom=$self->conf("ModuleMailEnvelopeFrom");
    
    $self->program_error( "ModuleMailEnvelopeFrom is not configured\n".
			  "\thaving a somewhat valide email address for this parameter is a prerequisite\n".
			  "\nfor running these tests.", 1 ) unless $mailfrom;
    
    
    
    $self->dprint( "Checking the rname mail ($email_address) can be delivered (SOA from $zone)", 1 );    
    
    
    $self->dprint("Determining MX for $domain ",4);

    my @mxIP;  # Array that will be used to store the IP addresses for all the 
               # MX records
    
    my $res=$self->get_resolver("local",0);
    my ($mx_answers,$mx_err, undef, $mx_rcode)=$self->lookup_records($res,$domain,"MX",1);
    
    if ( $mx_err ) {
	$self->p( 'PROBLEM_MAILCHECK_MX_LOOKUP', $problem_entity, $domain,  $mx_err );
	return 0;
    }

    
    if ( ! @$mx_answers ) {
	my $reason=$mx_rcode;
	$reason.=" and no data in the answer section" if ($reason eq "NOERROR");
	$self->p( 'PROBLEM_MAILCHECK_MX_LOOKUP', $problem_entity, $domain, $reason );
	return 0;
    }
    
  MX:    foreach my $mx (@$mx_answers){
      $self->dprint("Determining IP for ".$mx->exchange,5);
      
      my ($a_answers,$a_err, undef, $a_rcode)=$self->lookup_records($res,$mx->exchange,"A",1);
      if ( $a_err ) {
	  $self->p( 'PROBLEM_MAILCHECK_A_LOOKUP', $problem_entity, $domain,  $a_err );
	  next MX;
      }
      if ( ! @$a_answers ) {
	  my $reason=$a_rcode;
	  $reason.=" and no data in the answer section" if ($reason eq "NOERROR");
	  $self->p( 'PROBLEM_MAILCHECK_A_LOOKUP', $problem_entity, $domain, $reason );
	  next MX;
      }

      foreach my $a ( @$a_answers){
	  push @mxIP, $a->address;
      }
      # next MX;
  }
# end of loop over MXs. @mxIP now contains all the addresses for all MTAs
    
    
    if (!@mxIP){
	$self->dprint( "Terminating Mailcheck, could not find an IP address for the MX for $domain", 1 );    
	return;
    }
    
    
    
    # uniquify the list of IP addresses.
    # Perl cookbook ch04_07
    my %seen = ();
    my @uniquemxIP = grep { ! $seen{$_} ++ } @mxIP;
    
    $self->dprint("IP addresses of MTAs to check:".join(", ",@uniquemxIP ),3);
    
    my $errors_occured=0;
    foreach my $ip (@uniquemxIP){	
	
	$errors_occured=1 if $self->mailcheck_check_mail($ip,
							 $email_address,
							 $mailfrom,
							 $zone,
							 $problem_entity);

	
    }

    if ( !$errors_occured && defined ($cachedir) ){
	$self->dprint( "Caching successful verification", 5 );    	
	my $path=$cachedir."/";
	# Note that if the rname does not match we just do not cache.

	# we do not care if we miss out on caching records because
	# the test is not fine enough. We do care for writing files
	# in place we do not want files.

	# In theory we would never get a positive result on this check
	# as the rname results from parsing a DNS packet which would
	# never ever have a empty label. 
	

	# Unique filename 
	# the md5_hex will provide us with a unique filename that is 
	# 
	
	my $hexpath.="delcheck".md5_hex($rname);
	$hexpath=~/(.*)/;   # trivial, always clean 
	$path.=$1;

	if ( open(FILE,"> $path")){
	    print FILE gmtime() . " $rname\n" ;

	    close (FILE);
	}else{

	    $self->program_error( "Could not open $path for caching" ) unless $domain;


	}

    }
    $self->dprint( "Done with mailchecks", 1 );    
    return;
    
}



# Purpose:			Get the IP address for a domain.
# In:   $domain                 Domain to do MX query from

# In:	$problem_entity:	the things which the problem relates to e.g. nameserver
#				name, set of nameserver names
# Out:  $ip                     String containing one of the IP addresses for the
#                               lowest preference MTA or "0" on failure.


sub get_ip_for_mx_of_domain {
    my $self=shift;
    my $domain=shift;
    my $problem_entity=shift;
    $self->dprint("Determining MX for $domain ",4);
    my $res=$self->get_resolver("local",0);
    my ($mx_answers,$mx_err, undef, $mx_rcode)=$self->lookup_records($res,$domain,"MX",1);

    # select the lowest preference from the available 

    if ( $mx_err ) {
	$self->p( 'PROBLEM_MAILCHECK_MX_LOOKUP', $problem_entity, $domain,  $mx_err );
	return 0;
    }


    if ( ! @$mx_answers ) {
	my $reason=$mx_rcode;
	$reason.=" and no data in the answer section" if ($reason eq "NOERROR");
	$self->p( 'PROBLEM_MAILCHECK_MX_LOOKUP', $problem_entity, $domain, $reason );
	return 0;
    }



    my $mx=$mx_answers->[0];
MX:    foreach my $mx (@$mx_answers){
	$self->dprint("Determining IP for ".$mx->exchange,5);
	
	my ($a_answers,$a_err, undef, $a_rcode)=$self->lookup_records($res,$mx->exchange,"A",1);
	# select the lowest preference from the available 
	if ( $a_err ) {
	    $self->p( 'PROBLEM_MAILCHECK_A_LOOKUP', $domain,  $a_err );
	    next MX;
	}
	if ( ! @$a_answers ) {
	    my $reason=$a_rcode;
	    $reason.=" and no data in the answer section" if ($reason eq "NOERROR");
	    $self->p( 'PROBLEM_MAILCHECK_A_LOOKUP', $problem_entity, $domain, $reason );
	    next MX0;
	}
	
	return $a_answers->[0]->address;
    }
}



#
# Purpose:			Setup a socket to comunicate with an MTA.
# In:   $ip                     
# In:   $address                The email address to test
# In:   $mailfrom               The email address to be used in the from envelope
# In:	$zone                   The zone from which the SOA RR originated
# In:	$problem_entity:	the things which the problem relates to e.g. nameserver
#				name, set of nameserver names
# Out:                          Returns 0 on success non-0 when PROBLEMs occured

# Comments: This code implements the "control" messages.
# Code based on a contribution by Patrik Faltstrom


sub mailcheck_check_mail {
    my $self=shift;
    my ($ip, $address, $mymailaddress, $zone, $problem_entity) = @_;
    my ($e,$r_msg)=(0,"not defined");
    my ($error,$ret_msg)=(0,"not defined");
    
    my $dmsg='rname('.$zone.')->'.$address.', MTA: '.$ip;
    $self->dprint("MailCheck: checking rname from SOA from $zone ($address)  at MTA: $ip",1);    
    

    no strict ;
    my $timeout=$self->conf( 'ModuleMailTimeoutSocket');
    $timeout=5 unless defined $timeout;
    $timeout=~/(\d+)/;
    my $untainted_timeout=$1;
    $ip=~/^(\d+\.\d+\.\d+\.\d+)$/;
    my $untainted_ip=$1;
    

    # Open a socket to port 25 of the suposed MTA
    $self->dprint("MailCheck: timout on socket set to $timeout",5);
    my $s = IO::Socket::INET->new(PeerAddr => $untainted_ip,
				  PeerPort => 25,
				  Proto    => "tcp",
				  Type     => SOCK_STREAM,
				  Timeout  => $untainted_timeout);
    
    use strict ;
    if(!$s) {
	# Bail out completely if no connection was received.
	$self->dprint("Could not open connection to ${ip}:25 ($@)",1);
	$self->p('PROBLEM_MAILCHECK_CONNECTION_FAILED',$problem_entity,$ip,$address,$@);
	return 1;
	
    }

    my $myhostname = gethostbyaddr($s->sockaddr,&AF_INET);

    # We do not want to throw a fit if gethostbyaddr fails. We do
    # report it though and to STDERR     
    if (! defined ($myhostname)){
	$self->program_error("Could not determine hostname using gethostbyaddr");
	return(0);  # Return without errors. This is actually a program error.
    }

    ($e,$r_msg)=$self->mailcheck_smtp_msg($s,"",$dmsg,$problem_entity);
    goto ERROR if $e;
    

    # First try EHLO, if that fails because of old SMTP, reset (SMTP
    # RST) the connection, while keeping the TCP connection intact
    # and try HELO

    my $hello = "EHLO";
  HELLO:
    ($e,$r_msg)=(0,"");
    
    ($e,$r_msg) = $self->mailcheck_smtp_msg($s,$hello . " $myhostname",$dmsg,$problem_entity);
    if($e) {
	if($hello eq "EHLO") {
	    $self->dprint("MailCheck: ($dmsg) Server does not support $hello",5);
	    ($e,$r_msg)=(0,"");
	    ($e,$r_msg)=$self->mailcheck_smtp_msg($s,"RSET",$dmsg,$problem_entity);
	    goto ERROR if $e;
	    $hello = "HELO";
	    goto HELLO;
	} else {
	    $self->dprint("MailCheck: ($dmsg) Server does not support RSET after EHLO",5);
	    goto ERROR;
	}
    }
    

    # Trying to establish a connection with an empty mail envelope, if that 
    # fails RSET the connection and try again with the mailaddress provided
    # as 3rd argument.

    my $mail_from = "<>";
  FROM:
    ($e,$r_msg)=(0,"");
    ($e,$r_msg) = $self->mailcheck_smtp_msg($s,"MAIL FROM: " . $mail_from,$dmsg,$problem_entity);
    if($e) {
	if($mail_from eq "<>") {
	    $self->dprint("MailCheck:($dmsg) Server does not accept empty envelope from",1);
	    ($e,$r_msg)=$self->mailcheck_smtp_msg($s,"RSET",$dmsg,$problem_entity);
	    goto ERROR if $e;;
	    $mail_from = "<" . $mymailaddress . ">";
	    goto FROM;
	} else {
	    goto ERROR;
	}
    }
    ($e,$r_msg)=(0,"");
    ($e,$r_msg)=$self->mailcheck_smtp_msg($s,"RCPT TO: <" . $address . ">",$dmsg,$problem_entity);
    goto ERROR if $e;
    
    goto QUIT;
    
  ERROR:
    
    $self->dprint("MailCheck: Error: $e, $r_msg",5);
    
    $self->p('PROBLEM_MAILCHECK_SMTP_FAILURE',$problem_entity,$ip,$address,$r_msg);
    $self->dprint("MailCheck: ($dmsg) Closing socket",5);
    close($s) if $s;
    $s = 0;
    
    
  QUIT:
    # Failure to quit will not generate a problem code.
    if($s) {
	$self->dprint("MailCheck: ($dmsg) Trying to send a QUIT message",5);
	
	($e,$r_msg)=$self->mailcheck_smtp_msg($s,"QUIT",$dmsg,$problem_entity) ;
	$self->dprint("MailCheck: ($dmsg) Closing Socket",5);
	close($s) if $s;
    }
    
    $self->dprint("MailCheck: rname for $problem_entity  $address  at $dmsg",5);
    return $e;
};





#
# Purpose:			Control input and output of SMTP messages to a socket
# In:   $s                      An instance of a IO::Socket
# In:   $msg                    the message to be send
# In:   $dmsg                   text used in debuging messages (usually IP)
# In:	$problem_entity:	                which key to record the problem under
# Out:  $err                    returns 1 on failure, 0 on success
# Out:  $msg:                   The SMTP message  or an "undef"
# Comments: calls mailcheck_read_msg and mailcheck_write_msg.
# 
# Code based on a contribution by Patrik Faltstrom


sub mailcheck_smtp_msg {
    my $self=shift;
    my($s,$msg,$dmsg,$problem_entity) = @_;
    
    my $timeout=$self->conf( 'ModuleMailTimeoutRead' ) ;
    
    if(length($msg) > 0) {
	$self->mailcheck_write_msg($s,$msg,$timeout,$dmsg,$problem_entity) or return 0;
    }
    
    my ($error,$response) = $self->mailcheck_read_msg($s,$timeout,$dmsg,$problem_entity,$msg) or return 0;
    return (1,$response) if ($error);

    # Any non 2?? return code sets the first element of the return array to be false
    if(substr($response,0,1) ne "2") {
	$self->dprint("MailCheck: ($dmsg) Not accepted: ${msg}",5);
	return (1,$response);
    }
    return (0,$response);
}



#
# Purpose:			Read SMTP messages from a socket
# In:   $socket                 An instance of a IO::Socket
# In:   $timeout                Timeout while waiting for an answer
# In:   $dmsg                   text used in debuging messages (usually IP)
# In:	$problem_entity:	                Zone name for which this test is performed
# In:   $msg                    the message written to the socket before calling
#                               this read, used for error reporting
# Out:  $error                  returns 1 on failure, 0 on success
# Out:  $response:                   The SMTP message response or an "undef"
# 
# Code based on a contribution by Patrik Faltstrom


sub mailcheck_read_msg {
    my $self=shift;
    my ($socket, $timeout, $dmsg, $problem_entity,$msg) = @_;
    $msg="" unless defined ($msg);
    $socket->blocking(0);
    my $sel = new IO::Select($socket);
    
    my $response = "";
    my $separator = "-";
    my $line;
    while($separator eq "-") {

	my @ready = $sel->can_read($timeout);
	if(!@ready) {
	    $self->dprint("MailCheck: ($dmsg) Timeout when waiting for socket for reading",5);
	    return (1,"Timout while waiting for read");
	}
	my $nread = sysread($socket,$response,1024);
	if ( !defined($nread)) {
	    $self->dprint("MailCheck: ($dmsg), error reading from socket.",5);
	    return (1,"error reading from socket");
	}
	if (! $nread){
	    $self->dprint("MailCheck: MTA responded with empty message.",5);
	    return (1,"MTA returned an empty response after $msg");
	}	    
	$response =~ s/\015\012/!/g;
	my @lines = split('!',$response);
	foreach $line (@lines) {
	    $self->dprint("MailCheck: ($dmsg) < $line",5);
	    $line=~s/^[0-9][0-9][0-9]([ -]).*$/$1/;
	    $separator=$line;
	}
    }
    
#    $response =~ s/^([0-9][0-9][0-9]).*$/$1/;
    return (0,$response);
}



#
# Purpose:			Write SMTP messages to a socket
# In:   $socket                 An instance of a IO::Socket
# In:   $msg                    Message to be written to the socket.
# In:   $timeout                Timeout while waiting for an answer
# In:   $dmsg                   text used in debuging messages
# In:	$problem_entity:	which key to record the problem under
# Comments: calls mailcheck_read_msg and mailcheck_write_msg.
# 
# Code based on a contribution by Patrik Faltstrom


sub mailcheck_write_msg {
    my $self=shift;
    my ($socket, $msg, $timeout, $dmsg,$problem_entity) = @_;
    
    $socket->blocking(0);
    my $sel = new IO::Select($socket);
    
    $msg = $msg . "\015\012";
    my $offset = 0;
    my $len = length($msg);
    my @ready;
    my $written;
    
    while($len) {
	@ready = $sel->can_write(5);
	if(!@ready) {
	    $self->dprint("MailCheck: $problem_entity ($dmsg) Timeout when waiting for socket for writing",5);
	    return 0;
	}
	$written = syswrite($socket, $msg, $len, $offset);
	if(! defined( $written )) {
	    $self->dprint("MailCheck: $problem_entity ($dmsg) Error when writing to socket: $!",5);
	    return 0;
	}
	$offset += $written;
	$len -= $written;
    }
    chop $msg;
    chop $msg;
    $self->dprint("MailCheck: ($dmsg) > $msg",5);
    return 1;
}





# Get maname 
#
sub extract_usable_mname_from_soas
{
    my ( $self, $ip ) = @_;

    $self->dprint( "Extract usable mnames for $ip", 1 );

    my $ip_zone_data_ref = $self->{'data'}{'zone'}{$ip};

    my $soa_rrs = $ip_zone_data_ref->{'rrs'}{'soa_list'};

    my $ns_names_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );

    my $mname_to_return;

    # Ignore nameservers we couldn't get an SOA for
	
    if ( ! $soa_rrs || ! @$soa_rrs ) {
	$self->dprint( "$ip ($ns_names_text): ignoring, couldn't get an SOA.", 2 );
    }
    
    # Ignore nameservers with multiple SOA's
    
    elsif ( @$soa_rrs > 1 ) {
	$self->dprint( "$ip ($ns_names_text): ignoring, got multiple SOA's.", 2 );
    }
	
    # Ignore nameservers that we couldn't zone transfer an SOA, or
    # query for the 'SOA' query didn't give an AA answer
    
    elsif ( ! $ip_zone_data_ref->{'aa_flag'} ) {
	$self->dprint( "$ip ($ns_names_text): ignoring havn't got authoritative SOA", 2 );
    }
    else {
	my $soa_rr = $soa_rrs->[0];

	# Lee 20010723
	# Ignore empty mname field
    
	if ( $soa_rr->mname ) {
	    
	    # We have to normalise it, because otherwise if we compare
	    # two that are identical apart from case differences we'll
	    # get the wrong result
	    
	    $mname_to_return = &normalise_domain_name( $soa_rr->mname );

	    $self->dprint( "Got: $mname_to_return", 2 )
	}

	# Empty mname
	
	else {
	    $self->dprint( "$ip ($ns_names_text): ignoring empty mname field in SOA", 2 );
	}
    }

    return $mname_to_return;
}

############################## PRIVATE CLASS METHODS ######################################

sub read_configs
{
    my ( $config_ref, @config_specs ) = @_;

    foreach my $config_spec ( @config_specs ) {

	# If a config spec is a hash, treat keys as config variables and
	# values as settings

	if ( ref( $config_spec ) eq 'HASH' ) {
	    while ( my ( $config_var, $config_value ) = each %$config_spec ) {
		$config_ref->$config_var( $config_value );
	    }
	}

	# Otherwise it's a file, check exists, is a file, is readable

	elsif ( ! -e $config_spec ) {
	    &add_error( "$config_spec doesn't exist" );
	}
	elsif ( ! -f $config_spec ) {
	    &add_error( "$config_spec not a file" );
	}
	elsif ( ! -r $config_spec ) {
	    &add_error( "$config_spec not readable" ); 
	}
	else {
	    $config_ref->file( $config_spec );
	}
    }

    &test_config( $config_ref )	
	if ( ! &errors );
    
    return ( &errors );
}

sub initialise_template
{
    my $self = shift;

    my $parser = new XML::DOM::Parser;


    # Maybe config changed and we're reading a new template,
    # better dispose of the old one

    $self->{'template'}->dispose
	if ( defined $self->{'template'} );


    # Read in new one

    eval {
	$self->{'template'} = $parser->parsefile( $self->conf( 'ModuleTemplatePath' ) );
    };

    # If we managed to parse it, test the problem codes/arguments it contains

    if ( $@ ) {
	&add_template_error( $self->conf( 'ModuleTemplatePath' ), "parse error: $@" );

    }
    else {
	$self->test_template;
    }
}

# Purpose:            standardize format of errors produced when parsing a template file
# In: $file:          path to the file we found an error in
# In: $error_message: problem found
#
sub add_template_error
{
    my ( $file_path, $error_message ) = @_;

    &add_error( join ': ', $file_path, $error_message );
}

# Purpose:            handler for errors that AppConfig throws when we
# In: $error_message: the error
#
sub appconfig_error_handler
{
    my $error_message = shift;

    &add_error( $error_message );
}

# Purpose:         test that the config user has specified is valid
# In: $config_ref: reference to an AppConfig object with the settings
#                  to test
#
sub test_config
{
    my $config_ref = shift;

    # Directives which should have an unsigned integer value

    my @uint_directives = qw(
	ModuleDiagnosticLevel 
	ModuleQueryRemoteDefaultPortNumber ModuleQueryRemoteRetry 
	ModuleQueryRemoteRetrans ModuleQueryRemoteUDPTimeout

	ModuleQueryLocalPortNumber ModuleQueryLocalRetry 
	ModuleQueryLocalRetrans ModuleQueryLocalUDPTimeout

	ParamMaxDomainNameLength ParamMaxDomainNameLabelLength
	ParamMinimumNumberNameservers
	ParamFirstReasonableSerialYear ParamSerialFutureAllowanceHours
	ParamRefreshMin ParamRefreshMax
	ParamExpireMin ParamExpireMax
	ParamMinimumMin ParamMinimumMax	 
	ModuleMailTimeoutSocket ModuleMailTimeoutRead 

    );

    # Check that each directive which should contain a uint actually does

    foreach my $uint_directive ( @uint_directives ) {
	my $value = $config_ref->$uint_directive;

	if ( $value !~ /^\d+$/ ) {
	    &add_error( "$uint_directive ($value): must be an unsigned integer" );
	}
    }

    # Directives which should have a float value

    my @float_directives = qw( ParamRetryMinFractionRefresh ParamRetryMaxFractionRefresh );

    foreach my $float_directive ( @float_directives ) {
	my $value = $config_ref->$float_directive;

	if ( $value !~ /^\d+(?:\.\d+)?$/ ) {
	    &add_error( "$float_directive ($value): must be a float" );
	}
    }

    # ModuleTemplatePath: check specified, exists, a file, is readable

    my $to_test = $config_ref->ModuleTemplatePath;

    if ( $to_test ) {
	if ( ! -e $to_test ) {
	    &add_error( "ModuleTemplatePath: ($to_test): doesn't exist" );
	}
	elsif ( ! -f $to_test ) {
	    &add_error( "ModuleTemplatePath: ($to_test): not a file" );
	}
	elsif ( ! -r $to_test ) {
	    &add_error( "ModuleTemplatePath: ($to_test): not readable" );	
	}
    }
    else {
	&add_error( "ModuleTemplatePath: must be specified" );
    }

    # ModuleDiagnosticFormat: txt or html

    $to_test = $config_ref->ModuleDiagnosticFormat;

    if ( 
	 $to_test ne 'txt' && 
	 $to_test ne 'html'
    ) {
	&add_error( "ModuleDiagnosticFormat ($to_test): must be 'txt' or 'html'" );
    }

    # ParamProhibitedNameserverList/ParamRequiredNameserverList: no common elements

    foreach my $prohibited_server ( @{$config_ref->ParamProhibitedNameserverList} ) {
	foreach my $required_server ( @{$config_ref->ParamRequiredNameserverList} ) {
	    if ( &normalise_domain_name( $prohibited_server ) eq &normalise_domain_name( $required_server ) ) {
		&add_error( "ParamProhibitedNameserverList ($prohibited_server): server may not also occur in ParamRequiredNameserverList" );
	    }
	}
    }


    if ( ! (grep {  $config_ref->ParamRequiredNameserverSpec eq $_  } ( 'all', 'at-least-one', 'exactly-one' )) ) {
	&add_error( "ParamRequiredNameserverSpec: must be 'all', 'at-least-one' or 'exactly-one'" );

    }
    
    
    my $mail_envelope = $config_ref->ModuleMailEnvelopeFrom;
    if (defined $mail_envelope){
	# This uggly regexp might not be perfect but note that we only test
	# a configured e-mail address. 
	if ( $mail_envelope !~ /^[a-z0-9]([a-z0-9]?([\w\-]+[a-z0-9]))*(\.([a-z0-9]?[a-z0-9][\w\-]+[a-z0-9]))*@[\w\-]+\.([\w\-]+\.)*[a-z]{2,6}$/i ) {
	    &add_error( "ModuleEnvelopeFrom: ($mail_envelope): does not seem to be a valid e-mail address" );
	}
    }
    



    # ModuleMailCache: if this value is defined the directory must
    # be writeable

    my $mail_cache = $config_ref->ModuleMailCache;
    if ( $mail_cache ) {

	if ( ! -e $mail_cache ) {
	    &add_error( "ModuleMailCache: ($mail_cache):doesn't exist" );
	}
	elsif ( ! -d $mail_cache ) {
	    &add_error( "ModuleMailCache: ($mail_cache): is not a directory" );
	}
	elsif ( ! -w $mail_cache ) {
	    &add_error( "ModuleMailCache: ($mail_cache): is not writable" );	
	}


    }


    # ModuleUseIPv6: if config says to use it, but we don't have the
    # needed libraries

    if ( $config_ref->ModuleUseIPv6 && ! &ipv6_available ) {
	&add_error( "ModuleUseIPv6: IPv6 libraries are not available" );
    }
}

# Purpose:		given a reference to a list of L<Net::DNS::RR> records, 
#			return a list of names associated with any NS records
#			found in the RR's. 
# In:	$rr_list:	reference to a list of Net::DNS::RR's references
# Out:	@ns_name_list:	NS names from any NS RR's which were found in $rr_list
#
sub ns_rrs_to_ns_names
{
    my $ns_rrs = shift;

    my @nsnames;

    if ( defined $ns_rrs ) {
	foreach my $ns_rr ( @$ns_rrs ) {
	    push @nsnames, &normalise_domain_name( $ns_rr->nsdname )
		if ( $ns_rr->type eq 'NS' );
	}
    }

    return @nsnames;
}

# Purpose:		check that a prefix is one that is suitable
#                       for use as a nameserver IP.
# In:	$prefix:	prefix to test
# Out:	$reason:	reason it's bad, or undef if it's good
# Comments:             must be syntactically valid prefix at this point
#
sub is_unusable_prefix
{
    my $self = shift;

    my $prefix = shift;

    my $protocol_version = get_ip_version( $prefix );

    my $reason;

    if ( $protocol_version == 4 ) {

	my @octets = split /\./, $prefix;

	if ( @octets != 4 ) {
	    $reason = "it isn't 4 octets";
	}
# Dec 2004: removed these two tests as they are silly mechanism to determine
# network and broadcast address.
#	elsif ( $octets[3] == 0 ) {
#	    $reason = 'the final octet is 0';
#	}
#	elsif ( $octets[3] == 255 ) {
#	    $reason = "it's a broadcast address";
#	}
	elsif ( $prefix eq '127.0.0.1' ) {
	    $reason = "it's the loopback address";
	}
	elsif ( 
		( $octets[0] == 10 ) ||
		(
		   ( $octets[0] == 172 ) &&
		   ( $octets[1] >= 15 ) &&
		   ( $octets[1] <= 31 )
		 ) ||
		(
		   ( $octets[0] == 192 ) &&
		   ( $octets[1] == 168 )
		)
	) {
	    $reason = "it's from private address space (RFC 1597)";
	}
    }
    elsif ( $protocol_version == 6 ) {
	my $ip_obj = new Net::IP( $prefix );

	if ( defined $ip_obj ) {

	    my $bin_rep = $ip_obj->binip;
	    
	    if ( defined $bin_rep ) {
		# Global unicast is everything which isn't one of the below
		# (RFC3513)
	    
		if ( $bin_rep =~ /^0{127}0$/ ) {
		    $reason = "it's the unspecified address";
		}
		elsif ( $bin_rep =~ /^0{127}1$/ ) {
		    $reason = "it's the loopback address";
		}
		elsif ( $bin_rep =~ /^11111111/ ) {
		    $reason = "it's a multicast address";
		}
		elsif ( $bin_rep =~ /^1111111010/ ) {
		    $reason = "it's a link-local address";
		}
		elsif ( $bin_rep =~ /^1111111011/ ) {
		    $reason = "it's a site-local address";
		}
	    }
	    else {
		$self->program_error( "binary representation for IP $prefix undefined", 0 );
	    }
	}
    }

    return $reason;
}

# Purpose:		split off a scalar holding a pattern like nsname[:port]
#			into it's name and port
# In:	$ns_and_port
# Out:	$name		the name alone
# Out:	$port		the port
#
sub split_ns_and_port
{		
    my ( $ns_and_port, $default_port ) = @_;

    $ns_and_port =~ /
	            ^
			([^:]*)		# nameserver name
			:?		# possible divider for port
			(.*)		# port number
		    $
		  /x;

    my ( $name, $port ) = ( $1, $2 );

    $port = $default_port if ( ( ! defined $port ) || ( ! length $port ) );

    return ( $name, $port );
}

# Purpose:		decide whether an IP prefix is syntactically valid
# In:	$prefix:        the prefix
# Out:	                boolean yes/no
#
sub is_valid_prefix
{
    my $prefix = shift;

    my $valid_flag = 0;

    if ( defined $prefix ) {

	my $ip_obj = new Net::IP( $prefix );

	$valid_flag = ( defined $ip_obj );
    }

    return $valid_flag;
}

# Purpose:             join the members of a list up into a scalar to make
#                      code prettier
#
sub list_text
{
    return ( join ', ', @_ );
}

# Purpose:			return a list of nameserver names corresponding to a list
#				of nameserver IP's
# In:   $ip_to_name_map:	hash of arrays of names corresponding to each IP
# In:	$ips:			the IP's to convert
# Out:	@names:			the names corresponding to the IP's.
#
sub ip_list_to_name_list
{
    my ( $ip_to_name_map, @ips ) = @_;
    
    my @names;

    foreach my $ip ( @ips ) {
	push @names, @{ $ip_to_name_map->{$ip}{'name_list'} };
    }

    return @names;
}

# Purpose:              remove the taint on a scalar which may cause problems
#			when the module is used with -T flag. Of course, we
#			should think carefully before using this....
# In:	$to_clean:	the tainted thing
# Out:	$1:		the fresh sparkling version
#
sub remove_taint
{
    my $to_clean = shift;

    $to_clean =~ /^(.*)$/;

    return $1;
}

# Purpose:		check that a key exists in a hash without introducing
#                       new (intermediate) keys into the structure
# In:	$ref	        reference to the hash
# In:	@keys           from left to right, the shallowest to deepest key
#                       which should exist
#
sub hash_key_exists
{
    my ( $ref, @keys ) = @_;

    my $last_key = pop @keys;
    
    foreach my $key ( @keys ) {

	# Don't descend further if there's nowhere to go,
	# otherwise perl will create the key for you, being
	# friendly and all

	if (
            ( ref( $ref ) eq 'HASH' ) &&
            ( exists $ref->{$key}  )
	) {
	    $ref = $ref->{$key};
	}
	else {
            last;
	}
    }

    return ( exists $ref->{$last_key} );
}

sub check_rrsig_set 
{

  my $self = shift;

  $self->dprint( "Checking if server responds with RRSIG set in authority section", 0 );

  while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

  my $names_for_ip_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );

    #unless (exists $self->{'data'}{'zone'}{$ip}{'rrs'}{'any_rrsig_list'}) {
    unless (exists $self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_rrsig_list'}) {
      $self->p('PROBLEM_NOT_ALL_SERVERS_RETURN_DNSSECDATA', $names_for_ip_text, $names_for_ip_text);
    }

  }
  
}

sub check_dnskey_set
{
  my $self = shift;

  my (%keys_listed, %keys_not_listed);

  $self->dprint( "Checking consistency of DNSKEY lists returned", 0 );

  while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

    my $names_for_ip_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );

    foreach my $key_rr (@ { $self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_list'} } ) {
      $keys_listed{$key_rr->string()}{$ip} = $names_for_ip_text;
    }

  }

  foreach my $key (keys %keys_listed) {
    
    while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

       my $names_for_ip_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );
      
       unless ( exists $keys_listed{$key}{$ip}) {
         $keys_not_listed{$key}{$ip} = $names_for_ip_text;
       }

    }

  }

  # go through 'not listed' list

  foreach my $key (keys %keys_not_listed) {

     my @listed_names; 
     my @not_listed_names;

     foreach my $ip (sort keys %{ $keys_listed{$key}} ) {
       push @listed_names, $keys_listed{$key}{$ip};
     }
     foreach my $ip (sort keys %{ $keys_not_listed{$key}} ) {
       push @not_listed_names, $keys_not_listed{$key}{$ip};
     }

     $self->p('PROBLEM_INCONSISTENT_DNSKEY_LISTING', 'set',
     $key,
     &list_text (sort keys %{ $keys_listed{$key}} ),
     &list_text (@listed_names),
     &list_text (sort keys %{ $keys_not_listed{$key}} ),
     &list_text (@not_listed_names),
              );
  }
}

sub check_dnskey_for_ds 
{
  my $self = shift;
  
  my @ds = @_;

  my %verified;

  my %tagged; #( keytag => (ds => '', dnskey => '') ) 

  foreach my $key (@ds) {
    # $key contains RDATA from DS records. We add zone name and type to create key object.
    my $dsrr = new Net::DNS::RR($self->{'data'}{'delegation'}{'zone'}." DS ".$key);
    if ($dsrr) {

      while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

        foreach my $keyrr (@ { $self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_list'} } ) {
          if ($dsrr->algorithm() eq $keyrr->algorithm() &&
              $dsrr->keytag() eq $keyrr->keytag() ) {

            $verified{$dsrr->keytag()} ++;
            $tagged{$dsrr->keytag()}{'ds'} = $dsrr;
            $tagged{$dsrr->keytag()}{'dnskey'} = $keyrr;
          }
        }

      }
     unless ( exists $verified{$dsrr->keytag()} ) {
      $self->p('PROBLEM_NO_MATCHING_DNSKEY_FOR_DS', 'set', $dsrr->rdatastr());
     }
    }
  }

  return (\%tagged);
}

sub check_sep_flag
{
  my $self = shift;

  my ($tagged) = @_;

  my $sep_flag_used = 0;
  # check if any keys have SEP flag set
  while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

    foreach my $keyrr (@ { $self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_list'} } ) {
      if (($keyrr->flags() & 1) == 1 ) {
        $sep_flag_used = 1;
      }
    }

  }

  if ($sep_flag_used == 1) {
    foreach my $keytag (keys %{$tagged} )  {
      if (($tagged->{$keytag}{'dnskey'}->flags() & 1 ) == 0) {
        # issue a WARNING, sep flag is 0
        $self->p('PROBLEM_DS_POINTING_TO_NON_SEP_DNSKEY','set',
               $tagged->{$keytag}{'ds'}->rdatastr(), 
               $tagged->{$keytag}{'dnskey'}->rdatastr(),
               $keytag
               ) if ($self->check_enabled('CheckDSPointingToNonSEPDNSKEY') );
      }
    }
  }
  else {
    $self->p('PROBLEM_SEP_FLAG_NOT_USED', 'set') if ($self->check_enabled('CheckSepFlagNotUsed'));
  }
}

sub check_dnskey_rrsig
{

  my $self = shift;

  my ($tagged) = @_;

  foreach my $keytag (keys %{ $tagged } )  {

    # does this key has corresponding RRSIG record?
    while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

      my %verified;

      my $names_for_ip_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );

      foreach my $rrsigrr (@ { $self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_rrsig_list'} } ) {
        if ($rrsigrr->keytag() eq $keytag ) {
          $verified{$keytag} ++;
        }
      }
      unless ( exists $verified{$keytag} ) {
      $self->p('PROBLEM_SEP_KEY_NOT_SELF_SIGNED',$names_for_ip_text,
                $tagged->{$keytag}{'ds'}->rdatastr(),
                $tagged->{$keytag}{'dnskey'}->rdatastr(),
                $keytag
               ) if ($self->check_enabled('CheckSepKeyNotSelfSigned'));
      }

    }

  }

}

sub check_ds_hashes
{

  my $self = shift;
  my ($tagged) = @_;

  foreach my $keytag (keys % {$tagged} ) {

    if ($tagged->{$keytag}{'ds'}->digtype() != 1) {
      $self->p('PROBLEM_DS_DIGEST_ALGORITHM_NOT_KNOWN', 'set', 
               $tagged->{$keytag}{'ds'}->rdatastr()) if ($self->check_enabled ('CheckDSDigestAlgorithmNotKnown'));
    }
    else {
      unless ($tagged->{$keytag}{'ds'}->verify($tagged->{$keytag}{'dnskey'}) == 1) {
        $self->p('PROBLEM_DS_AND_DNSKEYRR_DO_NOT_MATCH', 'set',
                  $tagged->{$keytag}{'ds'}->rdatastr(),
                  $tagged->{$keytag}{'dnskey'}->rdatastr()
                 ) if ($self->check_enabled ('CheckDSAndDNSKEYRRDoNotMatch'));
      }
    }
  }

}

sub convert_string_to_time
{
  # no error checking here, relying on Net::DNS::RR
  my $string = shift;
  my @date = ($string =~ /(\d{4})(\d{2})(\d{2})(\d{2})(\d{2})(\d{2})/);
  $date[1] --; # in timelocal months are 0..11
  return timelocal(reverse @date);
}

sub check_signature_validity_soa
{
  my $self = shift;

  my %verified;

  while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

    foreach my $rrsigrr (@{$self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_rrsig_list'} } ) {
      unless (exists $verified{$rrsigrr->string()} ) {

      $verified{$rrsigrr->string()} = 1;
      my ($ins, $exp) = (convert_string_to_time($rrsigrr->siginceptation()),
                         convert_string_to_time($rrsigrr->sigexpiration())
                         );

      if ($rrsigrr->orgttl() > ( $exp - $ins ) /2 ) {
        $self->p('PROBLEM_TTL_LARGE_FRACTION_OF_SIGNATURE_VALIDITY', 'set',
                 $rrsigrr->orgttl(),
                 ($exp - $ins),
                ) if ($self->check_enabled('CheckTtlLargeFractionOfSignatureValidity'));
      }

      if ( $ins + 4*(($exp-$ins)/5) < time()) {
        $self->p('PROBLEM_SIGNATURE_VALIDITY_NEARS_EXPIRATION','set',
                $rrsigrr->keytag(),
                $rrsigrr->siginceptation(),
                $rrsigrr->sigexpiration()
                ) if ($self->check_enabled('CheckSignatureValidityNearsExpiration'));
      }

     }

    }

  }

}

sub find_dnskey
{

  my ($tag, $dnskey_list) = @_;
  if ($dnskey_list) {
    foreach my $dnskeyrr (@{$dnskey_list}) {
      if ($dnskeyrr->keytag() eq $tag) {
        return $dnskeyrr;
      }
    }
  }
  return undef;
}

sub validate_chains_of_trust
{
  my $self = shift;
  my ($tagged) = @_;

  my $validated = 0;
  my $terminated = 0;

  my %impl_algorithm = ('1' => '', '3' => '', '5' => ''); 

  foreach my $keytag ( %{ $tagged} ) {

    while ( my ( $ip, $ip_ref ) = each %{ $self->{'data'}{'zone'} } ) {

      my $names_for_ip_text = &list_text( @{ $self->{'data'}{'ns'}{'byip'}{$ip}{'name_list'} } );

      foreach my $rrsigrr (@{$self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_rrsig_list'} } ) {

        if ($rrsigrr->keytag() eq $keytag) {

          foreach my $soarrsigrr (@{$self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_rrsig_list'} } ) {

            my $broken = 0;
            my $algorithm_not_implemented = 0;

            # ksk: $keytag
            # dnskey rrsig: $rrsigrr
            # zsk: $soarrsigrr->keytag()
            # soa rrsig: $soarrsigrr
            # dnskey set: @{$self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_list'}}
            # soa set: @{$self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_list'}}

            # verify signature over DNSKEY RR set made by $keytag

            if ($self->check_enabled('CheckDNSSECAlgorithmNotImplemented')) {
              unless ( exists $impl_algorithm{$rrsigrr->algorithm()} ) {
                $self->p('PROBLEM_DNSSEC_ALGORITHM_NOT_IMPLEMENTED',$names_for_ip_text,
                       'DNSKEY',
                       $rrsigrr->algorithm(),
                       $rrsigrr->algorithm()
                      );
                $algorithm_not_implemented = 1;
              }
            }

            if ($algorithm_not_implemented == 0) {
              if ($self->check_enabled('CheckChainOfTrustDoesNotValidate')) {
                unless ( $rrsigrr->verify($self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_list'}, $tagged->{$keytag}{'dnskey'}) ) {

                  $self->p('PROBLEM_CHAIN_OF_TRUST_DOES_NOT_VALIDATE',$names_for_ip_text,
                         $tagged->{$keytag}{'ds'}->rdatastr(),
                         $soarrsigrr->keytag(),
                         $rrsigrr->vrfyerrstr()
                         ); 
                  $broken = 1;

                }      
              }
            }

            if ($algorithm_not_implemented == 0 && $broken == 0) {
              if ($self->check_enabled('CheckDNSSECAlgorithmNotImplemented')) {
                unless ( exists $impl_algorithm{$soarrsigrr->algorithm()} ) {
                  $self->p('PROBLEM_DNSSEC_ALGORITHM_NOT_IMPLEMENTED',$names_for_ip_text,
                           'SOA',
                           $soarrsigrr->algorithm(),
                           $soarrsigrr->algorithm()
                          );
                  $algorithm_not_implemented = 1;
                }
              }
            }
            my $zsk = find_dnskey($soarrsigrr->keytag(), $self->{'data'}{'zone'}{$ip}{'rrs'}{'dnskey_list'});

            if ($algorithm_not_implemented == 0 && $broken == 0) {
              if ($self->check_enabled('CheckChainOfTrustDoesNotValidate') ) {

                # maybe key signed the zone but is not found in dnskey rr set?
                unless ($zsk) {
                  $self->p('PROBLEM_CHAIN_OF_TRUST_DOES_NOT_VALIDATE',$names_for_ip_text,
                         $tagged->{$keytag}{'ds'}->rdatastr(),
                         $soarrsigrr->keytag(),
                         'Zone Signing Key is not found'
                         );
                  $broken = 1;
                }
              }
            }
           
            if ($algorithm_not_implemented == 0 && $broken == 0) {
              if ($self->check_enabled('CheckChainOfTrustDoesNotValidate')) {

                unless ($soarrsigrr->verify($self->{'data'}{'zone'}{$ip}{'rrs'}{'soa_list'}, $zsk) 
                      ) {

                  $self->p('PROBLEM_CHAIN_OF_TRUST_DOES_NOT_VALIDATE',$names_for_ip_text,
                         $tagged->{$keytag}{'ds'}->rdatastr(),
                         $soarrsigrr->keytag(),
                         $soarrsigrr->vrfyerrstr()
                         );
                  $broken = 1;
                }
              }
            }    
          
            if ($broken == 0 && $algorithm_not_implemented == 0) {
              # no invalid chains; no terminated chains;
              $validated = 1;
            }
            if ($broken == 0 && $algorithm_not_implemented == 1) {
              # no invalid chains; at least 1 teminated chain
              $terminated = 1;
            }

          }

        }

      }

    }

  }

  if ($validated == 0 && $terminated == 0) {
    # no valid chains and no terminated chains (which potentialy could validate 
    # if we would know the algorithm)
    $self->p('PROBLEM_NO_CHAINS_OF_TRUST_VALIDATE','set') if ($self->check_enabled('CheckNoChainsOfTrustValidate'));

  }
  # all other cases are not problematic.
}

sub check_dnssec_data
{
  my $self = shift;
  
  my @ds = @_;

  my $tagged;  # reference to hash: $keytag{'ds'} => dsrr, $keytag{'dnskey'} => dnskeyrr

  # do all servers respond with rrsig data?

  $self->check_rrsig_set() if ( $self->check_enabled( 'CheckNotAllServersReturnDNSSECData' ) );

  # do all servers respond with same DNSKEY RRset?

  $self->check_dnskey_set() if ( $self->check_enabled('CheckInconsistentDNSKEYListing') );

  # is there a DNSKEY for every DS?

  $tagged = $self->check_dnskey_for_ds(@ds) if ($self->check_enabled('CheckNoMatchingDNSKEYForDS') );

  # check the SEP flag on all DNSKEYs and on tagged keys

  $self->check_sep_flag($tagged);

  # check if all tagged keys sign the dnskey rrset
 
  $self->check_dnskey_rrsig($tagged);

  # check is DS hashes are correct

  $self->check_ds_hashes($tagged);

  # validate the chains of trust from DS to SOA
  # we only try it with DS that point to DNSKEY.

  $self->validate_chains_of_trust($tagged);

  # check signature validity interval on the SOA RR

  $self->check_signature_validity_soa();

  return 1;
}


1;

__END__

=head1 NAME

Net::DelCheck - DNS zone delegation checking

=head1 ABSTRACT

Provide routines for checking of DNS zone delegations.  It does data
collection via Net::DNS, performs a (configurable) variety of tests
and produces XML reports based on an XML template file containing
problem reports into which arguments are instantiated when problems
occur.

=head1 SYNOPSIS

use Net::DelCheck;

######### Simplest usage, no error checking ########

my $checker = new Net::DelCheck( { 'ModuleTemplatePath' => 'sample/template.xml' } );

$checker->get_data( 'a.zone.net', 'server1.bob.com', 'server2.bob.com' );

print $checker->check;

$checker->cleanup;

############### With error checking ################

my $checker = new Net::DelCheck( { 'ModuleTemplatePath' => 'sample/template.xml' } );

die "Oops: errors making DelCheck object:". ( join "\n\t", &Net::DelCheck::errors )
    if ( ! defined $checker );
    
my @errors = $checker->get_data( 'a.zone.net', 'server1.bob.com', 'server2.bob.com' );

die ( join "\n", @errors ) 
    if ( @errors );

my $xml_out = $checker->check;

if ( ! &my_verify_xml( $checker->schema(), $xml_out ) ) {
    print "oops bad output\n";
}

$checker->cleanup;
    
######### Multiple zones, no error checking ########

my $checker = new New::DelCheck( 'confs/base.conf' );

my %problems;

foreach my $zone_to_check ( @zones_to_check ) {

    # Pick config based on type of zone
    if ( &is_tld( $zone_to_check ) ) {
	$checker->init_config( { 'ModuleUseIPv6' => 0 }, 'confs/base.conf', 'confs/tld.conf' ) )
    }	
    else {
	$checker->init_config( 'confs/base.conf' );
    }

    # Calling get_data clears previous data and generated problems

    $checker->get_data( $zone_to_check, &get_zone_servers( $zone_to_check ) );

    $problems_xml{$zone_to_check} = $checker->check();
}

$checker->cleanup;

############### With error checking, DNSSEC checks enabled ################

my $checker = new Net::DelCheck( { 'ModuleTemplatePath' => 'sample/template.xml' } );

die "Oops: errors making DelCheck object:". ( join "\n\t", &Net::DelCheck::errors )
    if ( ! defined $checker );

my @errors = $checker->get_data( 'a.zone.net', 'server1.bob.com', 'server2.bob.com' );

die ( join "\n", @errors )
    if ( @errors );

my @ds = (<ds record>, <ds record>, ...);
my $xml_out = $checker->check(@ds);

if ( ! &my_verify_xml( $checker->schema(), $xml_out ) ) {
    print "oops bad output\n";
}

$checker->cleanup;


=head1 OVERVIEW

There are two phases to a check. The first involves gathering all the
necessary data, the second checking the data and producing a report.
Phase one is executed via the &L<"get_data"> call, phase two via
the &L<"check"> call.

B<TIP>: to see a lot more information about what is going on during
both phases, set L<"ModuleDiagnosticLevel"> to 5.

=head2 PROBLEM CODES AND ARGUMENTS

Whenever DelCheck finds a problem, it generates a B<problem code> and
a set of B<arguments>. The code indicates the problem which was found
and is the same each time that problem occurs. The arguments indicate
the precise circumstances under which the problem occured and give
further information about what is wrong e.g. which value was bad, what
the expected minimum value was, what the expected maximum value was
etc.

The user defines in a template file what XML should be output for
each problem code that can occur. The user includes placeholder
tags into which the actual arguments which occured are inserted.

=head2 TEMPLATE FILE FORMAT

The location of the template file is defined by ModuleTemplatePath in
the config. A sample (valid) template file called template.xml can be
found in the sample/ directory which comes with this module.

There are the following restrictions on the template file:

=over 4

=item *

it must contain a root tag (with any name)

=item *

the root tag must contain a 'version' attribute, and the value of the
version attribute must be the same as the $TEMPLATE_VERSION variable in
this module. This is to stop changes in problems or their arguments
taking template authors by surprise.

=item *

it must contain a set of <PROBLEM> tags (although this specific name can be
changed via the config), each with a 'code' attribute (also configurable)
specifying the problem code to which it applies.

=item *

there must be exactly one problem-code combination for each problem code
which can be generated by the module.

=item *

it B<MAY> contain exactly one NO_PROBLEM tag directly under the root node.
If such a tag is present, it will be inserted into the report if no
problem codes are generated by a check.

=back

The <PROBLEM> tag can contain any mixture of text and XML to any
depth.  <ARG num="X"> tags can be used anywhere to indicate that at
that point argument X to the problem code should be instantiated in
the output of the module. Note: any children of an ARG tag will NOT
be present in the module output.

=head2 OUTPUT FORMAT

The output format is decided largely by the user, but the skeleton
structure is the same every time. It's a set of problem elements
organised by 'problem entity' (the thing which had the problem).

You can see the output format by calling the &L<"schema"> object
method.

=head1 OBJECT METHODS

=head2 init_config

    $checker->init_config(  'ModuleTemplatePath' => 'sample/template.xml', 'file1', 'conf/file2' );

I<Purpose>: read in one or more configuration specifications, completely replacing
any previous configuration.

I<In>:  @config_specs: list of configuration specifications.
See &L<"new"> for more information.

=head2 get_data

    my @errors = $checker->get_data( $zone, 'ns.my.com', 'ns.their.com' );

I<Purpose>: pull in the necessary data about the specified
zone and nameservers from the network.

I<In>: $zone: zone to do checks on

I<In>: @nss: correct list of nameservers for this zone. A port
can be suffixed e.g. 'sample.name.server:5353'.

I<Out>: @errors: list of error messages occuring. If any errors occur,
the zone cannot be checked.

I<Comments>: 

DelCheck was developed for use by a DNS delegating authority.  In that
context, the 'correct' list of nameservers is the list of nameservers
which the requester has specified for the zone. In other contexts, the
list will be what the system administrator thinks the nameservers
should be, or could be filled in from e.g. the results of a query for
the nameserver list at the known primary server.

This method operates roughly as follows:

 - for each unique name in @nss, do an A and AAAA query for $ns_name at the local nameserver(s)
 - pick an IP for each $ns_name
 - for each unique IP picked
    - do a PTR query for IP at the local nameserver(s)
    - do a SOA and an NS query for $zone at $ip
 - if CheckMnameResolvable is set
    - for each valid SOA found at a server
    - do an A and AAAA query on the mname, retain records returned

No further network traffic is generated by DelCheck once &L<"get_data"> is
complete.

=head2 check

    print $checker->check();

I<Purpose>: check data and return an XML report.
Don't use this prior to calling &L<"get_data">!

I<Out>: $xml: scalar containing an XML DelCheck report document

I<Comments>: the report syntax can be verified using the output
of the object method &L<"schema"> if you wish.

The function generates a set of <PROBLEM> tags. Each problem
has a I<code> determining exactly what the problem was, and
a number of <ARG>s describing the circumstances under which
the problem occured.

The <PROBLEM> tags are organized within a sequence of <PROBLEM_ENTITY>
tags. Each problem entity is either a nameserver passed to &<<"get_data">
(possibly several comma-separated names, if they shared IP addresses),
or the word 'set'. I<set> problems relate to the entire set of
nameservers e.g. the problem code
PROBLEM_TOO_FEW_DELEGATED_NAMESERVERS relates to the entire set of
nameservers, rather than being associated with a particular
nameserver.

The actual tag and attribute names are configurable via 
the configuration directives ModuleReportXMLTags and
ModuleReportXMLAttributes.

If no problem tags are generated, the function checks for a
<NO_PROBLEM> tag in the template file, and will include this under the
root element if it is present.

The output of this function is minimally an XML declaration and
a root tag pair <DELCHECK_REPORT> (the name is configurable via
ModuleReportXMLTags).

=head2 schema

    print $checker->schema();

I<Purpose>: return a schema that can be used to validate
the output from &L<"check">

I<Out>: $schema: scalar containing an XML W3C schema document

=head2 unanimous_primary

    print 'Primary:', $checker->unanimous_primary(), "\n";

I<Purpose>: return the single unique primary nameserver for the
zone, or undef if no unanimous primary could be
determined

I<Comments>: You must run &L<"get_data"> to pull the zone data first! 

If each server returns an AA SOA record and lists the same mname, that
mname is returned, otherwise undef.

=head2 cleanup

    $checker->cleanup;

I<Purpose>: do garbage collection. You must call this when you're
finished with the object.

=head2 xml_attribute_name

    my $attribute_name = $checker->xml_attribute_name( 'problem_code' );

I<Purpose>: return the relevant attribute name that will appear in the output
of L<"check">.

I<In>: $attribute_type: the attribute of interest. Possible values are
    problem_entity
    problem_code

I<Out>: $attribute_name: the name of the attribute that will appear
in the relevant tag

I<Comments>: 

This routine was included primarily for testing purposes.

=head2 xml_tag_name

    my $tag_name = $checker->xml_tag_name( 'problem' );

I<Purpose>: return the XML tag that will appear in the output
of L<"check"> for a particular tag type.

I<In>: $tag_type: the type of tag. Possible values are
   report: report root node name
   problem_entity: problem entity node name
   problem: problem code node name
   no_problem: node name for report when there no problems occur

I<Out>: $tag_name: the associated tag name

I<Comments>: 

This routine was included primarily for testing purposes.

=head1 EXPORTED CLASS METHODS

All of the exported methods except &L<"new"> and &L<"errors"> are needed for the module's
internal workings, and just exported on the offchance that they are
useful to the calling program.

=head2 new

    my $checker1 = new DelCheck( { 'ModuleTemplateFile' => 'sample/template.xml' );

    my $checker2 = new DelCheck( 'confs/base.cnf', 'confs/tld.conf' );

I<Purpose>: Make a new checker

I<In>: @config_specs: (possibly empty) list of config file paths,
or references to hashes specifying config settings.

A config file must be in App::Config format.

For both config files and config hashes, the following applies:

=over 4

=item *

for array variables, another key/value pair must be given for each
array member, where key is the name of the array.

=item *

For hash variables, the key specifies the hash name and the key/value
pair is specified via key=value. See AppConfig for more details.

=back

If no config specifications are given, default values are used (the default
values are given later in this documentation). The only configuration
directive which must be specified is ModuleTemplateFile.

I<Out>: $self: reference to the new DelCheck object

=head2 errors

    my @error_messages = &DelCheck::errors;

I<Purpose>: access to error messages when a DelCheck object
could not be created.

I<Out>: Array context: List of error messages. The list
is cleared. Scalar context: number of errors in the list. The list
is not cleared.

The object should not be used further if this is not empty.

=head2 test_format

    $ok_flag = &test_format( $host, 'hostname' );

I<Purpose>: check a scalar holds an entity of a particular format

I<In>: $to_check: thing we want to check

I<In>: $expected_format: 'ip', 'port', 'domain' or 'hostname'

I<Out>: 1 = ok, 0 = bad

=head2 normalise_domain_name

    my $normalised_domain_name = &normalise_domain_name( $messy_domain_name );

I<Purpose>: remove preceeding/trailing dot, whitespace, convert to
lower case etc etc. Needed when comparing domain names, using them as hash
keys etc

I<In>: $messy_domain_name: name to normalise

I<Out>: $normalised_domain_nam: normalised version of input name

=head2 is_reverse_zone

   my $ip_version = &is_reverse_zone( $domainname );

   if ( $ip_version == 0 ) {
       print "It's not a reverse zone, Fred\n";
   } else {
       print "Reverse tree, IP $ip_version\n";
   }

I<Purpose>: decide whether a domain name is in a reverse tree,
and if it is, for which IP version.

I<In>: $domain_name: name to check

I<Out>: $ip_version: IP version of the reverse zone tree, or 0

=head2 ipv6_available

  my $is_available_flag = &Net::DelCheck::ipv6_available;

I<Purpose>: report whether the module can use IPv6 transport
or not

I<Out>: $is_available_flag: 0/1

=head2 get_ip_version

   my $ip_version = &is_reverse_zone( $domainname );

   if ( $ip_version == 0 ) {
       print "It's not a reverse zone, Fred\n";
   } else {
       print "Reverse tree, IP $ip_version\n";
   }

I<Purpose>: decide the IP protocol version of an IP address

I<In>: $ip: IP to examine (not a shortened version)

I<Out>: $version_no: IP version of $ip or 0 if we couldn't
decide (maybe it was syntactically bad).

=head2 problem_code_no_args

   my $no_args = &problem_code_no_args( 'PROBLEM_NO_REVERSE_MAPPING' );

I<Purpose>: get the number of arguments which the specified code
accepts

I<In>: $code: the problem code

I<Out>: $no_args: the number of arguments $code accepts or undef
if $code is unknown.

I<Comment>: mainly included for testing purposes

=head2 seconds_to_descr_string

   my $descr_string = &seconds_to_descr_string( $seconds );

I<Purpose>: convert a value in seconds to a descriptive
string like '3 weeks, 2 days'.

I<In>: $seconds: number of seconds to convert

I<Out>: $descr_string: the converted value

=head1 CONFIG FILE

Multiple config files can be read, either when creating a DelCheck
object and explicitly via &L<"init_config">. Note that &init_config wipes
any previous configuration.

Configuration files are in AppConfig format.

Configuration directives fall into three categories. Those beginning
with I<Check> are boolean options which enable or disable a particular
check. Directives beginning with I<Param> specify the parameters with
which particular checks are applied (if the check is enabled!).
Directives beginning with I<Module> control other aspects of module
operation. 

=head2 Check Directives

The I<Generates> field indicates the problem codes that a particular
check can generate. Note that even if all check directives are off,
problem codes can still be generated. For instance, the code
PROBLEM_LOOKUP can always be generated.

=head3 CheckHostnamesAgainstDomainNameSyntax

Check hostnames match required domain name syntax. This includes
checking

=over 4

=item *

the length of the name against L<"ParamMaxDomainNameLength">

=item *

the length of each label against L<"ParamMaxDomainNameLabelLength">

=item *

for empty labels

=back

Under 'hostnames' is included:  nameserver names, mname of SOA record
and rname host portion.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DOMAIN_NAME_TOO_LONG, PROBLEM_DOMAIN_NAME_LABEL_EMPTY, PROBLEM_DOMAIN_NAME_LABEL_TOO_LONG

=head3 CheckHostnamesForIllegalCharacters

Check that hostnames match ^[a-zA-Z0-9-]+$

Under 'hostnames' is included: nameserver names, mname of SOA record
and rname host portion.

This directive is ignored if
L<"CheckHostnamesAgainstDomainNameSyntax"> is disabled, or the domain
name check found problems.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_HOSTNAME_LABEL_ILLEGAL_CHARACTERS

=head3 CheckHostnameForIllegalStartEndCharacters

Check that the start and end of nameserver names are a letter or
number. 

Under 'hostnames' is included:  nameserver names, mname of SOA record
and rname host portion.

This directive is ignored if L<"CheckHostnamesAgainstDomainNameSyntax">
is disabled, or the domain name check found problems.

Note: disabling this check has no effect on L<CheckHostnameForIllegalStartEndCharacters>.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_HOSTNAME_LABEL_ILLEGAL_START_OR_END

=head3 CheckHostnamesHaveMultipleLabels

Check that hostnamess consist of more than one label. This is
technically OK, but highly unlikely to be correct.

Under 'hostnames' is included:  nameserver names, mname of SOA record
and rname host portion.

This directive is ignored if L<"CheckHostnamesAgainstDomainNameSyntax">
is disabled or the domain name check found problems.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_HOSTNAME_ONE_LABEL

=head3 CheckNameserverNameCanonical

Check that no CNAME resource record is returned for the nameserver name.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_NS_NAME_NOT_CANONICAL

=head3 CheckNameserverHasReverse

Check whether a PTR record exists in DNS for the IP that the nameserver
name resolved to.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_NO_REVERSE_MAPPING

=head3 CheckRRSetTTL

Checks if the TTLs in an RRset are the same (RFC2381 5.1).  Currently
only the NS RRsets returned by eacn name server are tested.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_RRSET_TTLS_NOT_EQUAL

=head3 CheckNameserverMinimumTTL

Check if the TTL on the name serverset is at least
L<"ParamMinimumNameserverTTL">. The test is performed at the stage where
the data received from each server is tested and skipped if the data from
the name server tested has a soa serial that has been seen previously.

I<Type>: Boolean

I<Default>: On

I<Generates>:  PROBLEM_NS_TTL_TOO_SHORT

=head3 CheckNameserverHasMatchingReverse

If we got one or more PTR records back for the IP that the nameserver
name resolved to, check that one of them maps back to the original name.
If no records are returned, this directive will B<not> produce a problem
code.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_WRONG_REVERSE_MAPPING

=head3 CheckNumberOfNameserverNames

Check that there are at least L<"ParamMinimumNumberNameservers"> nameservers
in the list passed to &L<"get_data">.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_TOO_FEW_DELEGATED_NAMESERVERS

=head3 CheckForDuplicateNameserverIPs

Check for two of the nameservers passed to &L<"get_ip"> resolving to the
same IP address.

If a nameserver name resolves to several IPs, a random one
is picked, so this test may fail.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DUPLICATE_DELEGATED_NS_IP

=head3 CheckIPProximity

Check whether any nameserver IPs are in the same /24.

IPv6 addresses are ignored for the purposes of this test, due to
limitations in Net::IP.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DELEGATED_NS_ON_SAME_SUBNET

=head3 CheckForDifferingSerialNumbers

Check whether the serial numbers for the zone at any of the
nameservers differ. The check takes place whether or not
this directive is set, but is no reported unless the
directive is set.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SERIAL_NUMBERS_DIFFER

=head3 CheckForDifferingSOAValues

Check whether any other SOA values differ between nameservers.
This check will only take place if the serial numbers differ.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SOA_VALUES_DIFFER

=head3 CheckForInconsistentNSListings

Check that the NS records for the zone listed at each nameserver
are consistent.

Only nameservers which returned NS records for the zone are
considered.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_INCONSISTENT_NS_LISTING

=head3 CheckForDuplicateDelegatedNameserverNames

Check for duplicate (normalized) nameserver names in the list 
passed to &L<"get_data">.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DUPLICATE_DELEGATED_NS_NAME

=head3 CheckForDuplicateNameserverNamesAtZone

Check for duplicate (normalized) nameserver names within the list 
returned by each nameserver.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DUPLICATE_NS_NAME_LISTED_AT_ZONE

=head3 CheckAllDelegatedNameserversListedAtZone

Check that each (normalized) nameserver name in the list
passed to &L<"get_data"> is also present in the list returned
by each nameserver.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DELEGATED_NS_NOT_LISTED_AT_ZONE

=head3 CheckListedNameserversInDelegatedList

Check that each (normalized) nameserver name returned
by each nameserver is also present in the list passed
to &L>"get_data">.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_LISTED_NAMESERVER_NOT_IN_DELEGATED_LIST


=head3 CheckSerialRecommendedFormat

Check that the serial for the zone returned by each nameserver matches
YYYYMMDDCC, where

=over 4

=item *

YYYY is a year greater than L<"ParamFirstReasonableSerialYear">

=item *

MM is a month between 1 and 12

=item *

DD is a day between 1 and 31

=item *

CC is a counter of any value

=back

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SERIAL_NOT_RECOMMENDED_FORMAT

=head3 CheckSerialNotInFuture

Check the serial number for the zone returned by each nameserver is
not in the future.

This requires a little care, because there's a decent chance
administrators are not thinking about GMT when entering serial
numbers, therefore we don't complain unless the (local time) start of
the day in the serial number is at least
L<"ParamSerialFutureAllowanceHours"> forward of the current local time.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SERIAL_IN_FUTURE

=head3 CheckMultipleSOARecordsAtServer

Check that server is not returning multiple SOA records for the zone.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_MULTIPLE_SOA_AT_SERVER

=head3 CheckNoSOARecordAtServer

Check that each server returns at least one SOA record for the zone.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_NO_SOA_AT_SERVER

=head3 CheckSOARecordAtServerAuthoritatative

Check that each server returns an authoritative answer when queried
about the zone (i.e. the AA flag is set).

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_NON_AA_NAMESERVER

=head3 CheckMnameInNSSetAtServer

Check that the mname field of the SOA record returned by each server is
in the list of nameservers returned by the same server.

Note: it's not required by the RFCs that the server listed in the
mname field is also listed in the nameserver list (RFC1996).

I<Type>: Boolean

I<Default>: On

=head3 CheckMnameResolvable

Check that the hostname listed in the mname field resolves to an IP
address.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_MNAME_NOT_LISTED_AS_NS

=head3 CheckRnameNoDuplicateZone

Check for the zone being mentioned twice in the rname of a returned
SOA record. This may be an indication that in the zone file the
terminating period (.) was accidentally ommitted.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_RNAME_DUPLICATE_ZONE

=head3 CheckRnameNotReverseZone

Check that the rname field does not contain a reverse zone.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_REVERSE_DOMAIN_IN_RNAME

=head3 CheckRnameHasNoAtSign

Check that the rname field does not contain an @ character.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_RNAME_HAS_AT_SIGN

=head3 CheckRnameAgainstDomainNameSyntax

First, the domain name is extracted from the rname. It's considered to
be the first non-backslashed period (.)

The domain name is then checked for:

=over 4

=item *

the length of the name being more than L<"ParamMaxDomainNameLength">

=item *

the length of each label being more than L<"ParamMaxDomainNameLabelLength">

=item *

empty labels

=back

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DOMAIN_NAME_TOO_LONG, PROBLEM_DOMAIN_NAME_LABEL_EMPTY, PROBLEM_DOMAIN_NAME_LABEL_TOO_LONG

=head3 CheckRnameMiscSyntax

Check the rname looked like 'mailboxname.domainname'.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_RNAME_BAD_SYNTAX

=head3 CheckRefreshRange

Check that the SOA refresh value returned by each nameserver
is between L<"ParamRefreshMin"> and L<"ParamRefreshMax">.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_REFRESH_OUT_OF_RANGE

=head3 CheckRetryRange

Check that the SOA retry value returned by each nameserver
is between L<"ParamRetryMin"> and L<"ParamRetryMax">.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_RETRY_OUT_OF_RANGE

=head3 CheckExpireRange

Check that the SOA expire value returned by each nameserver
is between L<"ParamExpireMin"> and L<"ParamExpireMax">.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_EXPIRE_OUT_OF_RANGE

=head3 CheckExpireAgainstRefresh

Check that the SOA expire value returned by each nameserver
is less than the SOA refresh returned by the same server.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_EXPIRE_LESS_THAN_REFRESH

=head3 CheckExpireAgainstRefreshRetrySum

Check that the sum of the SOA refresh and retry values returned by
each nameserver is greater than the SOA expire returned by the same
server.

I<Type>: Boolean

I<Default>: On

I<Generates>:PROBLEM_EXPIRE_LESS_THAN_REFRESH_RETRY_SUM

=head3 CheckMinimumRange

Check that the SOA minimum value returned by each nameserver
is between L<"ParamMinimumMin"> and L<"ParamMinimumMax">.

Note that the minimum field can have more than one application!

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_MINIMUM_OUT_OF_RANGE

=head3 CheckForRequiredNameservers

Check that all nameservers listed in L<"ParamRequiredNameserverList"> are
also present in the list passed to &L<"get_data">, according to the
specification in  L<"ParamRequiredNameserverSpec">.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_REQUIRED_NAMESERVER_MISSING

=head3 CheckForProhibitedNameservers

Check that none of the nameservers listed in L<"ParamProhibitedNameserverList"> are
present in the list passed to &L<"get_data">.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_PROHIBITED_NAMESERVER_PRESENT

=head3 CheckSkippedSerial

The module skips several tests when a serial number is encountered at
a nameserver which has previously been seen at another nameserver.
This directive turns on and off the informational report about whether
this has happened or not.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SKIPPED_SERIAL_NUMBER

=head3 CheckRnameMail

When this prameter is "on" _and_ L<"ModuleMailEnvelopeFrom"> has been
configured the module will perform several checks on the validity of
the email address in the rname. 

These test will not be done if one of the following problems has been
encountered.  PROBLEM_RNAME_DUPLICATE_ZONE, PROBLEM_RNAME_HAS_AT_SIGN,
PROBLEM_RNAME_BAD_SYNTAX, PROBLEM_REVERSE_DOMAIN_IN_RNAME.

First the program will perform an MX lookup, which can fail with
PROBLEM_MAILCHECK_MX_LOOKUP. Than it will lookup the A RR for each of
the MX record. When these lookup fail a PROBLEM_MAILCHECK_A_LOOKUP is
generated.

The program will then connect to port 25 of each listed MTA. If
multiple A RRs are returned for a specific MTA all these IPs will be
tried, this can fail with PROBLEM_MAILCHECK_CONNECTION_FAILED, and
start an SMTP message exchange.

The program first sends an EHLO to test if the MTA supports SMTP
extentions (rfc1869). If that does not successed HELO is used.  If a
RSET after EHLO fails PROBLEM_MAILCHECK_RSET_AFTER_EHLO_FAIL is
raised.  After the "HELO" the function will first try to set up a
session with an empty envelope from, if that fails (setting
PROBLEM_MAILCHECK_EMPTY_ENVELOP_FROM_FAILS) the program will send a
mail with the recipient configured in

If the RCPT TO <rname address> succeeds the routine RSETs the channel
and stops the SMTP transaction (no mail is being send).

If the RCPT TO fails the PROBLEM_MAILCHECK_RCPT_TO_RNAME_FAILS flag is set.

If the parameter L<"ModuleMailCache"> has been configured than e-mail
addresses for which the test was successfull are being cached so that
the test is not done by new instances of the checker. See
L<"ModuleMailCache">for details.


I<Type>: Boolean

I<Default>: Off

I<Generates>: PROBLEM_MAILCHECK_RSET_AFTER_EHLO_FAILS,
PROBLEM_MAILCHECK_EMPTY_ENVELOP_FROM_FAILS,
PROBLEM_MAILCHECK_RCPT_TO_RNAME_FAILS.

Also see L<"ModuleMailEnvelopeFrom"> wich will have to be configured
in order for these tests to be ran. If CheckRnameMail eq on and
ModuleMailEnvelopeFrom is not configured a fatal error will occur.

=head3 CheckNotAllServersReturnDNSSECData

Check if all servers return RRSIG records in authority section.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_NOT_ALL_SERVERS_RETURN_DNSSECDATA

=head3 CheckInconsistentDNSKEYListing

Checks if all servers return same set of DNSKEY records.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_INCONSISTENT_DNSKEY_LISTING

=head3 CheckNoMatchingDNSKEYForDS

Checks if every DS records supplied has a corresponding DNSKEY.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_NO_MATCHING_DNSKEY_FOR_DS

=head3 CheckDSPointingToNonSEPDNSKEY

DS points to DNSKEY that has no SEP flag set.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DS_POINTING_TO_NON_SEP_DNSKEY

=head3 CheckSepFlagNotUsed

None of the DNSKEYs that DS point to have SEP flag set.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SEP_FLAG_NOT_USED

=head3 CheckSepKeyNotSelfSigned

Sep DNSKEY doesn't sign the DNSKEY RR set.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SEP_KEY_NOT_SELF_SIGNED

=head3 CheckDSDigestAlgorithmNotKnown

Digest algorithm on DS record is not known.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DS_DIGEST_ALGORITHM_NOT_KNOWN

=head3 CheckDSAndDNSKEYRRDoNotMatch

DS doesn't match DNSKEY. probably copy-and-paste error.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DS_AND_DNSKEYRR_DO_NOT_MATCH

=head3 CheckDNSSECAlgorithmNotImplemented

Not possible to check RRSIG because unknown algorithm was used.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_DNSSEC_ALGORITHM_NOT_IMPLEMENTED

=head3 CheckChainOfTrustDoesNotValidate

The chain of trust is not validated. Usually points to the broken signatures.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_CHAIN_OF_TRUST_DOES_NOT_VALIDATE

=head3 CheckNoChainsOfTrustValidate

None of the chains of trust for the known algorithms validate.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_NO_CHAINS_OF_TRUST_VALIDATE

=head3 CheckSignatureValidityNearsExpiration

Signature has passed 4/5 of its validity interval.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SIGNATURE_VALIDITY_NEARS_EXPIRATION

=head3 CheckTtlLargeFractionOfSignatureValidity

TTL is > than half of signature validity interval.

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_TTL_LARGE_FRACTION_OF_SIGNATURE_VALIDITY

=head3 CheckSOAMnameDiffer

Check that all Mnames in SOA in all nameservers are the same.
This check always complains about Windows servers, in which every nameserver
it has its own name as Mname (requested by Brett).

I<Type>: Boolean

I<Default>: On

I<Generates>: PROBLEM_SOA_VALUES_DIFFER


=head2 Param Directives

The I<Used by> field indicates which checks use the parameters. In
most cases, the function of the Param is explained sufficiently in the
description of the Check(s) it is associatied with.

=head3 ParamMaxDomainNameLength

Maximum length of a domain name.

I<Type>: Integer

I<Default>: 255

I<Used By>: L<"CheckHostnamesAgainstDomainNameSyntax">, L<"CheckRnameAgainstDomainNameSyntax">

=head3 ParamMaxDomainNameLabelLength

Maximum length of a domain name label.

I<Type>: Integer

I<Default>: 63

I<Used By>: L<"CheckHostnamesAgainstDomainNameSyntax">, L<"CheckRnameAgainstDomainNameSyntax">

=head3 ParamMinimumNumberNameservers

I<Type>: Integer

I<Default>: 2

I<Used By>: L<"CheckNumberOfNameserverNames">

=head3 ParamFirstReasonableSerialYear

The YYYY field in a serial number should be at least this, otherwise we
suspect that it's not actually representing a year.

I<Type>: Integer

I<Default>: 1980

I<Used By>: L<"CheckSerialRecommendedFormat">

=head3 ParamSerialFutureAllowanceHours

I<Type>: Integer

I<Default>: 24

I<Used By>: L<"CheckSerialNotInFuture">

=head3 ParamRefreshMin

I<Type>: Integer

I<Default>: 1200

I<Used By>: L<"CheckRefreshRange">

=head3 ParamRefreshMax

I<Type>: Integer

I<Default>: 43200

I<Used By>: L<"CheckRefreshRange">

=head3 ParamExpireMin

I<Type>: Integer

I<Default>: 1209600

I<Used By>: L<"CheckExpireRange">

=head3 ParamExpireMax

I<Type>: Integer

I<Default>: 2419200

I<Used By>: L<"CheckExpireRange">

=head3 ParamMinimumMin

I<Type>: Integer

I<Default>: 3600

I<Used By>: L<"CheckMinimumRange">

=head3 ParamMinimumMax

I<Type>: Integer

I<Default>: 86400

I<Used By>: L<"CheckMinimumRange">


=head3 ParamNameServerMinimumTTL

I<Type>: Integer

I<Default>: 30

I<Used By>: L<"CheckNameserverMinimumTTL">

=head3 ParamRetryMinFractionRefresh

I<Type>: Float

I<Default>: 0.05

I<Used By>: L<"CheckRetryRange">

=head3 ParamRetryMaxFractionRefresh

I<Type>: Float

I<Default>: 1.0

I<Used By>: L<"CheckRetryRange">

=head3 ParamRequiredNameserverList

I<Type>: List

I<Default>: Empty

I<Used By>: L<"CheckForRequiredNameservers">

=head3 ParamRequiredNameserverSpec

How to interpret L<"ParamRequiredNameserverList"> when it contains
more than one entry.

all: all of the servers in ParamRequiredNameserverList must be present

at-least-one: at least one of the servers in ParamRequiredNameserverList must be present
              
exactly-one: guessable I hope

I<Type>: String

I<Default>: exactly-one

I<Used By>: L<"CheckForRequiredNameservers">

=head3 ParamProhibitedNameserverList

I<Type>: List

I<Default>: Empty

I<Used By>: L<"CheckForProhibitedNameservers">

=head2 Module Directives

=head3 ModuleQueryRemoteNameserverV4

If set, all queries to remote nameservers over IPv4 will go instead to
the IP address specified by this directive. This directive is intended
purely for testing to allow serveral nameservers to respond from different
ports of a single host.

It's value should be EMPTY FOR NORMAL OPERATION.

Queries to nameservers that take place due to a listing in
ModuleLocalResolvers are not affected by this setting.

I<Type>: String

I<Default>: none

=head3 ModuleQueryRemoteNameserverV6

The IPv6 equivalent of L<"ModuleQueryRemoteNameserverV6">. Please see that
directive for more information.

I<Type>: String

I<Default>: none

=head3 ModuleLocalResolvers

List of local resolving nameservers. Must have recursion enabled.
If this list is empty, the Net::DNS defaults are used (most likely
 F</etc/resolv.conf>). In previous tested versions, Net::DNS
broke when names were specified here, probably better to use IPs.

I<Type>: List

I<Default>: Empty

=head3 ModuleQueryRemoteDefaultPortNumber

If no port is specified when the nameserver name is passed to
&L<"get_data">, this value is passed to the Net::DNS::Resolver function I<port> when
querying local (resolving) nameservers.

I<Type>: Integer

I<Default>: 53

=head3 ModuleQueryRemoteRetry

Passed to the Net::DNS::Resolver function I<retry> when querying remote nameservers.

I<Type>: Integer

I<Default>: 3

=head3 ModuleQueryRemoteRetrans

Passed to the Net::DNS::Resolver function I<retrans> when querying remote nameservers.

I<Type>: Integer

I<Default>: 5

=head3 ModuleQueryRemoteUDPTimeout

Passed to the Net::DNS::Resolver function I<udp_timeout> when querying remote
nameservers.

I<Type>: Integer

I<Default>: 10

=head3 ModuleQueryLocalPortNumber

Passed to the Net::DNS::Resolver function I<port> when querying local (resolving)
nameservers.

I<Type>: Integer

I<Default>: 53

=head3 ModuleQueryLocalRetry

Passed to the Net::DNS::Resolver function I<retry> when querying local (resolving)
nameservers (see also: L<"ModuleLocalResolvers">).

I<Type>: Integer

I<Default>: 3

=head3 ModuleQueryLocalRetrans

Passed to the Net::DNS::Resolver function I<retrans> when querying local
(resolving) nameservers.

I<Type>: Integer,

I<Default>: 5

=head3 ModuleQueryLocalUDPTimeout

Passed to the Net::DNS::Resolver function I<udp_timeout> when querying local
(resolving) nameservers.

I<Type>: Integer

I<Default>: 10

=head3 ModuleNameserverIgnoreList

I<Type>: List

I<Default>: <empty>

Nameserver names in this list are ignored if they are passed via
&L<"get_data">. No queries are made to them. Problems are only reported for
them where it makes sense e.g.  L<"CheckNumberOfNameserverNames">,
L<"CheckForDuplicateDelegatedNameserverNames">,
L<"CheckAllDelegatedNameserversListedAtZone">.

=head3 ModuleUseIPv6

If enabled, the module will use AAAA records for queries when possible
(i.e. when resolution of a nameserver name returns AAAA record(s).

If this option is enabled, the following Perl libraries must be available
(because they are required by Net::DNS IPv6 support)

=over 4

=item *

   Socket6

=item *

   IO::Socket::INET6 v2.01

=back

This requires that the host that both Net::DNS and the host the caller
is running on support IPv6 queries.

I<Type>: Boolean

I<Default>: Off

=head3 ModuleReportXMLDeclaration

The XML declaration to use at the start of output from the &L<"check">
method. 

I<Type>: String

I<Default>: <?xml version="1.0"?>

=head3 ModuleDiagnosticLevel

If set above 0, the module will print diagnostic messages to STDOUT.
The higher the value, the more detailed the diagnostics. Values greater
than 5 have the same effect as 5.

I<Type>: Integer

I<Default>: 0

=head3 ModuleDiagnosticFormat

The format in which diagnostic messages should be printed. Options
are 'txt' or 'html'.

I<Type>: String

I<Default>: txt

=head3 ModuleDiagnosticPacketHeaders

Whether to print packet headers of sent and received DNS packets as a
diagnostic aid. Headers will only be printed if
L<"ModuleDiagnosticLevel"> is greater than 0.

I<Type>: Boolean

I<Default>: On

=head3 ModuleSchemaNamespaceURI

The URI of the namespace used in the output of &L<"schema">.

I<Type>: URI

I<Default>: http://www.w3.org/2001/XMLSchema

=head3 ModuleSchemaNamespaceName

The name of the namespace used in the output of &L<"schema">.

I<Type>: String

I<Default>: xsd

=head3 ModuleReportXMLTags

The XML tags used in the output from &L<"check"> and &L<"schema">. The key is
the entity type in the report, the value is the TAG to use for that
entity.

I<Type>: Hash

I<Default>:

    problem=PROBLEM
    problem_entity=PROBLEM_ENTITY
    report=DELCHECK_REPORT
    no_problem=NO_PROBLEM

=head3 ModuleReportXMLAttributes

The XML attributes names used in the output from &<"check"> and &L<"schema">. The key is
the tag it occurs in and the attribute, the value is the attribute name to use.

I<Type>: Hash

I<Default>:

    problem_entity_name=name
    problem_code=code

=head3 ModuleSchemaTypes

The schema output from &L<"schema"> defines several complex types making up
a report. This directive defines the names for the complex types which
are defined. Since the scope of the names is limited to the schema
itself, changing these values is probably only useful for
e.g. readability.

The key is the entity for which a complex type is being defined, the
value is the name used in the schema for the complex type describing
that entity.

I<Type>: Hash

I<Default>:

    problem_code=ProblemCodeDefinition
    problem=ProblemDefinition
    problem_entity=ProblemEntityDefinition
    report=DelcheckReportDefinition

=head3 ModuleSOAKeys

I<Type>: List
I<Default>: serial, mname, rname, refresh, retry, expire, minimum

List of SOA key types accessible via L<Net::DNS::RR:SOA>. Probably
better not change this unless Net::DNS::RR::SOA changes.

=head3 ModuleTemplatePath

I<Type>: String
I<Default>: NONE

Path to the template file to use for output. No default.

=head3 ModuleTemplateXMLArgTag

I<Type>: String
I<Default>: ARG

Tag name for tags which mark argument locations in a problem
template.

=head3 ModuleTemplateXMLArgNoAttr

I<Type>: String
I<Default>: 'no'

Attribute name for the attribute in an ARG tag in a problem
template which indicates the argument number which should be
instantiated in that place.

=head3	ModuleMailTimeoutSocket

I<type>: integer
I<Default>: 5

Parameter that determines the timeout for setting up a TCP session
to an  MTA during the L<"CheckRnameMail">.


=head3	ModuleMailTimeoutRead

I<type>: integer
I<Default>: 50

Parameter that determines the timeout for waiting for a read during 
the the transaction with MTAs during the L<"CheckRnameMail">.


=head3 ModuleMailEnvelopeFrom

I<type>: String
I<Default>: undef

Parameter  that determines the "from" address in SMTP envelops.
This parameter will have to be configured in order for any of the SMPT 
tests to succeed also see  L<"CheckRnameMail">.


=head3 ModuleMailCache

I<type>: String
I<Default>: undef

This parameter defines a path to a directory in which RNAME mail
addresses will be cached. The caching works by means of file entries
in a directory. Once the Mailchecks for a given rname have been
concluded without any errors a file named after the 'rname' is being
touched in the directory configured in ModuleMailCache. As long as
this file exists checks on that particular rname will not be done.

The filenames created in the ModuleMailCache directory are named
delcheck<hex-val> where <hex-val> is the hash value that is
constructed from the name of the zone that is tested. The content of
the files are the date the test was performed and the e-mail address
that was tested.  For instance the cache file for ripe.net would be
something like: cat /tmp/delcheck506dbfb4e32cd93a032f410f5e2672eb Thu
Mar 24 09:58:44 2005 ops.ripe.net



The Net::DelChecker module does not clean the ModuleMailCache. Use a
cronjob with a command like
  "find <ModuleMailCache> -name "delcheck*"  -ctime +<N> -exec rm {} \;"
to clean the cache. 

If left undefined no caching is performed.

=head1 SEE ALSO

=head2 RFCs

RFC 1912, Common DNS Operational and Configuration Errors, February 1996

RFC 2182 Selection and Operation of Secondary DNS Servers, July 1997

RFC 2181, Clarifications to the DNS Specification, July 1997

RFC 1123, Requirements for Internet Hosts -- Application and Support, October 1989

RFC 2317, Classless IN-ADDR.ARPA delegation, March 1998

RFC 2308, Negative Caching of DNS Queries (DNS NCACHE), March 1998

RFC 3513 Internet Protocol Version 6 (IPv6) Addressing Architecture, April 2003

RFC 1597 Address Allocation for Private Internets, March 1994

=head2 Modules

AppConfig, L<Net::IP>, L<Net::DNS>, 

=head1 LIMITATIONS

=over 4

=item *

the module cannot handle nameservers on the same IP but different
ports.

=back

=head1 TODO

=over 4

=item *

look for differing TTL values A/AAAA (watch for cached)

=item *

avoid resolving mname multiple times, look for previous answers

=item *

check zone at every IP for an NS, not just pick one.

=item *

IPv4: currently if we're checking a reverse zone and the NS name
resolves to an IP within the reverse zone, we don't complain about
missing reverse record.

This needs implementing for v6.

=item *

Test schema against output!

=item *

Check TTLs of RRsets

=item *

in sub get_zone_data_at_ip: Check the authority bit on both
queries. A problem should be flagged if both the NS and SOA queries
do not return the same authority flag


=back

=head1 AUTHOR

Lee Wilmot <lee@ripe.net>
With modifications by
Olaf Kolkman <olaf@ripe.net>



=head1 COPYRIGHT AND LICENSE

Copyright (c) 1999,2000,2001,2002,2003,2004                    RIPE NCC

All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of the author not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 



=cut
